// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
{$INCLUDE alisp3def.inc}
unit alisp3FilesU;

interface

uses
  Windows, SysUtils, Classes,
  alisp3U;


procedure ALDefineFilePrimitives ();


implementation


function ALPFOpen (n: Integer): Integer;
var
  f: Integer;
  h: THandle;
  fn, fm: string;
  acc, share, crdist: LongWord;
begin
  if not ALNextStrArg('FOPEN', n, fn, result) then exit;
  if not ALNextStrArg('FOPEN', n, fm, result) then exit;

  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  acc := 0; crdist := OPEN_EXISTING; share := FILE_SHARE_READ or FILE_SHARE_WRITE;
  for f := 1 to Length(fm) do
  begin
    case fm[f] of
      '-': share := share and (not FILE_SHARE_WRITE);
      '+': share := share and (not FILE_SHARE_READ);
      'r': acc := acc or GENERIC_READ;
      'w': acc := acc or GENERIC_WRITE;
      'c': begin crdist := CREATE_ALWAYS; acc := acc or GENERIC_WRITE; end;
    end;
  end;

  h := CreateFile(PChar(fn), acc, share, nil, crdist, 0, 0);
  if h = INVALID_HANDLE_VALUE then result := ALFalse()
  else result := ALNewIntAtom(h);
end;

function ALPFClose (n: Integer): Integer;
var
  h: Integer;
begin
  if not ALNextIntArg('FCLOSE', n, h, result) then exit;
  if CloseHandle(h) then result := ALTrue() else result := ALFalse();
end;

function ALPFTrunc (n: Integer): Integer;
var
  h: Integer;
begin
  if not ALNextIntArg('FTRUNC', n, h, result) then exit;
  if SetEndOfFile(h) then result := ALTrue() else result := ALFalse();
end;

function ALPFSize (n: Integer): Integer;
var
  oldFP: Integer;
  h: Integer;
begin
  if not ALNextIntArg('FSIZE', n, h, result) then exit;
  result := ALFalse();
  oldFP := SetFilePointer(h, 0, nil, FILE_CURRENT);
  if oldFP = -1 then exit;
  result := ALNewIntAtom(SetFilePointer(h, 0, nil, FILE_END));
  SetFilePointer(h, oldFP, nil, FILE_BEGIN);
end;

function ALPFCurPos (n: Integer): Integer;
var
  fp: Integer;
  h: Integer;
begin
  if not ALNextIntArg('FCURPOS', n, h, result) then exit;
  result := ALFalse();
  fp := SetFilePointer(h, 0, nil, FILE_CURRENT);
  if fp = -1 then exit;
  result := ALNewIntAtom(fp);
end;

function ALPFSeek (n: Integer): Integer;
var
  newFP: Integer;
  h, np, org: Integer;
begin
  if not ALNextIntArg('FSEEK', n, h, result) then exit;
  if not ALNextIntArg('FSEEK', n, np, result) then exit;
  if not ALIsNil(n) then
  begin
    if not ALNextIntArg('FSEEK', n, org, result) then exit;
  end
  else org := 1;
  if org < 0 then org := FILE_END
  else if org > 0 then org := FILE_BEGIN
  else org := FILE_CURRENT;

  result := ALFalse();
  newFP := SetFilePointer(h, np, nil, org);
  if newFP = -1 then exit;
  result := ALNewIntAtom(newFP);
end;

function ALPFReadByte (n: Integer): Integer;
var
  h: Integer;
  b: Byte;
  rs: LongWord;
  ok: Boolean;
begin
  if not ALNextIntArg('FREAD-BYTE', n, h, result) then exit;
  ok := ReadFile(h, b, 1, rs, nil);
  if ok and (rs = 1) then result := ALNewIntAtom(b) else result := ALFalse();
end;

function ALPFReadWord (n: Integer): Integer;
var
  h: Integer;
  w: Word;
  rs: LongWord;
  ok: Boolean;
begin
  if not ALNextIntArg('FREAD-WORD', n, h, result) then exit;
  ok := ReadFile(h, w, 2, rs, nil);
  if ok and (rs = 2) then result := ALNewIntAtom(w) else result := ALFalse();
end;

function ALPFReadInt (n: Integer): Integer;
var
  h: Integer;
  i: Integer;
  rs: LongWord;
  ok: Boolean;
begin
  if not ALNextIntArg('FREAD-INT', n, h, result) then exit;
  ok := ReadFile(h, i, 4, rs, nil);
  if ok and (rs = 4) then result := ALNewIntAtom(i) else result := ALFalse();
end;

function ALPFReadBuf (n: Integer): Integer;
// (fread-buf handle count [byte-buffer [shrink-flag]])
var
  len, nn: Integer;
  h: Integer;
  rs: LongWord;
  ok, shrink: Boolean;
  ptr: Pointer;
begin
  shrink := true;
  if not ALNextIntArg('FREAD-BUF', n, h, result) then exit;
  if not ALNextIntArg('FREAD-BUF', n, len, result) then exit;
  if not ALIsNil(n) then
  begin
    nn := ALCar(ALCdr(n));
    n := ALCar(n); if not ALIsNil(n) and (ALGetAtomType(n) <> DT_BYTEVECTOR) then
    begin
      result := ALThrowError('FREAD-BUF: byte vector expected', n);
      exit;
    end;
    shrink := (not ALIsNil(nn)) and (not ALIsFalse(nn));
  end;

  result := ALFalse();
  if (len < 1) or (len > 1024*1024) then exit;

  if ALIsNil(n) then
  begin
    // create new byte vector
    n := ALNewByteVector(len);
  end
  else
  begin
    // resize vector
    if ALGetVectorLength(n) < len then ALSetVectorLength(n, len)
    else if shrink and (ALGetVectorLength(n) > len) then ALSetVectorLength(n, len);
  end;
  ptr := ALGetByteVectorDataPtr(n);
  ok := ReadFile(h, ptr^, len, rs, nil);
  if ok and (rs = LongWord(len)) then result := n;
end;

function ALPFReadStr (n: Integer): Integer;
// (fread-str handle count)
var
  len: Integer;
  h: Integer;
  rs: LongWord;
  ok: Boolean;
  s: string;
begin
  if not ALNextIntArg('FREAD-STR', n, h, result) then exit;
  if not ALNextIntArg('FREAD-STR', n, len, result) then exit;

  result := ALFalse();
  if (len < 0) or (len > 1024*1024) then exit;

  SetLength(s, len);
  if len > 0 then
  begin
    ok := ReadFile(h, s[1], len, rs, nil);
    if ok and (rs = LongWord(len)) then result := ALNewStrAtom(s);
  end
  else result := ALNewStrAtom(s);
end;

function ALPFReadLn (n: Integer): Integer;
//TODO: SPEEDUP!
var
  h: Integer;
  rs: LongWord;
  ch: Char;
  ok: Boolean;
  s: string;
begin
  if not ALNextIntArg('FREADLN', n, h, result) then exit;
  s := ''; result := ALNil();
  repeat
    ok := ReadFile(h, ch, 1, rs, nil);
    if not ok or (rs <> 1) then begin if s = '' then exit else break; end;
    if ch = #0 then ch := ' ';

    if ch = #10 then break
    else if ch = #13 then
    begin
      ok := ReadFile(h, ch, 1, rs, nil);
      if ok and (rs = 1) and (ch <> #10) then
      begin
        SetFilePointer(h, -1, nil, FILE_CURRENT);
      end;
      break;
    end
    else s := s+ch;
  until false;
  result := ALNewStrAtom(s);
end;

function ALPFReadLnWhole (n: Integer): Integer;
var
  h, f: Integer;
  st: TStream;
  sl: TStringList;
  tl: TList;
begin
  if not ALNextIntArg('FREADLN-WHOLE', n, h, result) then exit;
  st := nil; sl := nil;
  try
    st := THandleStream.Create(h);
    sl := TStringList.Create();
    sl.LoadFromStream(st);
  except
    st.Free(); sl.Free(); result := ALFalse(); exit;
  end;
  st.Free();

  try
    result := ALNewVector(sl.Count);
    ALPushCell(result);
    tl := TList(ALCdr(result));
    for f := 0 to sl.Count-1 do
    begin
      h := ALNewStrAtom(sl[f]);
      //ALSetVectorItem(result, f, h);
      tl[f] := Pointer(h);
      {if f mod 128 = 1 then
        SetConsoleTitle(PChar('['+IntToStr(f)+'/'+IntToStr(sl.Count)+']'));}
    end;
    ALPopCells(1);
  finally
    sl.Free();
  end;
end;

function ALPFWriteByte (n: Integer): Integer;
var
  v: Integer;
  h: Integer;
  rs: LongWord;
  ok: Boolean;
begin
  if not ALNextIntArg('FWRITE-BYTE', n, h, result) then exit;
  if not ALNextIntArg('FWRITE-BYTE', n, v, result) then exit;
  ok := WriteFile(h, v, 1, rs, nil);
  if ok and (rs = 1) then result := ALTrue() else result := ALFalse();
end;

function ALPFWriteWord (n: Integer): Integer;
var
  v: Integer;
  h: Integer;
  rs: LongWord;
  ok: Boolean;
begin
  if not ALNextIntArg('FWRITE-WORD', n, h, result) then exit;
  if not ALNextIntArg('FWRITE-WORD', n, v, result) then exit;
  ok := WriteFile(h, v, 2, rs, nil);
  if ok and (rs = 2) then result := ALTrue() else result := ALFalse();
end;

function ALPFWriteInt (n: Integer): Integer;
var
  v: Integer;
  h: Integer;
  rs: LongWord;
  ok: Boolean;
begin
  if not ALNextIntArg('FWRITE-INT', n, h, result) then exit;
  if not ALNextIntArg('FWRITE-INT', n, v, result) then exit;
  ok := WriteFile(h, v, 4, rs, nil);
  if ok and (rs = 4) then result := ALTrue() else result := ALFalse();
end;

function ALPFWriteStr (n: Integer): Integer;
var
  h: Integer;
  rs: LongWord;
  ok: Boolean;
  s: string;
begin
  if not ALNextIntArg('FWRITE-STR', n, h, result) then exit;
  if not ALNextStrArg('FWRITE-STR', n, s, result) then exit;
  if s = '' then begin result := ALTrue(); exit; end;
  ok := WriteFile(h, s[1], Length(s), rs, nil);
  if ok and (rs = LongWord(Length(s))) then result := ALTrue() else result := ALFalse();
end;

function ALPFWriteLn (n: Integer): Integer;
var
  h: Integer;
  rs: LongWord;
  ok: Boolean;
  s: string;
  w: Word;
begin
  if not ALNextIntArg('FWRITELN', n, h, result) then exit;
  if not ALNextStrArg('FWRITELN', n, s, result) then exit;
  if s = '' then begin ok := true; rs := 0; end
  else
  begin
    ok := WriteFile(h, s[1], Length(s), rs, nil);
  end;
  w := $0A0D;
  if ok and (rs = LongWord(Length(s))) then ok := WriteFile(h, w, 2, rs, nil);
  if ok and (rs = 2) then result := ALTrue() else result := ALFalse();
end;

function ALPFWriteBuf (n: Integer): Integer;
// (write-buf handle count byte-buffer)
// count < 0: the whole buffer
var
  len: Integer;
  h: Integer;
  rs: LongWord;
  ok: Boolean;
  ptr: Pointer;
begin
  if not ALNextIntArg('FWRITE-BUF', n, h, result) then exit;
  if not ALNextIntArg('FWRITE-BUF', n, len, result) then exit;
  n := ALCar(n); if ALGetAtomType(n) <> DT_BYTEVECTOR then
  begin
    result := ALThrowError('FWRITE-BUF: byte vector expected', n);
    exit;
  end;

  result := ALFalse();

  if (len < 0) then len := ALGetVectorLength(n);
  if (len < 1) or (len > ALGetVectorLength(n)) then exit;
  ptr := ALGetByteVectorDataPtr(n);
  ok := WriteFile(h, ptr^, len, rs, nil);
  if ok and (rs = LongWord(len)) then result := ALTrue();
end;


function ALPDeleteFile (n: Integer): Integer;
var
  fn: string;
begin
  if not ALNextStrArg('DELETE-FILE', n, fn, result) then exit;
  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  if DeleteFile(fn) then result := ALTrue() else result := ALFalse();
end;


procedure ALDefineFilePrimitives ();
begin
  ALAddPrimitive('fopen', ALPFOpen, false, 2, 2);
  ALAddPrimitive('fclose', ALPFClose, false, 1, 1);
  ALAddPrimitive('ftrunc', ALPFTrunc, false, 1, 1);
  ALAddPrimitive('fsize', ALPFSize, false, 1, 1);
  ALAddPrimitive('fcurpos', ALPFCurPos, false, 1, 1);
  ALAddPrimitive('fseek', ALPFSeek, false, 2, 3);
  ALAddPrimitive('fread-byte', ALPFReadByte, false, 1, 1);
  ALAddPrimitive('fread-word', ALPFReadWord, false, 1, 1);
  ALAddPrimitive('fread-int', ALPFReadInt, false, 1, 1);
  ALAddPrimitive('fread-buf', ALPFReadBuf, false, 2, 4);
  ALAddPrimitive('fread-str', ALPFReadStr, false, 2, 2);
  ALAddPrimitive('freadln', ALPFReadLn, false, 1, 1);
  ALAddPrimitive('freadln-whole', ALPFReadLnWhole, false, 1, 1);
  ALAddPrimitive('fwrite-byte', ALPFWriteByte, false, 2, 2);
  ALAddPrimitive('fwrite-word', ALPFWriteWord, false, 2, 2);
  ALAddPrimitive('fwrite-int', ALPFWriteInt, false, 2, 2);
  ALAddPrimitive('fwrite-str', ALPFWriteStr, false, 2, 2);
  ALAddPrimitive('fwrite-buf', ALPFWriteBuf, false, 3, 3);
  ALAddPrimitive('fwriteln', ALPFWriteLn, false, 2, 2);
  ALAddPrimitive('delete-file', ALPDeleteFile, false, 1, 1);
end;


begin
  {$IFNDEF ALISP3_MULTIENGINE}
  ALDefineFilePrimitives();
  {$ENDIF}
end.
