// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
{$INCLUDE alisp3def.inc}
{$TYPEDADDRESS OFF}
unit alisp3RegU;

interface

uses
  Windows, SysUtils,
  alisp3U;


procedure ALDefineRegPrimitives ();


implementation


procedure ParseKeyName (var kn: string; out hk: HKEY; out vname: string);
var
  s: string;
  f: Integer;
begin
  vname := '';
  kn := StringReplace(kn, '/', '\', [rfReplaceAll]);
  f := Pos('\', kn);
  if f > 0 then
  begin
    s := Copy(kn, 1, f-1); Delete(kn, 1, f);
  end
  else s := '';

  if Comparetext(s, 'HKCR') = 0 then hk := HKEY_CLASSES_ROOT
  else if Comparetext(s, 'HKEY_CLASSES_ROOT') = 0 then hk := HKEY_CLASSES_ROOT
  else if Comparetext(s, 'HKCU') = 0 then hk := HKEY_CURRENT_USER
  else if Comparetext(s, 'HKEY_CURRENT_USER') = 0 then hk := HKEY_CURRENT_USER
  else if Comparetext(s, 'HKLM') = 0 then hk := HKEY_LOCAL_MACHINE
  else if Comparetext(s, 'HKEY_LOCAL_MACHINE') = 0 then hk := HKEY_LOCAL_MACHINE
  else if Comparetext(s, 'HKU') = 0 then hk := HKEY_USERS
  else if Comparetext(s, 'HKEY_USERS') = 0 then hk := HKEY_USERS
  else if Comparetext(s, 'HKPD') = 0 then hk := HKEY_PERFORMANCE_DATA
  else if Comparetext(s, 'HKEY_PERFORMANCE_DATA') = 0 then hk := HKEY_PERFORMANCE_DATA
  else if Comparetext(s, 'HKEY_PERF_DATA') = 0 then hk := HKEY_PERFORMANCE_DATA
  else if Comparetext(s, 'HKCC') = 0 then hk := HKEY_CURRENT_CONFIG
  else if Comparetext(s, 'HKEY_CURRENT_CONFIG') = 0 then hk := HKEY_CURRENT_CONFIG
  else if Comparetext(s, 'HKDD') = 0 then hk := HKEY_DYN_DATA
  else if Comparetext(s, 'HKEY_DYN_DATA') = 0 then hk := HKEY_DYN_DATA
  else
  begin
    hk := HKEY_CURRENT_USER;
    if s <> '' then kn := s+'\'+kn;
  end;

  vname := ExtractFileName(kn); kn := ExtractFileDir(kn);
  if kn = '' then begin kn := vname+'\'; kn := ''; end;
  //kn := kn+'\';
end;

function TypeToStr (ktype: LongWord): string;
begin
  case ktype of
    REG_BINARY: result := 'REG_BINARY';
    REG_DWORD: result := 'REG_DWORD';
    //REG_DWORD_LITTLE_ENDIAN: result := 'REG_DWORD_LITTLE_ENDIAN';
      // same as REG_DWORD in i86
    REG_DWORD_BIG_ENDIAN: result := 'REG_DWORD_BIG_ENDIAN';
    REG_EXPAND_SZ: result := 'REG_EXPAND_SZ';
    REG_LINK: result := 'REG_LINK';
    REG_MULTI_SZ: result := 'REG_MULTI_SZ';
    REG_NONE: result := 'REG_NONE';
    REG_RESOURCE_LIST: result := 'REG_RESOURCE_LIST';
    REG_SZ: result := 'REG_SZ';
    REG_FULL_RESOURCE_DESCRIPTOR: result := 'REG_FULL_RESOURCE_DESCRIPTOR';
    REG_RESOURCE_REQUIREMENTS_LIST: result := 'REG_RESOURCE_REQUIREMENTS_LIST';
    else result := IntToStr(ktype);
  end;
end;

function StrToType (tn: string): LongWord;
var
  e: Integer;
begin
  if CompareText(tn, 'REG_BINARY') = 0 then result := REG_BINARY
  else if CompareText(tn, 'REG_DWORD') = 0 then result := REG_DWORD
  //else if CompareText(tn, 'REG_DWORD_LITTLE_ENDIAN') = 0 then result := REG_DWORD_LITTLE_ENDIAN
    // same as REG_DWORD in i86
  else if CompareText(tn, 'REG_DWORD_BIG_ENDIAN') = 0 then result := REG_DWORD_BIG_ENDIAN
  else if CompareText(tn, 'REG_EXPAND_SZ') = 0 then result := REG_EXPAND_SZ
  else if CompareText(tn, 'REG_LINK') = 0 then result := REG_LINK
  else if CompareText(tn, 'REG_MULTI_SZ') = 0 then result := REG_MULTI_SZ
  else if CompareText(tn, 'REG_NONE') = 0 then result := REG_NONE
  else if CompareText(tn, 'REG_RESOURCE_LIST') = 0 then result := REG_RESOURCE_LIST
  else if CompareText(tn, 'REG_SZ') = 0 then result := REG_SZ
  else if CompareText(tn, 'REG_FULL_RESOURCE_DESCRIPTOR') = 0 then result := REG_FULL_RESOURCE_DESCRIPTOR
  else if CompareText(tn, 'REG_RESOURCE_REQUIREMENTS_LIST') = 0 then result := REG_RESOURCE_REQUIREMENTS_LIST
  else
  begin
    Val(tn, result, e);
    if e <> 0 then result := REG_NONE;
  end;
end;

function ALPKeySize (n: Integer): Integer;
// (key-size pathname)
// result: int or *f*
var
  kn, vn: string;
  kh: HKEY;
  hk: HKEY;
  ktype: LongWord;
  dSize: LongWord;
begin
  if not ALNextStrArg('REG::KEY-SIZE', n, kn, result) then exit;
  result := ALFalse();
  ParseKeyName(kn, hk, vn);

  if RegOpenKeyEx(hk, PChar(kn), 0, KEY_READ, kh) <> ERROR_SUCCESS then exit;
  try
    dSize := 0;
    if RegQueryValueEx(kh, PChar(vn), nil, @ktype, nil, @dSize) <> ERROR_SUCCESS then exit;
    result := ALNewIntAtom(Integer(dSize));
  finally
    RegCloseKey(kh);
  end;
end;

function ALPKeyType (n: Integer): Integer;
// (key-type pathname)
// result: string or *f*
var
  kn, vn: string;
  kh: HKEY;
  hk: HKEY;
  ktype: LongWord;
  dSize: LongWord;
begin
  if not ALNextStrArg('REG::KEY-TYPE', n, kn, result) then exit;
  result := ALFalse();
  ParseKeyName(kn, hk, vn);

  if RegOpenKeyEx(hk, PChar(kn), 0, KEY_READ, kh) <> ERROR_SUCCESS then exit;
  try
    dSize := 0;
    if RegQueryValueEx(kh, PChar(vn), nil, @ktype, nil, @dSize) <> ERROR_SUCCESS then exit;
    result := ALNewStrAtom(TypeToStr(ktype));
  finally
    RegCloseKey(kh);
  end;
end;

function ALPGetKey (n: Integer): Integer;
// (key pathname)
// result: (valuestr . typestr) or *f*
var
  kn, vn: string;
  kh: HKEY;
  hk: HKEY;
  ktype: LongWord;
  data: packed array of Byte;
  dSize: LongWord;
  f: Integer;
begin
  SetLength(data, 0);
  if not ALNextStrArg('REG::KEY', n, kn, result) then exit;
  result := ALFalse();
  ParseKeyName(kn, hk, vn);

  if RegOpenKeyEx(hk, PChar(kn), 0, KEY_READ, kh) <> ERROR_SUCCESS then
  begin
    if RegOpenKeyEx(hk, PChar(kn), 0, KEY_QUERY_VALUE, kh) <> ERROR_SUCCESS then exit;
  end;
  try
    dSize := 0;
    if RegQueryValueEx(kh, PChar(vn), nil, @ktype, nil, @dSize) <> ERROR_SUCCESS then exit;
    if dSize > 0 then
    begin
      SetLength(data, dSize);
      if RegQueryValueEx(kh, PChar(vn), nil, @ktype, @(data[0]), @dSize) <> ERROR_SUCCESS then exit;
    end;
    result := ALNewStrAtom(TypeToStr(ktype)); ALPushCell(result);
    if ((ktype = REG_DWORD) or (ktype = REG_DWORD_BIG_ENDIAN)) and (dSize = 4) then
    begin
      Move(data[0], dSize, 4);
      if ktype = REG_DWORD_BIG_ENDIAN then
      begin
        asm
          mov   eax,dSize
          bswap eax
          mov   dSize,eax
        end;
      end;
      kn := IntToStr(dSize);
    end
    else
    begin
      SetLength(kn, dSize);
      if dSize > 0 then Move(data[0], kn[1], dSize);
      if (ktype = REG_SZ) or (ktype = REG_EXPAND_SZ) then
      begin
        f := 0; while (f <= High(data)) and (data[f] <> 0) do Inc(f);
        kn := Copy(kn, 1, f);
      end
      else if ktype = REG_MULTI_SZ then
      begin
        f := 0;
        while f <= High(data) do
        begin
          if f < High(data) then
          begin
            if (data[f] = 0) and (data[f+1] = 0) then
            begin
              Inc(f); break;
            end;
          end;
          Inc(f);
        end;
        kn := Copy(kn, 1, f);
      end;
    end;
    result := ALNewStrAtom(kn);
    result := ALCons(result, ALPopCells(1));
  finally
    RegCloseKey(kh);
  end;
end;

function ALPSetKey (n: Integer): Integer;
// (set-key pathname value [type])
// default type is REG_SZ for string and REG_DWORD for int
// type can be string or integer
// result: *t* or *f*
var
  kn, vn: string;
  vals: string;
  kh: HKEY;
  hk: HKEY;
  ktype: LongWord;
  dSize: LongWord;
  f: Integer;
begin
  if not ALNextStrArg('REG::SET-KEY', n, kn, result) then exit;
  if ALGetatomType(ALCar(n)) = DT_STRING then
  begin
    if not ALNextStrArg('REG::SET-KEY', n, vals, result) then exit;
    ktype := REG_SZ;
  end
  else
  begin
    if not ALNextIntArg('REG::SET-KEY', n, f, result) then exit;
    SetLength(vals, 4); Move(f, vals[1], 4);
    ktype := REG_DWORD;
  end;
  if not ALIsNil(n) then
  begin
    if ALGetatomType(ALCar(n)) = DT_STRING then
    begin
      if not ALNextStrArg('REG::SET-KEY', n, vn, result) then exit;
      ktype := StrToType(vn);
    end
    else
    begin
      if not ALNextIntArg('REG::SET-KEY', n, Integer(ktype), result) then exit;
    end;
  end;
  result := ALFalse();
  ParseKeyName(kn, hk, vn);

  if RegCreateKeyEx(hk, PChar(kn), 0, nil, REG_OPTION_NON_VOLATILE,
    KEY_WRITE, nil, kh, @dSize) <> ERROR_SUCCESS then
  begin
    if RegCreateKeyEx(hk, PChar(kn), 0, nil, REG_OPTION_NON_VOLATILE,
      KEY_SET_VALUE, nil, kh, @dSize) <> ERROR_SUCCESS then exit;
  end;
  try
    if ((ktype = REG_DWORD) or (ktype = REG_DWORD_BIG_ENDIAN)) and (Length(vals) <> 4) then exit;
    if ktype = REG_DWORD_BIG_ENDIAN then
    begin
      Move(vals[1], f, 4);
      asm
        mov   eax,f
        bswap eax
        mov   f,eax
      end;
      UniqueString(vals); Move(f, vals[1], 4);
    end
    else if (ktype = REG_SZ) or (ktype = REG_EXPAND_SZ) or (ktype = REG_MULTI_SZ) then
    begin
      vals := vals+#0;
    end;
    if vals = '' then vals := ' ';
    if RegSetValueEx(kh, PChar(vn), 0, ktype, @(vals[1]), Length(vals)) <> ERROR_SUCCESS then exit;
    result := ALTrue();
  finally
    RegCloseKey(kh);
  end;
end;

function ALPDeleteValue (n: Integer): Integer;
// (delete-value pathname)
// result: *t* or *f*
var
  kn, vn: string;
  kh: HKEY;
  hk: HKEY;
begin
  if not ALNextStrArg('REG::DELETE-VALUE', n, kn, result) then exit;
  result := ALFalse();
  ParseKeyName(kn, hk, vn);
  if RegOpenKeyEx(hk, PChar(kn), 0, KEY_SET_VALUE, kh) <> ERROR_SUCCESS then exit;
  try
    if RegDeleteValue(kh, PChar(vn)) = ERROR_SUCCESS then result := ALTrue();
  finally
    RegCloseKey(kh);
  end;
end;

function ALPDeleteKey (n: Integer): Integer;
// (delete-key pathname)
// result: *t* or *f*
var
  kn, vn: string;
  kh: HKEY;
  hk: HKEY;
begin
  if not ALNextStrArg('REG::DELETE-KEY', n, kn, result) then exit;
  result := ALFalse();
  ParseKeyName(kn, hk, vn);
  if RegOpenKeyEx(hk, PChar(kn), 0, KEY_WRITE, kh) <> ERROR_SUCCESS then exit;
  try
    if RegDeleteKey(kh, PChar(vn)) = ERROR_SUCCESS then result := ALTrue();
  finally
    RegCloseKey(kh);
  end;
end;

function ALPEnumKeys (n: Integer): Integer;
// (enum-keys pathname)
// result: list or *f*
// list item: keyname
var
  kn, vn: string;
  kh: HKEY;
  hk: HKEY;
  index: Integer;
  nameBuf: packed array of Char;
  nbsz: LongWord;
  ft: Windows.TFileTime;
begin
  SetLength(nameBuf, 0);
  if not ALNextStrArg('REG::ENUM-KEYS', n, kn, result) then exit;
  if (kn <> '') and not (kn[Length(kn)] in ['/','\']) then kn := kn+'\';
  result := ALFalse();
  ParseKeyName(kn, hk, vn);
  if RegOpenKeyEx(hk, PChar(kn), 0, KEY_ENUMERATE_SUB_KEYS, kh) <> ERROR_SUCCESS then exit;
  try
    index := 0;
    SetLength(nameBuf, 64);
    n := ALCons(ALNil(), ALNil()); ALPushCell(n);
    repeat
      nbsz := Length(nameBuf);
      case RegEnumKeyEx(kh, index, @(nameBuf[0]), nbsz, nil, nil, nil, @ft) of
        ERROR_NO_MORE_ITEMS: break;
        ERROR_MORE_DATA: SetLength(nameBuf, Length(nameBuf)+64);
        ERROR_SUCCESS:
          begin
            SetString(kn, PChar(@(nameBuf[0])), nbsz);
            ALSetCdr(n, ALCons(ALNewStrAtom(kn), ALNil())); n := ALCdr(n);
            Inc(index);
          end;
        else begin ALPopCells(1); exit; end;
      end;
    until false;
    result := ALCdr(ALPopCells(1));
  finally
    RegCloseKey(kh);
  end;
end;

function ALPEnumValues (n: Integer): Integer;
// (enum-values pathname)
// result: list or *f*
// list item: valname
var
  kn, vn: string;
  kh: HKEY;
  hk: HKEY;
  index: Integer;
  nameBuf: packed array of Char;
  nbsz: LongWord;
begin
  SetLength(nameBuf, 0);
  if not ALNextStrArg('REG::ENUM-VALUES', n, kn, result) then exit;
  if (kn <> '') and not (kn[Length(kn)] in ['/','\']) then kn := kn+'\';
  result := ALFalse();
  ParseKeyName(kn, hk, vn);
  if RegOpenKeyEx(hk, PChar(kn), 0, KEY_QUERY_VALUE, kh) <> ERROR_SUCCESS then exit;
  try
    index := 0;
    SetLength(nameBuf, 64);
    n := ALCons(ALNil(), ALNil()); ALPushCell(n);
    repeat
      nbsz := Length(nameBuf);
      case RegEnumValue(kh, index, @(nameBuf[0]), nbsz, nil, nil, nil, nil) of
        ERROR_NO_MORE_ITEMS: break;
        ERROR_MORE_DATA: SetLength(nameBuf, Length(nameBuf)+64);
        ERROR_SUCCESS:
          begin
            SetString(kn, PChar(@(nameBuf[0])), nbsz);
            ALSetCdr(n, ALCons(ALNewStrAtom(kn), ALNil())); n := ALCdr(n);
            Inc(index);
          end;
        else begin ALPopCells(1); exit; end;
      end;
    until false;
    result := ALCdr(ALPopCells(1));
  finally
    RegCloseKey(kh);
  end;
end;


procedure ALDefineRegPrimitives ();
begin
  ALAddPrimitive('reg::key', ALPGetKey, false, 1, 1);
  ALAddPrimitive('reg::key-size', ALPKeySize, false, 1, 1);
  ALAddPrimitive('reg::key-type', ALPKeyType, false, 1, 1);
  ALAddPrimitive('reg::set-key', ALPSetKey, false, 2, 3);
  ALAddPrimitive('reg::delete-value', ALPDeleteValue, false, 1, 1);
  ALAddPrimitive('reg::delete-key', ALPDeleteKey, false, 1, 1);
  ALAddPrimitive('reg::enum-keys', ALPEnumKeys, false, 1, 1);
  ALAddPrimitive('reg::enum-values', ALPEnumValues, false, 1, 1);
end;


begin
  {$IFNDEF ALISP3_MULTIENGINE}
  ALDefineRegPrimitives();
  {$ENDIF}
end.
