// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
function LispLoad (n: Integer; check: Boolean; checkid: string): Integer;
var
  s, t: string;
  oalist: Integer;
begin
  if ALGetAtomType(n) <> DT_STRING then
  begin
    if check then result := ALThrowError('CLOAD: string expected', n)
    else result := ALThrowError('LOAD: string expected', n);
    exit;
  end;

  s := StringReplace(al_StringPool[al_CDR[n].vi], '/', '\', [rfReplaceAll]);
  if ExtractFileExt(s) = '' then s := s+'.l';
  if not FileExists(s) then
  begin
    t := FileSearchEx(s, al_DefPathList);
    if t <> '' then s := t;
  end;

  result := al_strue;
  if check then
  begin
    if checkid = '' then t := ANSILowerCase(ChangeFileExt(ExtractFileName(s), ''))
    else t := checkid;
    if (t <> '') and (ALFindGlobalSymbol(t) <> al_Nil) then exit;
  end;

  // let ALISP3 think that we are at the top level
  oalist := al_Frames; ALPushCell(oalist); al_Frames := al_Nil;
  try
    result := ALLoadFile(s);
  except
    if check then result := ALThrowError('CLOAD: exception')
    else result := ALThrowError('LOAD: exception');
  end;
  if al_BreakFlag <> ALBREAK_ERROR then
  begin
    al_Frames := ALPopCells(1);
  end;
  if (al_BreakFlag <> ALBREAK_THROW) and
     (al_BreakFlag <> ALBREAK_ERROR) and
     (al_BreakFlag <> ALBREAK_QUIT) and
     (al_BreakFlag <> ALBREAK_NONE) then
  begin
    if check then result := ALThrowError('CLOAD: error')
    else result := ALThrowError('LOAD: error');
    //alBreakFlag := 0;
    //result := s_true;
  end;
end;

function ALPLoad (n: Integer): Integer;
// (load fname)
begin
  result := LispLoad(al_CAR[n], false, '');
end;

function ALPCLoad (n: Integer): Integer;
// (cload fname [id])
begin
  result := al_CAR[al_CDR[n].vi];
  if result <> al_Nil then
  begin
    if ALGetAtomType(result) <> DT_STRING then
    begin
      result := ALThrowError('CLOAD: id expected');
      exit;
    end;
    result := al_CDR[result].vi;
  end
  else result := 0;
  result := LispLoad(al_CAR[n], true, al_StringPool[result]);
end;

function ALPVerbLoad (n: Integer): Integer;
// (sys-verbose-load [flag])
begin
  if al_VerbLoadFlag then result := al_strue else result := al_sfalse;
  if n <> al_Nil then
  begin
    n := al_CAR[n];
    al_VerbLoadFlag := (n <> al_Nil) and (n <> al_sfalse);
  end;
end;


function ALPList (n: Integer): Integer;
// (list a ...)
begin
  result := n;
end;

function ALPCons (n: Integer): Integer;
// (cons a b)
begin
  result := ALCons(al_CAR[n], al_CAR[al_CDR[n].vi]);
end;

function ALPCar (n: Integer): Integer;
// (car a)
var
  nn: Integer;
begin
  nn := al_CAR[n];
  if (nn = al_Nil) or ALIsAtom(nn) then result := ALThrowError('car: bad pair', n)
  else result := al_CAR[nn];
end;

function ALPCdr (n: Integer): Integer;
// (cdr a)
var
  nn: Integer;
begin
  nn := al_CAR[n];
  if (nn = al_Nil) or ALIsAtom(nn) then result := ALThrowError('cdr: bad pair', n)
  else result := al_CDR[nn].vi;
end;

function ALPEnum (n: Integer): Integer;
// (enum name <num> name ...)
// no evaluated args here
var
  inum, c: Integer;
begin
  result := al_Nil; inum := 0;
  while n <> al_Nil do
  begin
    c := al_CAR[n]; n := al_CDR[n].vi;
    if ALGetAtomType(c) <> DT_STRING then c := ALEval(c);
    if al_BreakFlag <> ALBREAK_NONE then begin result := c; exit; end;
    case ALGetAtomType(c) of
      DT_STRING:
        begin
          ALSetSymbolEx(al_StringPool[al_CDR[c].vi],
            ALNewIntAtom(inum), al_CDR[c].vi, true, true, true);
          Inc(inum);
        end;
      DT_INTEGER: inum := al_CDR[c].vi;
      else begin result := ALThrowError('ENUM: identifier or int expected', c); exit; end;
    end;
    //if al_BreakFlag <> ALBREAK_NONE then exit; // breaks?
  end;
end;

function ALPUndefine (n: Integer): Integer;
// (undefine name [name ...])
var
  c: Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    c := al_CAR[n]; n := al_CDR[n].vi;
    if ALGetAtomType(c) <> DT_STRING then begin result := ALThrowError('UNDEFINE: identifier expected', c); exit; end;
    ALDeleteSymbolFromCurrentContext(al_StringPool[al_CDR[c].vi]);
  end;
end;

function ALPSet (n: Integer): Integer;
// (set name value <name value>)
// all args already evaluated
var
  c: Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    c := al_CAR[n];
    if ALGetAtomType(c) <> DT_STRING then begin result := ALThrowError('SET: identifier expected', c); exit; end;
    n := al_CDR[n].vi;
    if n = al_Nil then begin result := ALThrowError('SET: missing value', c); exit; end;
    result := al_CAR[n];
    //ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, true);
    ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, false, true);
    n := al_CDR[n].vi;
  end;
end;

function ALPSetQ (n: Integer): Integer;
// (setq name value <name value>)
// no evaluated args here
var
  c: Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    c := al_CAR[n];
    if ALGetAtomType(c) <> DT_STRING then begin result := ALThrowError('SETQ: identifier expected', c); exit; end;
    n := al_CDR[n].vi;
    if n = al_Nil then begin result := ALThrowError('SETQ: missing value', c); exit; end;
    //WriteLn('SETQ: "', stringPool[cdr[c]], '"');
    //Write('SETQ_x: '); ALPrintNoQ(eng car[n]); WriteLn;
    result := ALEval(al_CAR[n]); // evaluate value
    //Write('SETQ_v: '); ALPrintNoQ(eng result); WriteLn;
    if (al_BreakFlag = ALBREAK_ERROR) or (al_BreakFlag = ALBREAK_THROW) then exit;
    //ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, true);
    ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, false, true);
    if al_BreakFlag <> ALBREAK_NONE then exit; // breaks?
    n := al_CDR[n].vi;
  end;
end;

function ALPUpSet (n: Integer): Integer;
// (up-set name value <name value>)
// all args already evaluated
var
  c: Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    c := al_CAR[n];
    if ALGetAtomType(c) <> DT_STRING then begin result := ALThrowError('UP-SET: identifier expected', c); exit; end;
    n := al_CDR[n].vi;
    if n = al_Nil then begin result := ALThrowError('UP-SET: missing value', c); exit; end;
    result := al_CAR[n];
    ALPushCell(al_Frames); al_Frames := al_CDR[al_Frames].vi;
    //ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, true);
    ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, false, true);
    al_Frames := ALPopCells(1);
    n := al_CDR[n].vi;
  end;
end;

function ALPUpSetQ (n: Integer): Integer;
// (up-setq name value <name value>)
// no evaluated args here
var
  c: Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    c := al_CAR[n];
    if ALGetAtomType(c) <> DT_STRING then begin result := ALThrowError('UP-SETQ: identifier expected', c); exit; end;
    n := al_CDR[n].vi;
    if n = al_Nil then begin result := ALThrowError('UP-SETQ: missing value', c); exit; end;
    result := ALEval(al_CAR[n]); // evaluate value
    if (al_BreakFlag = ALBREAK_ERROR) or (al_BreakFlag = ALBREAK_THROW) then exit;
    ALPushCell(al_Frames); al_Frames := al_CDR[al_Frames].vi;
    //ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, true);
    ALSetSymbolEx(al_StringPool[al_CDR[c].vi], result, al_CDR[c].vi, false, true);
    al_Frames := ALPopCells(1);
    if al_BreakFlag <> ALBREAK_NONE then exit; // breaks?
    n := al_CDR[n].vi;
  end;
end;

function DefunOrMacro (n: Integer; macro: Boolean): Integer;
// (defun name (args) body)
// (macro name (args) body)
// no evaluated args here
// result: lambda-list
var
  c, si: Integer;
begin
  c := al_CAR[n];
  if ALGetAtomType(c) <> DT_STRING then begin result := ALThrowError('DEFUN/MACRO: identifier expected', c); exit; end;
  // build lambda list
  si := al_CDR[c].vi;
  if macro then result := ALCons(al_smlambda, al_CDR[n].vi)
  else result := ALCons(al_slambda, al_CDR[n].vi);
  result := ALEval(result);  // optimize lambda
  if al_BreakFlag = ALBREAK_NONE then
  begin
    result := al_CAR[ALSetSymbolEx(al_StringPool[si], result,
      -1{al_CDR[c].vi}, true, true, true)];
  end;
end;

function ALPDefun (n: Integer): Integer;
// (defun name (args) body)
// no evaluated args here
// result: lambda-list
begin
  result := DefunOrMacro(n, false);
end;

function ALPMacro (n: Integer): Integer;
// (macro name (args) body)
// no evaluated args here
// result: lambda-list
begin
  result := DefunOrMacro(n, true);
end;

function ALPBegin (n: Integer): Integer;
// (begin expr [expr ...])
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    result := al_CAR[n]; n := al_CDR[n].vi;
  end;
end;

function LPrint (n: Integer; nl: Boolean): Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]); n := al_CDR[n].vi;
    if (al_BreakFlag = ALBREAK_ERROR) or (al_BreakFlag = ALBREAK_THROW) then exit;
    ALPrint(result); Write(' ');
    if al_BreakFlag <> ALBREAK_NONE then break;
  end;
  if nl then WriteLn;
end;

function Print (n: Integer; nl: Boolean): Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]); n := al_CDR[n].vi;
    if (al_BreakFlag = ALBREAK_ERROR) or (al_BreakFlag = ALBREAK_THROW) then exit;
    Write(ALGetAtomStr(result));
    if al_BreakFlag <> ALBREAK_NONE then break;
  end;
  if nl then WriteLn;
end;

function ALPLPrint (n: Integer): Integer;
// (lprint expr-list)
begin
  result := LPrint(n, false);
end;

function ALPLPrintLn (n: Integer): Integer;
// (lprintln expr-list)
begin
  result := LPrint(n, true);
end;

function ALPPrint (n: Integer): Integer;
// (print expr-list)
begin
  result := Print(n, false);
end;

function ALPPrintLn (n: Integer): Integer;
// (println expr-list)
begin
  result := Print(n, true);
end;

function ALPLPrintToStr (n: Integer): Integer;
// (lprint-to-str expr-list)
var
  s: string;
begin
  s := '';
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]); n := al_CDR[n].vi;
    if (al_BreakFlag = ALBREAK_ERROR) or (al_BreakFlag = ALBREAK_THROW) then exit;
    s := s+ALPrintToStr(result)+' ';
    if al_BreakFlag <> ALBREAK_NONE then break;
  end;
  result := ALNewStrAtom(s);
end;


function ALPCatch (n: Integer): Integer;
// (catch expr-list)
begin
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]);
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag = ALBREAK_THROW then al_BreakFlag := ALBREAK_NONE;
      exit;
    end;
    n := al_CDR[n].vi;
  end;
  result := al_Nil;
end;

function ALPIf (n: Integer): Integer;
// (if cond-expr cond-body [cond-expr cond-body]* [else-body])
begin
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]); n := al_CDR[n].vi; // cond
    if al_BreakFlag <> ALBREAK_NONE then exit;
    if n = al_Nil then exit; // no body (else)?
    if (result = al_Nil) or (result = al_sfalse) then n := al_CDR[n].vi // skip body
    else
    begin
      // eval body and exit
      result := ALEval(al_CAR[n]);
      exit;
    end;
  end;
  result := al_Nil;
end;

function ALPSwitch (n: Integer): Integer;
// (switch expr  cond-expr cond-body [cond-expr cond-body]* [else-body])
var
  typ: Integer;
  ni: Integer;
  nf: TALFloat;
  xn, cnd: Integer;
  ok, testBool: Boolean;
begin
  {nn := n; }result := ALEval(al_CAR[n]); n := al_CDR[n].vi;
  if al_BreakFlag <> ALBREAK_NONE then exit;

  ALPushCell(result); cnd := result;
  typ := ALGetAtomType(result);
  case typ of
    DT_INTEGER: begin ni := al_CDR[result].vi; nf := 0; end;
    DT_FLOAT: begin nf := al_CDR[result].vf; ni := 0; end;
    DT_STRING: begin ni := al_CDR[result].vi; nf := 0; end;
    DT_VECTOR, DT_BYTEVECTOR, DT_TOBJECT: begin ni := al_CDR[result].vi; nf := 0; end;
    else begin ni := result; nf := 0; end;
  end;

  //ok := false;
  while n <> al_Nil do
  begin
    xn := al_CAR[n];
    if ALGetAtomType(xn) = -1 then
    begin
      //Write('non-atom: '); ALPrintNoQ(xn); WriteLn;
      // not an atom, so build the function call
      if al_CAR[xn] = al_slambda then
      begin
        result := ALCons(al_squote, ALCons(cnd, al_Nil));
        result := ALCons(xn, ALCons(result, al_Nil));
        testBool := true;
      end
      else if al_CAR[xn] = al_smlambda then
      begin
        result := ALCons(xn, ALCons(cnd, al_Nil));
        testBool := true;
      end
      else if al_CDR[xn].vi <> al_Nil then
      begin
        // list of several elements: have to evaluate it
        result := xn;
        testBool := false;
      end
      else
      begin
        // list of one element -- function name
        result := ALCons(al_squote, ALCons(cnd, al_Nil));
        result := ALCons(al_CAR[xn], ALCons(result, al_Nil));
        testBool := true;
      end;
      result := ALEval(result);
    end
    else
    begin
      //Write('atom: '); ALPrintNoQ(xn); WriteLn;
      result := ALEval(xn);
      testBool := false;
    end;
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag <> ALBREAK_ERROR then ALPopCells(1);
      exit;
    end;
    n := al_CDR[n].vi; // body
    if n = al_Nil then
    begin
      ALPopCells(1);
      exit; // was else
    end;
    if testBool then
    begin
      ok := (result <> al_Nil) and (result <> al_sfalse);
    end
    else
    begin
      case ALGetAtomType(result) of
        DT_INTEGER:
          if typ = DT_INTEGER then ok := (ni = al_CDR[result].vi)
          else ok := false;
        DT_FLOAT:
          case typ of
            DT_INTEGER: ok := (ni = al_CDR[result].vf);
            DT_FLOAT: ok := (nf = al_CDR[result].vf);
            else ok := false;
          end;
        DT_STRING:
          if typ = DT_STRING then ok := (ni = al_CDR[result].vi)
          else ok := false;
        DT_VECTOR, DT_BYTEVECTOR, DT_TOBJECT: ok := (ni = al_CDR[result].vi);
        else ok := (ni = result);
      end;
    end;
    if not ok then n := al_CDR[n].vi
    else
    begin
      // found!
      ALPopCells(1);
      result := ALEval(al_CAR[n]);
      exit;
    end;
  end;
  ALPopCells(1);
  result := al_Nil;
end;

function ALPFor (n: Integer): Integer;
// (for init-expr cont-expr inc-expr [body-expr]*)
label forcond;
var
  ce, ie, nn: Integer;
begin
  result := ALEval(al_CAR[n]); if al_BreakFlag <> ALBREAK_NONE then exit; // init
  n := al_CDR[n].vi; ce := al_CAR[n];
  n := al_CDR[n].vi; ie := al_CAR[n];
  n := al_CDR[n].vi; // bodies

  repeat
    // evaluate cont-expr
    result := ALEval(ce);
    if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; goto forcond; end; // cont
    if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
    if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
    if (result = al_Nil) or (result = al_sfalse) then exit; // loop quit
    // evaluate bodies
    nn := n;
    while nn <> al_Nil do
    begin
      result := ALEval(al_CAR[nn]); nn := al_CDR[nn].vi;
      if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; goto forcond; end; // cont
      if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
      if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
    end;
    // evaluate inc-expression
forcond:
    result := ALEval(ie);
    if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; goto forcond; end; // cont
    if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
    if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
  until false;
end;

function ALPWhile (n: Integer): Integer;
// (while cont-expr [body-expr]*)
label again;
var
  ce, nn: Integer;
begin
  ce := al_CAR[n]; n := al_CDR[n].vi; // bodies
  repeat
again:
    // evaluate cont-expr
    result := ALEval(ce);
    if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; continue; end; // cont
    if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
    if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
    if (result = al_Nil) or (result = al_sfalse) then exit; // loop quit
    // evaluate bodies
    nn := n;
    while nn <> al_Nil do
    begin
      result := ALEval(al_CAR[nn]); nn := al_CDR[nn].vi;
      if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; goto again; end; // cont
      if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
      if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
    end;
  until false;
end;

function ALPUntil (n: Integer): Integer;
// (until quit-expr [body-expr]*)
label untilcond;
var
  qe, nn: Integer;
begin
  qe := al_CAR[n]; n := al_CDR[n].vi; // bodies
  repeat
    // evaluate bodies
    nn := n;
    while nn <> al_Nil do
    begin
      result := ALEval(al_CAR[nn]); nn := al_CDR[nn].vi;
      if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; goto untilcond; end; // cont
      if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
      if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
    end;

    // evaluate quit-expr
untilcond:
    result := ALEval(qe);
    if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; goto untilcond; end; // cont
    if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
    if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
    if (result <> al_Nil) and (result <> al_sfalse) then exit; // loop quit
  until false;
end;

function ALPLoop (n: Integer): Integer;
// (loop [body-expr]+)
label again;
var
  nn: Integer;
begin
  repeat
again:
    // evaluate bodies
    nn := n;
    while nn <> al_Nil do
    begin
      result := ALEval(al_CAR[nn]); nn := al_CDR[nn].vi;
      if al_BreakFlag = ALBREAK_CONT then begin al_BreakFlag := ALBREAK_NONE; goto again; end; // cont
      if al_BreakFlag = ALBREAK_BREAK then begin al_BreakFlag := ALBREAK_NONE; exit; end; // break
      if al_BreakFlag <> ALBREAK_NONE then exit; // error or return or ...
    end;
  until false;
end;

function ALPLOr (n: Integer): Integer;
// (lor expr [expr]*)
// early quit
begin
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]); if al_BreakFlag <> ALBREAK_NONE then exit;
    if (result <> al_Nil) and (result <> al_sfalse) then begin result := al_strue; exit; end;
    n := al_CDR[n].vi;
  end;
  result := al_sfalse;
end;

function ALPLAnd (n: Integer): Integer;
// (land expr [expr]*)
// early quit
begin
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]); if al_BreakFlag <> ALBREAK_NONE then exit;
    if (result = al_Nil) or (result = al_sfalse) then begin result := al_sfalse; exit; end;
    n := al_CDR[n].vi;
  end;
  result := al_strue;
end;

function ALPNot (n: Integer): Integer;
// (not expr)
begin
  n := al_CAR[n];
  if (n = al_Nil) or (n = al_sfalse) then result := al_strue else result := al_sfalse;
end;

function ALPIsNull (n: Integer): Integer;
// (null? expr)
begin
  n := al_CAR[n];
  if n = al_Nil then result := al_strue else result := al_sfalse;
end;

function ALPIsNotNull (n: Integer): Integer;
// (not-null? expr)
begin
  n := al_CAR[n];
  if n <> al_Nil then result := al_strue else result := al_sfalse;
end;

function ALPFindSymbol (n: Integer): Integer;
// (find-symbol expr)
// returns *NIL* or symbol pair
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> DT_STRING then result := ALThrowError('FIND-SYMBOL: invalid symbol name', n)
  else
  begin
    result := ALFindSymbolEx(al_StringPool[al_CDR[n].vi]);
  end;
end;

function ALPIsAtom (n: Integer): Integer;
// (atom? expr)
begin
  n := al_CAR[n];
  if (n = al_Nil) or ALIsAtom(n) then result := al_strue else result := al_sfalse;
end;

function ALPIsNotAtom (n: Integer): Integer;
// (not-atom? expr)
begin
  n := al_CAR[n];
  if (n = al_Nil) or ALIsAtom(n) then result := al_sfalse else result := al_strue;
end;

function ALPIsNumber (n: Integer): Integer;
// (number? expr)
begin
  n := ALGetAtomType(al_CAR[n]);
  if (n = DT_INTEGER) or (n = DT_FLOAT) then result := al_strue else result := al_sfalse;
end;

function ALPIsInt (n: Integer): Integer;
// (int? expr)
begin
  if ALGetAtomType(al_CAR[n]) = DT_INTEGER then result := al_strue else result := al_sfalse;
end;

function ALPIsFloat (n: Integer): Integer;
// (float? expr)
begin
  if ALGetAtomType(al_CAR[n]) = DT_FLOAT then result := al_strue else result := al_sfalse;
end;

function ALPIsString (n: Integer): Integer;
// (string? expr)
begin
  if ALGetAtomType(al_CAR[n]) = DT_STRING then result := al_strue else result := al_sfalse;
end;

function ALPIsEmptyStr (n: Integer): Integer;
// (empty-str? expr)
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> DT_STRING then result := ALThrowError('EMPTY-STR?: string expected', n)
  else
  begin
    if al_CDR[n].vi = 0 then result := al_strue else result := al_sfalse;
  end;
end;

function ALPIsNotEmptyStr (n: Integer): Integer;
// (not-empty-str? expr)
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> DT_STRING then result := ALThrowError('NOT-EMPTY-STR?: string expected', n)
  else
  begin
    if al_CDR[n].vi <> 0 then result := al_strue else result := al_sfalse;
  end;
end;

function ALPIsVector (n: Integer): Integer;
// (vector? expr)
begin
  if ALGetAtomType(al_CAR[n]) = DT_VECTOR then result := al_strue else result := al_sfalse;
end;

function ALPIsByteVector (n: Integer): Integer;
// (byte-vector? expr)
begin
  if ALGetAtomType(al_CAR[n]) = DT_BYTEVECTOR then result := al_strue else result := al_sfalse;
end;

function ALPIsObject (n: Integer): Integer;
// (object? expr)
begin
  if ALGetAtomType(al_CAR[n]) = DT_TOBJECT then result := al_strue else result := al_sfalse;
end;

function ALPTypeQ (n: Integer): Integer;
// (type? expr)
begin
  result := ALNewIntAtom(ALGetAtomType(al_CAR[n]));
end;

function ALPConcat (n: Integer): Integer;
// (|| str [str]*)
var
  s: string;
begin
  s := '';
  while n <> al_Nil do
  begin
    if ALGetAtomType(al_CAR[n]) <> DT_STRING then
    begin
      result := ALThrowError('CONCAT: string expected', al_CAR[n]);
      exit;
    end;
    s := s+al_StringPool[al_CDR[al_CAR[n]].vi]; n := al_CDR[n].vi;
  end;
  result := ALNewStrAtom(s);
end;

function ALPNewVector (n: Integer): Integer;
// (new-vector size)
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> DT_INTEGER then result := ALThrowError('NEW-VECTOR: integer expected', n)
  else result := ALNewVector(al_CDR[n].vi);
end;

function ALPNewByteVector (n: Integer): Integer;
// (new-bytevector size)
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> DT_INTEGER then result := ALThrowError('NEW-BYTEVECTOR: integer expected', n)
  else result := ALNewByteVector(al_CDR[n].vi);
end;

function ALPVectorFromList (n: Integer): Integer;
// (vector-from-list list)
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> -1 then result := ALThrowError('VECTOR-FROM-LIST: list expected', n)
  else result := ALNewVectorFromList(n);
end;

function ALPLength (n: Integer): Integer;
// (length list|vector|string)
begin
  n := al_CAR[n];
  case ALGetAtomType(n) of
    DT_STRING: result := ALNewIntAtom(Length(al_StringPool[al_CDR[n].vi]));
    DT_VECTOR: result := ALNewIntAtom(TList(al_CDR[n].vi).Count);
    DT_BYTEVECTOR: result := ALNewIntAtom(Length(TLispByteVector(al_CDR[n].vi).vector));
    -1:
      begin
        result := 0;
        while n <> al_Nil do
        begin
          if ALGetAtomType(n) <> -1 then
          begin
            result := ALThrowError('LENGTH: invalid list', n);
            exit;
          end;
          n := al_CDR[n].vi; Inc(result);
        end;
        result := ALNewIntAtom(result);
      end;
    else result := ALThrowError('LENGTH: list, vector or string expected', n);
  end;
end;

function ALPSetLength (n: Integer): Integer;
// (set-length vector|string newlength)
var
  nl: Integer;
  s: string;
begin
  nl := al_CAR[al_CDR[n].vi];
  if ALGetAtomType(nl) <> DT_INTEGER then
  begin
    result := ALThrowError('SET-LENGTH: integer length expected');
    exit;
  end;
  nl := al_CDR[nl].vi;
  if (nl < 0) or (nl > 1024*1024) then
  begin
    result := ALThrowError('SET-LENGTH: invalid length');
    exit;
  end;
  n := al_CAR[n];
  case ALGetAtomType(n) of
    DT_STRING:
      begin
        s := al_StringPool[al_CDR[n].vi];
        if Length(s) = nl then result := n
        else
        begin
          n := Length(s);
          SetLength(s, nl);
          if n < nl then
          begin
            UniqueString(s);
            FillChar(s[n+1], nl-n, ' ');
          end;
          result := ALNewStrAtom(s);
        end;
      end;
    DT_VECTOR, DT_BYTEVECTOR:
      begin
        ALSetVectorLength(n, nl);
        result := n;
      end;
    else result := ALThrowError('SET-LENGTH: vector or string expected', n);
  end;
end;

function ALPNth (n: Integer): Integer;
// (nth list|vector|string index)
// index < 0: from the last item (-1: the latest one)
// for strings: first char index is 1, returns charcode or -1
// for lists: returns new list with the <index> item as head
var
  nn, ind: Integer;
  s: string;
begin
  ind := al_CAR[al_CDR[n].vi];
  if ALGetAtomType(ind) <> DT_INTEGER then
  begin
    result := ALThrowError('NTH: integer expected', ind);
    exit;
  end;
  ind := al_CDR[ind].vi; n := al_CAR[n];
  case ALGetAtomType(n) of
    DT_STRING:
      begin
        s := al_StringPool[al_CDR[n].vi];
        if ind < 0 then ind := Length(s)+ind+1;
        if (ind < 1) or (ind > Length(s)) then result := ALNewIntAtom(-1)
        else result := ALNewIntAtom(ord(s[ind]));
      end;
    DT_VECTOR:
      begin
        nn := TList(al_CDR[n].vi).Count;
        if ind < 0 then ind := nn+ind;
        if (ind < 0) or (ind >= nn) then result := al_Nil
        else result := Integer(TList(al_CDR[n].vi)[ind]);
      end;
    DT_BYTEVECTOR:
      begin
        nn := Length(TLispByteVector(al_CDR[n].vi).vector);
        if ind < 0 then ind := nn+ind;
        if (ind < 0) or (ind >= nn) then result := ALNewIntAtom(0)
        else result := ALNewIntAtom(TLispByteVector(al_CDR[n].vi).vector[ind]);
      end;
    -1:
      begin
        if ind < 0 then
        begin
          result := 0; nn := n;
          while nn <> al_Nil do
          begin
            if ALGetAtomType(nn) <> -1 then
            begin
              result := ALThrowError('NTH: invalid list', n);
              exit;
            end;
            nn := al_CDR[nn].vi; Inc(result);
          end;
          ind := result+ind;
          if ind < 0 then begin result := al_Nil; exit; end;
        end;
        result := n;
        while (result <> al_Nil) and (ind > 0) do
        begin
          if ALGetAtomType(n) <> -1 then
          begin
            result := ALThrowError('NTH: invalid list', n);
            exit;
          end;
          result := al_CDR[result].vi; Dec(ind);
        end;
      end;
    else begin result := ALThrowError('NTH: list, vector or string expected', n); exit; end;
  end;
end;

function ALPNthChar (n: Integer): Integer;
// (nth-char string index)
// index < 0: from the last item (-1: the latest one)
// first char index is 1, returns string with one char or empty string
var
  ind: Integer;
  s: string;
begin
  ind := al_CAR[n];
  if ALGetAtomType(ind) <> DT_STRING then
  begin
    result := ALThrowError('NTH-CHAR: string expected', ind);
    exit;
  end;
  s := al_StringPool[al_CDR[ind].vi];
  ind := al_CAR[al_CDR[n].vi];
  if ALGetAtomType(ind) <> DT_INTEGER then
  begin
    result := ALThrowError('NTH-CHAR: integer expected', ind);
    exit;
  end;
  ind := al_CDR[ind].vi;
  if ind < 0 then ind := Length(s)+ind+1;
  if (ind < 1) or (ind > Length(s)) then result := ALNewStrAtom('')
  else result := ALNewStrAtom(s[ind]);
end;

function ALPSetNth (n: Integer): Integer;
// (set-nth vector|string index newvalue)
// index < 0: from the last item (-1: the latest one)
// for strings: first char index is 1, returns charcode
var
  nn, ind, val: Integer;
  s: string;
begin
  ind := al_CAR[al_CDR[n].vi]; // index
  if ALGetAtomType(ind) <> DT_INTEGER then
  begin
    result := ALThrowError('SET-NTH: integer index expected', ind);
    exit;
  end;
  ind := al_CDR[ind].vi;
  val := al_CAR[al_CDR[al_CDR[n].vi].vi]; // value
  n := al_CAR[n]; // obj
  case ALGetAtomType(n) of
    DT_STRING:
      begin
        case ALGetAtomType(val) of
          DT_INTEGER: val := al_CDR[val].vi;
          DT_STRING:
            begin
              s := al_StringPool[al_CDR[val].vi];
              if s = '' then val := 0 else val := ord(s[1]);
            end;
          else
            begin
              result := ALThrowError('SET-NTH: char or integer value expected', val);
              exit;
            end;
        end;
        s := al_StringPool[al_CDR[n].vi];
        if ind < 0 then ind := Length(s)+ind+1;
        if (ind < 1) or (ind > Length(s)) then result := n
        else
        begin
          s[ind] := chr(val and $FF);
          result := ALNewStrAtom(s);
        end;
      end;
    DT_VECTOR:
      begin
        nn := TList(al_CDR[n].vi).Count;
        if ind < 0 then ind := nn+ind;
        if (ind >= 0) and (ind < nn) then TList(al_CDR[n].vi)[ind] := Pointer(val);
        result := n;
      end;
    DT_BYTEVECTOR:
      begin
        if ALGetAtomType(val) <> DT_INTEGER then
        begin
          result := ALThrowError('SET-NTH: integer value expected', val);
          exit;
        end;
        val := al_CDR[val].vi;
        nn := Length(TLispByteVector(al_CDR[n].vi).vector);
        if ind < 0 then ind := nn+ind;
        if (ind >= 0) and (ind < nn) then TLispByteVector(al_CDR[n].vi).vector[ind] := val;
        result := n;
      end;
    else result := ALThrowError('SET-NTH: vector or string expected', n);
  end;
end;

function ALPStrMid (n: Integer): Integer;
// (str-mid string from [length])
// first char index is 1
// index < 0: from the last char (-1: the latest one)
// if no length, then length is Length(string)
var
  from, len: Integer;
  s: string;
begin
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('STR-MID: string expected', al_CAR[n]);
    exit;
  end;
  s := al_StringPool[al_CDR[al_CAR[n]].vi];

  n := al_CDR[n].vi; from := al_CAR[n];
  if ALGetAtomType(from) <> DT_INTEGER then
  begin
    result := ALThrowError('STR-MID: integer index expected', from);
    exit;
  end;
  from := al_CDR[from].vi;

  n := al_CDR[n].vi;
  if n = al_Nil then len := Length(s)
  else
  begin
    len := al_CAR[n];
    if ALGetAtomType(len) <> DT_INTEGER then
    begin
      result := ALThrowError('STR-MID: integer length expected', len);
      exit;
    end;
    len := al_CDR[len].vi;
  end;

  if from < 0 then from := Length(s)+from+1;
  if (from < 1) or (from > Length(s)) then s := ''
  else s := Copy(s, from, len);
  result := ALNewStrAtom(s);
end;

function ALPStrDel (n: Integer): Integer;
// (str-del string from [length])
// first char index is 1
// index < 0: from the last char (-1: the latest one)
// if no length, then length is Length(string)
var
  from, len: Integer;
  s: string;
begin
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('STR-DEL: string expected', al_CAR[n]);
    exit;
  end;
  s := al_StringPool[al_CDR[al_CAR[n]].vi];

  n := al_CDR[n].vi; from := al_CAR[n];
  if ALGetAtomType(from) <> DT_INTEGER then
  begin
    result := ALThrowError('STR-DEL: integer index expected', from);
    exit;
  end;
  from := al_CDR[from].vi;

  n := al_CDR[n].vi;
  if n = al_Nil then len := Length(s)
  else
  begin
    len := al_CAR[n];
    if ALGetAtomType(len) <> DT_INTEGER then
    begin
      result := ALThrowError('STR-DEL: integer length expected', len);
      exit;
    end;
    len := al_CDR[len].vi;
  end;

  if from < 0 then from := Length(s)+from+1;
  Delete(s, from, len);
  result := ALNewStrAtom(s);
end;

function ALPStrIns (n: Integer): Integer;
// (str-ins what where from)
// first char index is 1, returns charcode
// index < 0: from the last char (-1: the latest one)
// if no length, then length is Length(string)
var
  from: Integer;
  what, where: string;
begin
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('STR-INS: string expected', al_CAR[n]);
    exit;
  end;
  what := al_StringPool[al_CDR[al_CAR[n]].vi];

  n := al_CDR[n].vi;
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('STR-INS: string expected', al_CAR[n]);
    exit;
  end;
  where := al_StringPool[al_CDR[al_CAR[n]].vi];

  n := al_CDR[n].vi; from := al_CAR[n];
  if ALGetAtomType(from) <> DT_INTEGER then
  begin
    result := ALThrowError('STR-INS: integer index expected', from);
    exit;
  end;
  from := al_CDR[from].vi;

  if from < 0 then from := Length(where)+from+1;
  Insert(what, where, from);
  result := ALNewStrAtom(where);
end;

function ALPStrPos (n: Integer): Integer;
// (str-pos what where [from])
// first char index is 1, returns pos or 0
// from < 0: from the last char (-1: the latest one)
var
  from: Integer;
  what, where: string;
begin
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('STR-POS: string expected', al_CAR[n]);
    exit;
  end;
  what := al_StringPool[al_CDR[al_CAR[n]].vi];

  n := al_CDR[n].vi;
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('STR-POS: string expected', al_CAR[n]);
    exit;
  end;
  where := al_StringPool[al_CDR[al_CAR[n]].vi];

  n := al_CDR[n].vi;
  if n = al_Nil then from := 1
  else
  begin
    from := al_CAR[n];
    if ALGetAtomType(from) <> DT_INTEGER then
    begin
      result := ALThrowError('STR-POS: integer index expected', from);
      exit;
    end;
    from := al_CDR[from].vi;
  end;

  if from < 0 then from := Length(where)+from+1;
  if (from < 1) or (from > Length(where)) then from := 0
  else from := PosEx(what, where, from);
  result := ALNewIntAtom(from);
end;

function ALPStrDup (n: Integer): Integer;
// (str-dup str cnt)
var
  cnt, p: Integer;
  s, t: string;
begin
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('STR-DUP: string expected', al_CAR[n]);
    exit;
  end;
  s := al_StringPool[al_CDR[al_CAR[n]].vi];

  n := al_CAR[al_CDR[n].vi];
  if ALGetAtomType(n) <> DT_INTEGER then
  begin
    result := ALThrowError('STR-DUP: integer index expected', n);
    exit;
  end;
  cnt := al_CDR[n].vi;

  if cnt < 1 then cnt := 0;
  //t := ''; while cnt > 0 do begin t := t+s; Dec(cnt); end;
  SetLength(t, Length(s)*cnt);
  if t <> '' then
  begin
    UniqueString(t);
    p := 1;
    while cnt > 0 do
    begin
      Move(s[1], t[p], Length(s)); Inc(p, Length(s)); Dec(cnt);
    end;
  end;
  result := ALNewStrAtom(t);
end;

function ALPOrd (n: Integer): Integer;
// (ord string)
// code of the 1st char (or -1 for empty string)
var
  s: string;
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> DT_STRING then
  begin
    result := ALThrowError('ORD: string expected', n);
    exit;
  end;
  s := al_StringPool[al_CDR[n].vi];
  if s = '' then result := -1 else result := ord(s[1]);
  result := ALNewIntAtom(result);
end;

function ALPChr (n: Integer): Integer;
// (chr code)
begin
  if ALGetAtomType(al_CAR[n]) <> DT_INTEGER then
  begin
    result := ALThrowError('CHR: string expected', al_CAR[n]);
    exit;
  end;
  result := ALNewStrAtom(chr(al_CDR[al_CAR[n]].vi and $FF));
end;

function ALPEvalStr (n: Integer): Integer;
// (eval-str string [ctx])
var
  ctxc: Integer;
begin
  //WriteLn('EVAL-STR');
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('EVAL-STR: string expected', al_CAR[n]);
    exit;
  end;

  if al_CDR[n].vi = al_Nil then
  begin
    // one arg
    result := ALEvalStr(al_StringPool[al_CDR[al_CAR[n]].vi]);
    exit;
  end;

  // eval in new context...
  ctxc := al_CAR[al_CDR[n].vi]; // 2nd arg (context)
  // context?
  if not ALIsContext(ctxc) then
  begin
    result := ALThrowError('EVAL-STR: context expected', ctxc);
    exit;
  end;
  result := al_CDR[al_CAR[n]].vi; // 1st arg - si
  if not ALEnterContext(ctxc) then
  begin
    result := ALThrowError('EVAL-STR: context error', ctxc);
    exit;
  end;
  result := ALEvalStr(al_StringPool[result]);
  if al_BreakFlag = ALBREAK_ERROR then exit;
  ALLeaveContexts(1);
end;

function ALPUpEvalStr (n: Integer): Integer;
// (up-eval-str string [ctx])
var
  ctxc: Integer;
begin
  if ALGetAtomType(al_CAR[n]) <> DT_STRING then
  begin
    result := ALThrowError('UP-EVAL-STR: string expected', al_CAR[n]);
    exit;
  end;

  if al_CDR[n].vi = al_Nil then
  begin
    // one arg
    ALPushCell(al_Frames); al_Frames := al_CDR[al_Frames].vi;
    result := ALEvalStr(al_StringPool[al_CDR[al_CAR[n]].vi]);
  end
  else
  begin
    // eval in new context...
    ctxc := al_CAR[al_CDR[n].vi]; // 2nd arg (context)
    // context?
    if not ALIsContext(ctxc) then
    begin
      result := ALThrowError('UP-EVAL-STR: context expected', ctxc);
      exit;
    end;
    result := al_CDR[al_CAR[n]].vi; // 1st arg - si
    if not ALEnterContext(ctxc) then
    begin
      result := ALThrowError('UP-EVAL-STR: context error', ctxc);
      exit;
    end;
    ALPushCell(al_Frames); al_Frames := al_CDR[al_Frames].vi;
    result := ALEvalStr(al_StringPool[result]);
    if al_BreakFlag <> ALBREAK_ERROR then ALLeaveContexts(1);
  end;

  if al_BreakFlag = ALBREAK_ERROR then exit;
  al_Frames := ALPopCells(1);
end;

function ALPUpEval (n: Integer): Integer;
// (up-eval expr [ctx])
var
  ctxc: Integer;
begin
  if al_CDR[n].vi = al_Nil then
  begin
    // one arg
    ALPushCell(al_Frames); al_Frames := al_CDR[al_Frames].vi;
    result := ALEval(al_CAR[n]);
  end
  else
  begin
    // eval in new context...
    ctxc := al_CAR[al_CDR[n].vi]; // 2nd arg (context)
    // context?
    if not ALIsContext(ctxc) then
    begin
      result := ALThrowError('UP-EVAL: context expected', ctxc);
      exit;
    end;
    result := al_CAR[n]; // 1st arg
    if not ALEnterContext(ctxc) then
    begin
      result := ALThrowError('UP-EVAL: context error', ctxc);
      exit;
    end;
    ALPushCell(al_Frames); al_Frames := al_CDR[al_Frames].vi;
    result := ALEval(result);
    if al_BreakFlag <> ALBREAK_ERROR then ALLeaveContexts(1);
  end;

  if al_BreakFlag = ALBREAK_ERROR then exit;
  al_Frames := ALPopCells(1);
end;

function ALPAppend (n: Integer): Integer;
// (append list [list]*)
var
  nn: Integer;
begin
  result := al_Nil;
  while n <> al_Nil do
  begin
    nn := al_CAR[n]; n := al_CDR[n].vi;
    if ALGetAtomType(nn) <> -1 then
    begin
      result := ALThrowError('APPEND: list expected', n);
      exit;
    end;
    while nn <> al_Nil do
    begin
      if ALGetAtomType(nn) <> -1 then
      begin
        result := ALThrowError('APPEND: invalid list item', nn);
        exit;
      end;
      if result = al_Nil then
      begin
        result := ALCons(al_CAR[nn], al_Nil);
        ALPushCell(result);
      end
      else
      begin
        al_CDR[result].vi := ALCons(al_CAR[nn], al_Nil);
        result := al_CDR[result].vi;
      end;
      nn := al_CDR[nn].vi;
    end;
  end;
  if result <> al_Nil then result := ALPopCells(1);
end;

function ALPReverse (n: Integer): Integer;
// (reverse list)
begin
  n := al_CAR[n];
  if ALGetAtomType(n) <> -1 then
  begin
    result := ALThrowError('REVERSE: list expected', n);
    exit;
  end;
  result := al_Nil;
  while n <> al_Nil do
  begin
    if ALGetAtomType(n) <> -1 then
    begin
      result := ALThrowError('REVERSE: invalid list item', n);
      exit;
    end;
    result := ALCons(al_CAR[n], result);
    n := al_CDR[n].vi;
  end;
end;

function ALPEq (n: Integer): Integer;
// (eq expr0 expr1)
// does not performs recursive list comparison
var
  nn: Integer;
begin
  result := al_sfalse; nn := al_CAR[al_CDR[n].vi]; n := al_CAR[n];
  case ALGetAtomType(n) of
    DT_INTEGER:
      begin
        case ALGetAtomType(nn) of
          DT_INTEGER: if al_CDR[n].vi <> al_CDR[nn].vi then exit;
          DT_FLOAT: if al_CDR[n].vi <> al_CDR[nn].vf then exit;
          else exit;
        end;
      end;
    DT_FLOAT:
      begin
        case ALGetAtomType(nn) of
          DT_INTEGER: if al_CDR[n].vf <> al_CDR[nn].vi then exit;
          DT_FLOAT: if al_CDR[n].vf <> al_CDR[nn].vf then exit;
          else exit;
        end;
      end;
    DT_STRING:
      begin
        if (ALGetAtomType(nn) <> DT_STRING) or (al_CDR[n].vi <> al_CDR[nn].vi) then exit;
      end;
    DT_VECTOR, DT_BYTEVECTOR, DT_TOBJECT:
      if (al_CAR[n] <> al_CAR[nn]) or (al_CDR[n].vi <> al_CDR[nn].vi) then exit;
    else if n <> nn then exit;
  end;
  result := al_strue;
end;

function ALPNEq (n: Integer): Integer;
// (neq expr0 expr1)
begin
  result := ALPEq(n);
  if (result = al_Nil) or (result = al_sfalse) then result := al_strue else result := al_sfalse;
end;

function ALPLowerCase (n: Integer): Integer;
// (lower-case str)
var
  s, t: string;
begin
  n := al_CAR[n]; if ALGetAtomType(n) = DT_STRING then
  begin
    s := al_StringPool[al_CDR[n].vi];
    t := AnsiLowerCase(s);
    if t <> s then result := ALNewStrAtom(t) else result := n;
  end
  else result := ALThrowError('LOWER-CASE: string expected');
end;

function ALPUpperCase (n: Integer): Integer;
// (upper-case str)
var
  s, t: string;
begin
  n := al_CAR[n]; if ALGetAtomType(n) = DT_STRING then
  begin
    s := al_StringPool[al_CDR[n].vi];
    t := AnsiUpperCase(s);
    if t <> s then result := ALNewStrAtom(t) else result := n;
  end
  else result := ALThrowError('UPPER-CASE: string expected');
end;

function ALPCompareText (n: Integer): Integer;
// (compare-text str0 str1)
begin
  if (ALGetAtomType(al_CAR[n]) <> DT_STRING) or (al_CAR[n] <> al_CAR[al_CDR[n].vi]) then
    result := ALThrowError('COMPARE-TEXT: invalid argument type')
  else
    result :=
      ALNewIntAtom(
        CompareStr(AnsiUpperCase(al_StringPool[al_CDR[al_CAR[n]].vi]),
          AnsiUpperCase(al_StringPool[al_CDR[al_CAR[al_CDR[n].vi]].vi])));
end;

function ALPCompareStr (n: Integer): Integer;
// (compare-str str0 str1)
begin
  if (ALGetAtomType(al_CAR[n]) <> DT_STRING) or (al_CAR[n] <> al_CAR[al_CDR[n].vi]) then
    result := ALThrowError('COMPARE-STR: invalid argument type')
  else
    result :=
      ALNewIntAtom(CompareStr(al_StringPool[al_CDR[al_CAR[n]].vi],
        al_StringPool[al_CDR[al_CAR[al_CDR[n].vi]].vi]));
end;

function ALPStrAnsi2Oem (n: Integer): Integer;
// (str-ansi-to-oem str)
var
  s, t: string;
begin
  n := al_CAR[n]; if ALGetAtomType(n) = DT_STRING then
  begin
    s := al_StringPool[al_CDR[n].vi];
    if s <> '' then
    begin
      t := s; UniqueString(t);
      CharToOemBuff(@(t[1]), @(t[1]), Length(t));
      if t <> s then result := ALNewStrAtom(t) else result := n;
    end
    else result := n;
  end
  else result := ALThrowError('STR-ANSI-TO-OEM: string expected');
end;

function ALPStrOem2Ansi (n: Integer): Integer;
// (str-oem-to-ansi str)
var
  s, t: string;
begin
  n := al_CAR[n]; if ALGetAtomType(n) = DT_STRING then
  begin
    s := al_StringPool[al_CDR[n].vi];
    if s <> '' then
    begin
      t := s; UniqueString(t);
      OemToCharBuff(@(t[1]), @(t[1]), Length(t));
      if t <> s then result := ALNewStrAtom(t) else result := n;
    end
    else result := n;
  end
  else result := ALThrowError('STR-OEM-TO-ANSI: string expected');
end;


function ALPSysLispOpt (n: Integer): Integer;
// (lisp-optimization [flag])
begin
  if al_Optimize then result := al_strue else result := al_sfalse;
  if n <> al_Nil then
  begin
    n := al_CAR[n];
    al_Optimize := (n <> al_Nil) and (n <> al_sfalse);
  end;
end;

function ALPSysThrowErr (n: Integer): Integer;
// (sys-throw-errors [flag])
begin
  if al_ThrowSysError then result := al_strue else result := al_sfalse;
  if n <> al_Nil then
  begin
    n := al_CAR[n];
    al_ThrowSysError := (n <> al_Nil) and (n <> al_sfalse);
  end;
end;


function ALPDumpFrames (n: Integer): Integer;
// (dump-frames)
begin
  result := al_Nil;
  ALDumpFrames();
end;


function ALPGC (n: Integer): Integer;
begin
  ALGC();
  result := al_Nil;
end;


function ALPNewContext (n: Integer): Integer;
// (new-context)
begin
  result := ALNewContextAtom();
end;

function ALPInContext (n: Integer): Integer;
// (in-context ctx expr [expr ...])
var
  ctx: Integer;
begin
  result := ALEval(al_CAR[n]);
  if al_BreakFlag <> ALBREAK_NONE then exit;
  ctx := result;
  if not ALIsContext(ctx) then
  begin
    result := ALThrowError('IN-CONTEXT: context expected', ctx);
    exit;
  end;
  n := al_CDR[n].vi; result := al_Nil;
  if not ALEnterContext(ctx) then
  begin
    result := ALThrowError('IN-CONTEXT: context error', ctx);
    exit;
  end;
  while n <> al_Nil do
  begin
    result := ALEval(al_CAR[n]);
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag <> ALBREAK_ERROR then ALLeaveContexts(1);
      exit;
    end;
    n := al_CDR[n].vi;
  end;
  ALLeaveContexts(1);
end;

function ALPContextDup (n: Integer): Integer;
// (context-dup [ctx])
// resulting context symbols are in reverse order
var
  ctxc: Integer;
  ctx: TLispContextObject;
begin
  ctxc := ALCar(n);
  if ALIsNil(n) or ALIsNil(ctxc) then
  begin
    // dup current context
    ctx := al_CurrentCtx;
  end
  else
  begin
    if not ALIsContext(ctxc) then
    begin
      result := ALThrowError('CONTEXT-DUP: context expected', ctxc);
      exit;
    end;
    ctx := TLispContextObject(al_CDR[ctxc].vi);
  end;

  n := ctx.fSymbols;

  result := ALNewContextAtom();
  ALPushCell(result);
  if not ALEnterContext(result) then
  begin
    ALPopCells(1);
    result := ALThrowError('CONTEXT-DUP: context error');
    exit;
  end;

  while n <> al_Nil do
  begin
    ctxc := al_CAR[n]; n := al_CDR[n].vi;
    ALSetGlobalSymbol(al_StringPool[al_CDR[al_CAR[ctxc]].vi], al_CDR[ctxc].vi, true);
    //WriteLn('"', al_StringPool[al_CDR[al_CAR[ctxc]]], '" duped');
  end;
  ALLeaveContexts(1);
  result := ALPopCells(1);
end;

function ALPContextAdd (n: Integer): Integer;
// (context-add dest [src])
// appended symbols are in reverse order
var
  ctxc, srccx: Integer;
  ctx, srcc: TLispContextObject;
begin
  srccx := ALCar(n); n := al_CDR[n].vi;
  if ALIsNil(srccx) then
  begin
    // current context
    srcc := al_CurrentCtx;
  end
  else
  begin
    if not ALIsContext(srccx) then
    begin
      result := ALThrowError('CONTEXT-ADD: context expected', srccx);
      exit;
    end;
    srcc := TLispContextObject(al_CDR[srccx].vi);
  end;

  ctxc := ALCar(n);
  if ALIsNil(n) or ALIsNil(ctxc) then
  begin
    // current context
    ctx := al_CurrentCtx;
  end
  else
  begin
    if not ALIsContext(ctxc) then
    begin
      result := ALThrowError('CONTEXT-ADD: context expected', ctxc);
      exit;
    end;
    ctx := TLispContextObject(al_CDR[ctxc].vi);
  end;
  result := ALTrue();

  if ctx = srcc then exit;

  n := ctx.fSymbols;

  if not ALEnterContext(srccx) then
  begin
    result := ALThrowError('CONTEXT-ADD: context error');
    exit;
  end;

  while n <> al_Nil do
  begin
    ctxc := al_CAR[n]; n := al_CDR[n].vi;
    ALSetGlobalSymbol(al_StringPool[al_CDR[al_CAR[ctxc]].vi], al_CDR[ctxc].vi, true);
  end;
  ALLeaveContexts(1);
end;

function ALPClearContextStack (n: Integer): Integer;
// (clear-context-stack)
begin
  al_ContextStack.Clear();
  al_CurrentCtx := al_MainContextObj;
  result := al_Nil;
end;


{$INCLUDE alisp3fnmath.pas}
{$INCLUDE alisp3fnvec.pas}


procedure ALDefineStdPrimitives ();
begin
  ALAddPrimitive('car', ALPCar, false, 1, 1);
  ALAddPrimitive('cdr', ALPCdr, false, 1, 1);
  ALAddPrimitive('setq', ALPSetQ, true, 2, -1);
  ALAddPrimitive('begin', ALPBegin, false, 0, -1);
  ALAddPrimitive('defun', ALPDefun, true, 3, -1);
  ALAddPrimitive('if', ALPIf, true, 2, -1);
  ALAddPrimitive('enum', ALPEnum, true, 1, -1);

  ALAddPrimitive('switch', ALPSwitch, true, 2, -1);
  ALAddPrimitive('for', ALPFor, true, 3, -1);
  ALAddPrimitive('while', ALPWhile, true, 1, -1);
  ALAddPrimitive('until', ALPUntil, true, 1, -1);
  ALAddPrimitive('loop', ALPLoop, true, 1, -1);

  ALAddPrimitive('macro', ALPMacro, true, 3, -1);
  ALAddPrimitive('set', ALPSet, false, 2, -1);
  ALAddPrimitive('undefine', ALPUndefine, false, 1, -1);
  ALAddPrimitive('lprint', ALPLPrint, true, 0, -1);
  ALAddPrimitive('lprintln', ALPLPrintLn, true, 0, -1);
  ALAddPrimitive('lprint-to-str', ALPLPrintToStr, true, 0, -1);
  ALAddPrimitive('print', ALPPrint, true, 0, -1);
  ALAddPrimitive('println', ALPPrintLn, true, 0, -1);
  ALAddPrimitive('catch', ALPCatch, true, 1, -1);

  ALAddPrimitive('not', ALPNot, false, 1, 1);
  ALAddPrimitive('null?', ALPIsNull, false, 1, 1);
  ALAddPrimitive('not-null?', ALPIsNotNull, false, 1, 1);
  ALAddPrimitive('atom?', ALPIsAtom, false, 1, 1);
  ALAddPrimitive('not-atom?', ALPIsNotAtom, false, 1, 1);
  ALAddPrimitive('number?', ALPIsNumber, false, 1, 1);
  ALAddPrimitive('int?', ALPIsInt, false, 1, 1);
  ALAddPrimitive('float?', ALPIsFloat, false, 1, 1);
  ALAddPrimitive('string?', ALPIsString, false, 1, 1);
  ALAddPrimitive('vector?', ALPIsVector, false, 1, 1);
  ALAddPrimitive('byte-vector?', ALPIsByteVector, false, 1, 1);
  ALAddPrimitive('object?', ALPIsObject, false, 1, 1);
  ALAddPrimitive('type?', ALPTypeQ, false, 1, 1);
  ALAddPrimitive('find-symbol', ALPFindSymbol, false, 1, 1);

  // math
  ALAddPrimitive('incr', ALPIncr, false, 1, 2);
  ALAddPrimitive('decr', ALPDecr, false, 1, 2);
  ALAddPrimitive('inc-v', ALPIncVar, true, 1, 2);
  ALAddPrimitive('dec-v', ALPDecVar, true, 1, 2);
  ALAddPrimitive('post-inc-v', ALPPostIncVar, true, 1, 2);
  ALAddPrimitive('post-dec-v', ALPPostDecVar, true, 1, 2);
  ALAddPrimitive('~', ALPBitNot, false, 1, 1);
  ALAddPrimitive('rnd', ALPRnd, false, 1, 1);
  ALAddPrimitive('randomize', ALPRandomize, false, 0, 1);
  ALAddPrimitive('abs', ALPAbs, false, 1, 1);
  ALAddPrimitive('neg', ALPNeg, false, 1, 1);
  ALAddPrimitive('sqr', ALPSqr, false, 1, 1);
  ALAddPrimitive('sqrt', ALPSqrt, false, 1, 1);
  ALAddPrimitive('sign', ALPSign, false, 1, 1);
  ALAddPrimitive('neg?', ALPIsNeg, false, 1, 1);
  ALAddPrimitive('pos?', ALPIsPos, false, 1, 1);
  ALAddPrimitive('zero?', ALPIsZero, false, 1, 1);
  ALAddPrimitive('not-zero?', ALPIsNotZero, false, 1, 1);
  ALAddPrimitive('empty-str?', ALPIsEmptyStr, false, 1, 1);
  ALAddPrimitive('not-empty-str?', ALPIsNotEmptyStr, false, 1, 1);

  ALAddPrimitive('toint', ALPToInt, false, 1, 1);
  ALAddPrimitive('tofloat', ALPToFloat, false, 1, 1);
  ALAddPrimitive('tonumber', ALPToNumber, false, 1, 1);
  ALAddPrimitive('tostr', ALPToStr, false, 1, 1);

  ALAddPrimitive('+', ALPPlus, false, 2, -1);
  ALAddPrimitive('-', ALPMinus, false, 2, -1);
  ALAddPrimitive('*', ALPMul, false, 2, -1);
  ALAddPrimitive('/', ALPDiv, false, 2, -1);
  ALAddPrimitive('idiv', ALPIDiv, false, 2, -1);
  ALAddPrimitive('mod', ALPMod, false, 2, -1);
  ALAddPrimitive('min', ALPMin, false, 2, -1);
  ALAddPrimitive('max', ALPMax, false, 2, -1);
  ALAddPrimitive('&', ALPBitAnd, false, 2, -1);
  ALAddPrimitive('|', ALPBitOr, false, 2, -1);
  ALAddPrimitive('^', ALPBitXor, false, 2, -1);
  ALAddPrimitive('||', ALPConcat, false, 1, -1);

  // logic, comparisons
  ALAddPrimitive('<', ALPLess, false, 2, -1);
  ALAddPrimitive('<=', ALPLessEqu, false, 2, -1);
  ALAddPrimitive('>', ALPGreat, false, 2, -1);
  ALAddPrimitive('>=', ALPGreatEqu, false, 2, -1);
  ALAddPrimitive('=', ALPEqu, false, 2, -1);
  ALAddPrimitive('<>', ALPNotEqu, false, 2, -1);
  ALAddPrimitive('lor', ALPLOr, true, 1, -1);
  ALAddPrimitive('land', ALPLAnd, true, 1, -1);
  ALAddPrimitive('s<', ALPSLess, false, 2, -1);
  ALAddPrimitive('s<=', ALPSLessEqu, false, 2, -1);
  ALAddPrimitive('s>', ALPSGreat, false, 2, -1);
  ALAddPrimitive('s>=', ALPSGreatEqu, false, 2, -1);
  ALAddPrimitive('s=', ALPSEqu, false, 2, -1);
  ALAddPrimitive('s<>', ALPSNotEqu, false, 2, -1);

  ALAddPrimitive('cons', ALPCons, false, 2, 2);
  ALAddPrimitive('list', ALPList, false, 1, -1);

  ALAddPrimitive('new-vector', ALPNewVector, false, 1, 1);
  ALAddPrimitive('new-bytevector', ALPNewByteVector, false, 1, 1);
  ALAddPrimitive('vector-from-list', ALPVectorFromList, false, 1, 1);
  ALAddPrimitive('length', ALPLength, false, 1, 1);
  ALAddPrimitive('set-length', ALPSetLength, false, 2, 2);
  ALAddPrimitive('nth', ALPNth, false, 2, 2);
  ALAddPrimitive('nth-char', ALPNthChar, false, 2, 2);
  ALAddPrimitive('set-nth', ALPSetNth, false, 3, 3);

  ALAddPrimitive('str-mid', ALPStrMid, false, 2, 3);
  ALAddPrimitive('str-del', ALPStrDel, false, 2, 3);
  ALAddPrimitive('str-ins', ALPStrIns, false, 3, 3);
  ALAddPrimitive('str-pos', ALPStrPos, false, 2, 3);
  ALAddPrimitive('str-dup', ALPStrDup, false, 2, 2);
  ALAddPrimitive('ord', ALPOrd, false, 1, 1);
  ALAddPrimitive('chr', ALPChr, false, 1, 1);

  ALAddPrimitive('str-ansi-to-oem', ALPStrAnsi2Oem, false, 1, 1);
  ALAddPrimitive('str-oem-to-ansi', ALPStrOem2Ansi, false, 1, 1);

  ALAddPrimitive('eval-str', ALPEvalStr, false, 1, 2);
  ALAddPrimitive('up-eval-str', ALPUpEvalStr, false, 1, 2);
  ALAddPrimitive('up-eval', ALPUpEval, false, 1, 2);
  ALAddPrimitive('up-setq', ALPUpSetQ, true, 2, -1);
  ALAddPrimitive('up-set', ALPUpSet, false, 2, -1);

  ALAddPrimitive('append', ALPAppend, false, 1, -1);
  ALAddPrimitive('reverse', ALPReverse, false, 1, 1);
  ALAddPrimitive('eq', ALPEq, false, 2, 2);
  ALAddPrimitive('neq', ALPNEq, false, 2, 2);

  ALAddPrimitive('lower-case', ALPLowerCase, false, 1, 1);
  ALAddPrimitive('upper-case', ALPUpperCase, false, 1, 1);
  ALAddPrimitive('compare-text', ALPCompareText, false, 2, 2);
  ALAddPrimitive('compare-str', ALPCompareStr, false, 2, 2);

  ALAddPrimitive('load', ALPLoad, false, 1, 1);
  ALAddPrimitive('cload', ALPCLoad, false, 1, 2);
  ALAddPrimitive('sys-verbose-load', ALPVerbLoad, false, 0, 1);

  ALAddPrimitive('sys-lisp-optimization', ALPSysLispOpt, false, 0, 1);
  ALAddPrimitive('sys-throw-errors', ALPSysThrowErr, false, 0, 1);
  ALAddPrimitive('dump-frames', ALPDumpFrames, false, 0, 0);
  ALAddPrimitive('gc', ALPGC, false, 0, 0);

  ALAddPrimitive('vec-delete', ALPVecDelete, false, 2, 3);
  ALAddPrimitive('vec-extract', ALPVecExtract, false, 2, 3);
  ALAddPrimitive('vec-copy', ALPVecCopy, false, 2, 3);
  ALAddPrimitive('vec-extract-nth', ALPVecExtractNth, false, 2, 2);
  ALAddPrimitive('vec-concat', ALPVecConcat, false, 2, 2);
  ALAddPrimitive('vec-insert-nth', ALPVecInsertNth, false, 3, 3);
  ALAddPrimitive('vec-insert', ALPVecInsert, false, 3, 3);

  ALAddPrimitive('new-context', ALPNewContext, false, 0, 0);
  ALAddPrimitive('in-context', ALPInContext, true, 2, -1);
  ALAddPrimitive('context-dup', ALPContextDup, false, 0, 1);
  ALAddPrimitive('context-add', ALPContextAdd, false, 1, 2);
  ALAddPrimitive('clear-context-stack', ALPClearContextStack, false, 0, 0);

  al_sdefun := al_CDR[ALFindGlobalSymbol('defun')].vi;
  al_smacro := al_CDR[ALFindGlobalSymbol('macro')].vi;
  al_ssetq := al_CDR[ALFindGlobalSymbol('setq')].vi;
end;
