// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
{$R-,I-}
unit dspLabelsU;

interface

uses
  SysUtils, Classes;


procedure ZClearLabels (sysLabels: Boolean = false);
procedure ZLoadLabels (const fName: string; sysLabels: Boolean = false);
procedure ZSaveLabels (const fName: string; sysLabels: Boolean = false);
procedure ZAddLabel (const name: string; addr: Word; size: Integer; sysLabel: Boolean = false);
procedure ZDeleteLabel (const name: string; sysLabel: Boolean = false);

function ZFindLabelByName (const name: string; out addr: Word; out size: Integer; out sys: Boolean): Boolean;
function ZFindCodeLabel (addr: Word; out found: Boolean): string;
function ZFindDataLabel (addr: Word; out found: Boolean): string;


implementation


type
  PLabelInfo = ^TLabelInfo;
  TLabelInfo = record
    sys: Boolean;
    addr: Word;
    dataSize: Integer; // 0: code label
    name: string[127];
  end;


var
  zLabMap: TList;
    // objects: PLabelInfo
    // sorted by address (to fasten search)

procedure ZClearLabels (sysLabels: Boolean = false);
var
  f: Integer;
begin
  f := zLabMap.Count-1;
  while f >= 0 do
  begin
    if PLabelInfo(zLabMap[f]).sys = sysLabels then
    begin
      Dispose(PLabelInfo(zLabMap[f]));
      zLabMap.Delete(f);
    end;
    Dec(f);
  end;
end;

(*
function FindLabel (addr: Word): Integer;
var
  l, h, i: Integer;
  a: Word;
begin
  result := -1; l := 0; h := zLabMap.Count-1;
  while l <= h do
  begin
    i := (l+h) shr 1;
    a := PLabelInfo(zLabMap[i]).addr;
    if addr < a then l := i+1
    else
    begin
      h := i-1;
      if addr = a then begin result := l; exit; end;
    end;
  end;
end;
*)

function FindLabel (addr: Word): Integer;
begin
  result := zLabMap.Count-1;
  while (result >= 0) and (PLabelInfo(zLabMap[result]).addr <> addr) do Dec(result);
end;

procedure ZAddLabel (const name: string; addr: Word; size: Integer; sysLabel: Boolean = false);
var
  ind: Integer;
  linf: PLabelInfo;
begin
  if name = '' then exit;
  if (size < 0) or (size > 65535) then size := 0;

  ind := FindLabel(addr);
  if ind = -1 then
  begin
    // TODO: write more efficient code! %-)
    ind := zLabMap.Count-1;
    while (ind >= 0) and (addr < PLabelInfo(zLabMap[ind]).addr) do Dec(ind);
    Inc(ind);
    New(linf);
    linf.name := name;
    linf.datasize := size;
    linf.sys := sysLabel;
    linf.addr := addr;
    try
      zLabMap.Insert(ind, linf);
    except
      Dispose(linf);
      raise;
    end;
    //WriteLn(' added');
  end
  else
  begin
    //Write(' ');
    WriteLn('duplicate labels at #', IntToHex(addr, 4),
      ': ', PLabelInfo(zLabMap[ind]).name, ', ', name);
  end;
end;

procedure ZLoadLabels (const fName: string; sysLabels: Boolean = false);
var
  fi: Text;
  s, t, name: string;
  addr, size: Integer;
  f: Integer;
begin
  IOResult();
  AssignFile(fi, fName);
  Reset(fi);
  if IOResult() <> 0 then exit;
  try
    while not EOF(fi) do
    begin
      ReadLn(fi, s); if IOResult() <> 0 then exit;
      // trim left spaces
      f := 1; while (f <= Length(s)) and (s[f] <= ' ') do Inc(f);
      Delete(s, 1, f-1);
      if (s = '') or (s[1] = ';') then continue;

      // get first word (address)
      f := 1; while (f <= Length(s)) and (s[f] > ' ') do Inc(f);
      if f > Length(s) then continue;

      // decode address
      t := Copy(s, 1, f-1);
      if t[1] = '#' then t[1] := '$';
      addr := StrToIntDef(t, -1);
      if (addr < 0) or (addr > 65535) then continue;

      // get label name
      while (f <= Length(s)) and (s[f] <= ' ') do Inc(f);
      Delete(s, 1, f-1);
      f := 1; while (f <= Length(s)) and (s[f] > ' ') do Inc(f);
      if f = 1 then continue;
      name := Copy(s, 1, f-1);

      // get data size (if any)
      while (f <= Length(s)) and (s[f] <= ' ') do Inc(f);
      Delete(s, 1, f-1); s := TrimRight(s);
      if s <> '' then
      begin
        if s[1] = '#' then s[1] := '$';
        size := StrToIntDef(s, -1);
        if size < 0 then size := 0
        else if size > 65535 then size := 65535;
      end
      else size := 0;

      ZAddLabel(name, addr, size, sysLabels);
    end;
  finally
    CloseFile(fi);
    IOResult();
  end;
end;

procedure ZSaveLabels (const fName: string; sysLabels: Boolean = false);
var
  fo: Text;
  f: Integer;
  linf: PLabelInfo;
begin
  IOResult();
  AssignFile(fo, fName);
  Rewrite(fo);
  if IOResult() <> 0 then begin WriteLn('can''t save labels'); exit; end;
  for f := 0 to zLabMap.Count-1 do
  begin
    linf := PLabelInfo(zLabMap[f]);
    if linf.sys <> syslabels then continue;
    Write(fo, '#', IntToHex(linf.addr, 4), ' ', linf.name);
    if linf.dataSize > 0 then Write(fo, ' ', linf.dataSize);
    WriteLn(fo);
  end;
  CloseFile(fo);
end;

function ZFindCodeLabel (addr: Word; out found: Boolean): string;
var
  ind: Integer;
begin
  ind := FindLabel(addr);
  if (ind <> -1) {and (PLabelInfo(zLabMap[ind]).datasize = 0)} then
  begin
    found := true; result := PLabelInfo(zLabMap[ind]).name;
  end
  else
  begin
    found := false; result := '';
  end;
end;

function ZFindDataLabel (addr: Word; out found: Boolean): string;
var
  f, a, sz: Integer;
begin
  result := ''; found := true;
  f := zLabMap.Count-1;
  while f >= 0 do
  begin
    sz := PLabelInfo(zLabMap[f]).dataSize;
    if sz > 0 then
    begin
      a := PLabelInfo(zLabMap[f]).addr;
      if (addr >= a) and (addr < a+sz) then
      begin
        result := PLabelInfo(zLabMap[f]).name;
        if a <> addr then result := result+'+'+IntToStr(addr-a);
        exit;
      end;
    end;
    Dec(f);
  end;

  found := false;
end;

function ZFindLabelByName (const name: string; out addr: Word; out size: Integer; out sys: Boolean): Boolean;
var
  f: Integer;
begin
  f := zLabMap.Count-1;
  while (f >= 0) and (ANSICompareText(PLabelInfo(zLabMap[f]).name, name) <> 0) do Dec(f);
  if f >= 0 then
  begin
    result := true;
    addr := PLabelInfo(zLabMap[f]).addr;
    size := PLabelInfo(zLabMap[f]).dataSize;
    sys := PLabelInfo(zLabMap[f]).sys;
  end
  else
  begin
    result := true;
    addr := 0;
    size := 0;
    sys := false;
  end;
end;

procedure ZDeleteLabel (const name: string; sysLabel: Boolean = false);
var
  f: Integer;
begin
  f := zLabMap.Count-1;
  while (f >= 0) and (ANSICompareText(PLabelInfo(zLabMap[f]).name, name) <> 0) do Dec(f);
  if f >= 0 then
  begin
    Dispose(PLabelInfo(zLabMap[f]));
    zLabMap.Delete(f);
  end;
end;


initialization
  zLabMap := TList.Create();
finalization
  ZClearLabels(false);
  ZClearLabels(true);
  zLabMap.Free();
end.
