// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// variables
//
{$INCLUDE options.inc}
unit dspEmuVarsU;

interface

uses
  Windows, SysUtils, Classes,
  z80U, dspAYEngineU, dspTZXU, dspBetaU,
  dzsLibU, {$IFDEF USE_HP_TIMER}hpTimerU{$ELSE}qCounterU{$ENDIF}, dspMachConstU;


const
  emuName = 'DarkSpecter v0.2.1 RC1';
  emuSSCtrl = 1;
  emuSSAlt = 2;
  emuSSShift = 4;
  wconMaxHeight = 28;

  //initTapeBit = $FE;
  initTapeBit = $00;


type
  PByte = ^Byte;

  TEmuEvent = (eeBeforeExec, eeAfterExec, eeBeforePeek, eeAfterPeek,
    eeBeforePoke, eeAfterPoke, eeBeforeIn, eeAfterIn);

  TEmuConfig = record
    useKempstonMouse: Boolean;
    useKempstonJoystick: Boolean;
    usePCJoystick: Boolean;
    mouseSwap: Boolean;
    useKeystick: Boolean;
    keystickKempston: Boolean;
    emulateMatrixEffect: Boolean;
    contendedMemory: Boolean;
    pauseOnDefocus: Boolean;

    startingModel: TZXModel;
    autolockMouse: Boolean;

    default128k: TZXModel;

    //accurateScreenUpdate: Boolean;
      // update screen byte-by-byte instead of line-by-line (slow!)

    conDumpLastLines: Boolean;
      // show the last n console lines (if changed)? -- tan
    conDumpDelay: LongWord;
      // how long each line lasts (ms) -- 3000
    conLinesToDump: Integer;
      // # of lines to dump -- 2
    conHeight: Integer;
      // console height (24)
    conSlide: Boolean;
    conSlideDelay: Integer; // ms

    defaultBoot: string;
      // empty: built in
    autobootSCL: Boolean;
    autobootTRD: Boolean;

    useTRDOS: Boolean;
    tapeAutostart: Boolean;
    tapeAutomax: Boolean;   // max speed on tape loading
    tapeAutopause: Boolean; // pause after a tape was stopped
    tapePrepause: Boolean;  // pause before the tape stops
    tapeFastLoader: Boolean; // use "FastLoader(utm)"?

    emulatePortFF: Boolean;

    autofiring: Boolean;
    autofireRate: Integer;
    autofireKeepPressed: Integer;

    showClock: Integer;
      // 0: no; 1: fullscreen; 2: always

    tapeBit: Integer;
      // 0: 0; 1: 1; other -- take from tape emu

    pent512: Boolean; // emulate 512K on port #7FFD

    detectEdgeRoutine: Boolean;
    detectKnownLoaders: Boolean;

    rom48k: string;
    rom128k_0: string;
    rom128k_1: string;
    romtrdos: string;
  end;

  TPCKeyMapItem = record
    ports: array [0..1] of PByte;
    masks: array [0..1] of Byte;
  end;


var
  config: TEmuConfig;

  conScrollDelta: Integer = 0;

  emuJoyMask: Byte;
    // kempston joystick port value
  emuKeybMask: array [0..7] of Byte;
    // keyboard ports values
  emuKeysLocked: Boolean = false;
    // setting to TRUE locks ZX keyboard updating
  pcKeyMap: array [Boolean, 0..255] of TPCKeyMapItem;
    // mapping PC keys to ZX ones
    // false: normal; true: when SHIFT pressed
  pcKeys: array [0..255] of Boolean;
    // is corresponding PC key pressed?
  pcJoyMap: array [0..1, 0..7] of TPCKeyMapItem;
    // l,r,u,d,b1,b2,b3,b4
  pcJoys: array [0..1, 0..7] of Boolean;
    // is corresponding joystick action active?
  pcJoyCount: Integer = 0;
  pcJoyActive: Byte = $03; // bit 0: j1; bit 1: j2
  pcJoyXCen, pcJoyYCen: array [0..1] of Integer;
  pcJoyXMinMax, pcJoyYMinMax: array [0..1, 0..1] of Integer;

  // the following is for kempston mouse emulation
  // currently unused
  emuMousePos: TPoint; // scaled into byte!
  emuMouseX, emuMouseY: Integer; // for windowing system
  winMousePos: TPoint; // real
  emuMouseButtons: Byte; // 0: left; 1: right; 2: middle
  emuZXWActive: Boolean; // windowing system has control?

  emuZ80: PZ80CPU;
    // cpu
  emuAY: TAY8912;
    // ym/ay
  emuMemory: packed array [0..34, 0..16383] of Byte;
    // pages 0-7: RAM;
    // 8: 128k rom, part 1
    // 9: 48k rom/128k standard 48k rom
    // 10: tr-dos rom
    // 11-34: pentagon512 memory pages
  emuPaging: array [0..4] of Integer;
    // current stamped pages
  emuScreenPage: Integer;
    // current screen page (visible one)
  emuLastOut7FFD: Byte;
  emuAYRegIndex: Integer;
    // last out to 0FFFDh

  emuFramesToSkip: Integer = -1;
  emuFrameSkip: Integer = -1;
    //-1: auto
    // 0: never
    // >0: # of frames to skip

  // tape control
  emuTapeInfo:
    record
      tape: TZXRawPulses;
      playing: Boolean;
      nextEdgeT: Integer;
      tapeBit: Byte;
      ticksPassed: Integer; // after starting the tape
      ticksLeft: Integer; // paused, ticks unwasted %-)
    end;

  // disk interface objects

  emuDisks: TBetaDisk = nil;
  emuWasDiskActivity: Boolean = false;
    // was disk accesses during the last frame?

  emuScale: Integer = 1;
  emuWWidth, emuWHeight: Integer;
    // entire main window width & height (NOT CLIENT!)

  emuMaxSpeed: Boolean = false;
    // setting to TRUE runs emulator at maximum speed
  emuSpeedFactor: Double = 1.0;
    // speed factor (2 = double speed)

  emuIntrCount: Integer = 0;
    // increments on each speccy frame

  emuPaused: Boolean = false;
    // setting to TRUE pauses emulator
    // but DON'T do this manually!
  emuTempPaused: Boolean = false;
    // setting to TRUE pauses emulator
    // but DON'T do this manually!
  emuDoTapeAutopause: Boolean = false;
    // tape autopause?

  emuReanimated: Boolean = false;
    // TRUE -- program was reanimated from DI/HALT

  emuSpeakerState: Boolean = true;
    // speccy speaker bit

  emuSoundEnabled: Boolean = true;
    // FALSE -- sound totally disabled
  emuAYEnabled: Boolean = true;
    // FALSE -- ay disabled
  emuSyncWithSound: Boolean = false;
    // TRUE -- sync with sound, not with <emuHQCounter>
    // don't work for double or maximum speed modes
    // alas, this is not works at all...

  emuPrevMixAdvCounter: Integer = 0;
    // value of <mixBufAdvanceCounter> for the previous frame
  {$IFDEF USE_HP_TIMER}
  emuHQCounter: TX86LargeInteger;
  emuFPSHQCounter: TX86LargeInteger;
  emuFPSHQCounterCounting: Boolean = false;
  {$ELSE}
  emuHQCounter: THQCounter;
  emuFPSHQCounter: THQCounter;
  {$ENDIF}

  emuShowFPS: Boolean = false;

  emuInfoMsgHideTime: LongWord = 0;
  emuInfoMsgX: Integer;
  emuInfoMsg: string = '';

  // inner
  emuSetModel: procedure (model: TZXModel);

  emuKeyScreen: packed array [0..6911] of Byte;
  emuShowKeyHelp: Boolean = false;

  emuDbgActive: Boolean = false;
  emuSingleStep: Boolean = false; // activate debugger if TRUE

  emuConWinH: THandle;

  emuWConDumpScroll: LongWord = 0;
    // time of the next dump scrolling
  emuWConDumpLinesLeft: Integer = 0;
    // # of dump lines that are still on the screen

  emuWConVisible: Boolean = false;
  emuWConHeight: Integer = 0;
    // current console window height
  emuWConNextSlide: LongWord; // time to next slide %-)
  emuConInputLine: string = '';

  emuSavedBorderColor: Byte;

  emuShiftsState: Byte;
  emuFPSStIntrCount: Integer;
  emuFPS: Double = 0;
  emuPrevFPSSlows: Integer = 0;
  emuSlowIconHide: LongWord = 0;
  emuDiskIconHide: LongWord = 0;

  emuFrameComplete: Boolean = true;
    // true: EmuExec completes the frame

  emuInDebugger: Boolean = false;


  emuAltDebugKey: Word = 0;
  emuAltDebugShift: Byte = 0;

  emuAutofireDelay: Integer = 1;

  conKeyBinds: array [0..7, 0..255] of string;
    // for each VK in each of the shift states %-)

  emuUseDDraw: Boolean = false;
  emuNoTurtle: Boolean = false;

  (*emuEvents: array [TEmuEvent, 0..65535] of string;
  emuEventAddr: Integer;
  emuEventValue: Integer;
  emuEventAborted: Boolean;*)
  emuEventDontCall: Boolean = false;


type
  TEmuKeyMouseInfo = record
    mouse: Boolean;  // mouse event?
    // mouse event
    x, y: Integer;
    buttons: Byte;

    // keyboard event
    ascii: Char;
    vk: Word;
    shifts: Byte;
    pressed: Boolean;
  end;


var
  emuQuit: Boolean = false;  // true: WM_QUIT catched by EmuProcessMessages()
  EmuProcessMessages: function (): Boolean = nil;
  EmuReadKeyMouse: function (out msg: TMsg): Boolean; // no waiting
  EmuProcessWinMessage: procedure (var msg: TMsg) = nil;
  EmuProcessConMessage: procedure (var msg: TMsg) = nil;
  EmuDrawCursor: procedure = nil;
    // returns true if there is event

  hackZXSetProcs: procedure = nil;


procedure WaitKeybRelease ();


implementation


procedure WaitKeybRelease ();
begin
  while
    (GetAsyncKeyState(VK_RETURN) and $8000 <> 0) or
    (GetAsyncKeyState(VK_ESCAPE) and $8000 <> 0) do EmuProcessMessages();
end;

procedure EmuClearEvents ();
(*var
  f: Integer;
  c: TEmuEvent;*)
begin
  (*for c := eeBeforeExec to eeAfterIn do
    for f := 0 to 65535 do
      emuEvents[c, f] := '';*)
end;


initialization
  Z80ClearTraps(); Z80ClearTrapsED();
  New(emuZ80);

  FillChar(pcKeys, SizeOf(pcKeys), 0);
  FillChar(pcJoys, SizeOf(pcJoys), 0);
  emuMousePos.x := 0; emuMousePos.y := 0; emuMouseButtons := 0;

  emuTapeInfo.playing := false;

  emuDisks := TBetaDisk.Create();

  emuAY := TAY8912.Create();
  emuAY.InitClock(1773000); // speccy
  //emuAY.InitClock(1750000); // pentagon
  emuSetModel := nil;
  emuZXWActive := true;

  FillChar(emuTapeInfo, SizeOf(emuTapeInfo), 0);
  emuTapeInfo.tapeBit := initTapeBit;
  emuTapeInfo.tape := TZXRawPulses.Create();

  FillChar(config, SizeOf(config), 0);

  config.autofiring := false;
  config.autofireRate := 2;
  config.autofireKeepPressed := 1;

  config.conDumpLastLines := true;
    // show last two console lines (if changed)?
  config.conDumpDelay := 3000;
    // how long each line lasts (ms)
  config.conLinesToDump := 2;
    // # of lines to dump
  config.conHeight := 24;
    // console height (24)
  config.conSlide := true;
  config.conSlideDelay := 30;

  config.useKempstonMouse := true;
  config.useKempstonJoystick := true;
  config.usePCJoystick := true;
  config.mouseSwap := true;
  config.useKeystick := false;
  config.emulateMatrixEffect := true;
  config.contendedMemory := true;

  config.pauseOnDefocus := true;
   // pause DSP when it's main window lost focus
  config.useTRDOS := true;
  config.tapeFastLoader := true;
  config.tapeAutostart := true;
    // start tape when ROM LOAD_BYTES called
  config.tapeAutomax := true;
    // automatically switch to MaxSpeed when tape was autostarted
  config.tapeAutopause := false;
    // pause emu just after the tape stops
    // this is useful to take the snapshot of the loaded game
    // you'll miss some frames anyway, but not so many as
    // if you'll try to do this by hands %-)
  config.tapePrepause := false;
  config.startingModel := zxm128k;
  config.autolockMouse := true;
    // automatically lock the mouse when one of
    // mouse ports was read

  config.default128k := zxmPentagon128k;
  config.defaultBoot := ''; // empty: built in
  config.autobootSCL := true;
  config.autobootTRD := true;

  config.emulatePortFF := true;
  //config.accurateScreenUpdate := false;

  config.showClock := 1;
  config.pent512 := true;

  // default romz
  config.rom48k := 'romz\48.rom';
  config.rom128k_0 := 'romz\128_0_tr.rom';
  config.rom128k_1 := 'romz\128_1.rom';
  config.romtrdos := 'romz\trdos504f.rom';

  config.tapeBit := -1;

  config.detectEdgeRoutine := true;
  config.detectKnownLoaders := true;

  EmuClearEvents();

finalization
  //Dispose(emuZ80);
end.
