// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// PC & Speccy key names
// key names was taken from UnrealSpectrum by SMT and slightly modified
//
unit dspKeyNamesU;

interface

uses
  Windows, SysUtils, dspEmuVarsU;


const
  pcVKNames: array [0..83] of record name: string; code: Word; end = (
    (name:'esc'; code:VK_ESCAPE),
    (name:'escape'; code:VK_ESCAPE),
    (name:'f1'; code:VK_F1),
    (name:'f2'; code:VK_F2),
    (name:'f3'; code:VK_F3),
    (name:'f4'; code:VK_F4),
    (name:'f5'; code:VK_F5),
    (name:'f6'; code:VK_F6),
    (name:'f7'; code:VK_F7),
    (name:'f8'; code:VK_F8),
    (name:'f9'; code:VK_F9),
    (name:'f10'; code:VK_F10),
    (name:'f11'; code:VK_F11),
    (name:'f12'; code:VK_F12),
    (name:'prscr'; code:VK_SNAPSHOT),
    (name:'printscr'; code:VK_SNAPSHOT),
    (name:'printscreen'; code:VK_SNAPSHOT),
    (name:'scrlock'; code:VK_SCROLL),
    (name:'slock'; code:VK_SCROLL),
    (name:'sclock'; code:VK_SCROLL),
    (name:'scrollock'; code:VK_SCROLL),
    (name:'pause'; code:VK_PAUSE),
    (name:'minus'; code:$BD),
    (name:'plus'; code:$BB),
    (name:'back'; code:VK_BACK),
    (name:'tab'; code:VK_TAB),
    (name:'lb'; code:$DB),
    (name:'rb'; code:$DD),
    (name:'caps'; code:VK_CAPITAL),
    (name:'tilda'; code:$C0),
    (name:'space'; code:VK_SPACE),
    (name:'colon'; code:$BA),
    (name:'quote'; code:$DE),
    (name:'enter'; code:VK_RETURN),
    (name:'return'; code:VK_RETURN),
    (name:'comma'; code:$BC),
    (name:'point'; code:$BE),
    (name:'slash'; code:$BF),
    (name:'backslash'; code:$DC),
    (name:'backsl'; code:$DC),
    (name:'shift'; code:VK_SHIFT),
    (name:'alt'; code:VK_MENU),
    (name:'ctrl'; code:VK_CONTROL),
    (name:'control'; code:VK_CONTROL),
    (name:'leftshift'; code:VK_LSHIFT),
    (name:'lshift'; code:VK_LSHIFT),
    (name:'lalt'; code:VK_LMENU),
    (name:'leftalt'; code:VK_LMENU),
    (name:'leftctrl'; code:VK_LCONTROL),
    (name:'lctrl'; code:VK_LCONTROL),
    (name:'lcontrol'; code:VK_LCONTROL),
    (name:'rshift'; code:VK_RSHIFT),
    (name:'ralt'; code:VK_RMENU),
    (name:'rightctrl'; code:VK_RCONTROL),
    (name:'rightcontrol'; code:VK_RCONTROL),
    (name:'rcontrol'; code:VK_RCONTROL),
    (name:'ins'; code:VK_INSERT),
    (name:'home'; code:VK_HOME),
    (name:'pgup'; code:VK_PRIOR),
    (name:'del'; code:VK_DELETE),
    (name:'end'; code:VK_END),
    (name:'pgdn'; code:VK_NEXT),
    (name:'up'; code:VK_UP),
    (name:'down'; code:VK_DOWN),
    (name:'left'; code:VK_LEFT),
    (name:'right'; code:VK_RIGHT),
    (name:'numlock'; code:VK_NUMLOCK),
    (name:'grdiv'; code:VK_DIVIDE),
    (name:'grmul'; code:VK_MULTIPLY),
    (name:'grsub'; code:VK_SUBTRACT),
    (name:'gradd'; code:VK_ADD),
    (name:'clear'; code:VK_CLEAR),
    (name:'num5'; code:VK_CLEAR),
    (name:'n0'; code:VK_NUMPAD0),
    (name:'n1'; code:VK_NUMPAD1),
    (name:'n2'; code:VK_NUMPAD2),
    (name:'n3'; code:VK_NUMPAD3),
    (name:'n4'; code:VK_NUMPAD4),
    (name:'n5'; code:VK_NUMPAD5),
    (name:'n6'; code:VK_NUMPAD6),
    (name:'n7'; code:VK_NUMPAD7),
    (name:'n8'; code:VK_NUMPAD8),
    (name:'n9'; code:VK_NUMPAD9),
    (name:'np'; code:VK_DECIMAL));

const
  zxKNames: array [0..44] of record name: string; pvptr: PByte; mask: Byte; end = (
    // kempston
    (name:'kright'; pvptr:@emuJoyMask; mask: Byte(not 1)),
    (name:'kleft';  pvptr:@emuJoyMask; mask: Byte(not 2)),
    (name:'kdown';  pvptr:@emuJoyMask; mask: Byte(not 4)),
    (name:'kup';    pvptr:@emuJoyMask; mask: Byte(not 8)),
    (name:'kfire';  pvptr:@emuJoyMask; mask: Byte(not 16)),
    // keyboard
    (name:'cap'; pvptr:@emuKeybMask[0]; mask: Byte(not $01)),
    (name:'z';   pvptr:@emuKeybMask[0]; mask: Byte(not $02)),
    (name:'x';   pvptr:@emuKeybMask[0]; mask: Byte(not $04)),
    (name:'c';   pvptr:@emuKeybMask[0]; mask: Byte(not $08)),
    (name:'v';   pvptr:@emuKeybMask[0]; mask: Byte(not $10)),
    (name:'a';   pvptr:@emuKeybMask[1]; mask: Byte(not $01)),
    (name:'s';   pvptr:@emuKeybMask[1]; mask: Byte(not $02)),
    (name:'d';   pvptr:@emuKeybMask[1]; mask: Byte(not $04)),
    (name:'f';   pvptr:@emuKeybMask[1]; mask: Byte(not $08)),
    (name:'g';   pvptr:@emuKeybMask[1]; mask: Byte(not $10)),
    (name:'q';   pvptr:@emuKeybMask[2]; mask: Byte(not $01)),
    (name:'w';   pvptr:@emuKeybMask[2]; mask: Byte(not $02)),
    (name:'e';   pvptr:@emuKeybMask[2]; mask: Byte(not $04)),
    (name:'r';   pvptr:@emuKeybMask[2]; mask: Byte(not $08)),
    (name:'t';   pvptr:@emuKeybMask[2]; mask: Byte(not $10)),
    (name:'1';   pvptr:@emuKeybMask[3]; mask: Byte(not $01)),
    (name:'2';   pvptr:@emuKeybMask[3]; mask: Byte(not $02)),
    (name:'3';   pvptr:@emuKeybMask[3]; mask: Byte(not $04)),
    (name:'4';   pvptr:@emuKeybMask[3]; mask: Byte(not $08)),
    (name:'5';   pvptr:@emuKeybMask[3]; mask: Byte(not $10)),
    (name:'0';   pvptr:@emuKeybMask[4]; mask: Byte(not $01)),
    (name:'9';   pvptr:@emuKeybMask[4]; mask: Byte(not $02)),
    (name:'8';   pvptr:@emuKeybMask[4]; mask: Byte(not $04)),
    (name:'7';   pvptr:@emuKeybMask[4]; mask: Byte(not $08)),
    (name:'6';   pvptr:@emuKeybMask[4]; mask: Byte(not $10)),
    (name:'p';   pvptr:@emuKeybMask[5]; mask: Byte(not $01)),
    (name:'o';   pvptr:@emuKeybMask[5]; mask: Byte(not $02)),
    (name:'i';   pvptr:@emuKeybMask[5]; mask: Byte(not $04)),
    (name:'u';   pvptr:@emuKeybMask[5]; mask: Byte(not $08)),
    (name:'y';   pvptr:@emuKeybMask[5]; mask: Byte(not $10)),
    (name:'ent'; pvptr:@emuKeybMask[6]; mask: Byte(not $01)),
    (name:'l';   pvptr:@emuKeybMask[6]; mask: Byte(not $02)),
    (name:'k';   pvptr:@emuKeybMask[6]; mask: Byte(not $04)),
    (name:'j';   pvptr:@emuKeybMask[6]; mask: Byte(not $08)),
    (name:'h';   pvptr:@emuKeybMask[6]; mask: Byte(not $10)),
    (name:'spc'; pvptr:@emuKeybMask[7]; mask: Byte(not $01)),
    (name:'sym'; pvptr:@emuKeybMask[7]; mask: Byte(not $02)),
    (name:'m';   pvptr:@emuKeybMask[7]; mask: Byte(not $04)),
    (name:'n';   pvptr:@emuKeybMask[7]; mask: Byte(not $08)),
    (name:'b';   pvptr:@emuKeybMask[7]; mask: Byte(not $10)));


function GetPCKeyName (vk: Word; shifts: Byte): string;
function PCKeyNameToCode (name: string; out vk: Word; out shifts: Byte): Boolean;
function BindPCKeyToZX (vk: Word; shifts: Byte; zxname: string): Boolean;

function PCJoyNameValid (name: string): Boolean;
function BindPCJoyToZX (name, zxname: string): Boolean;


implementation


function GetPCKeyName (vk: Word; shifts: Byte): string;
var
  f: Integer;
begin
  if (vk = 0) and (shifts = 0) then
  begin
    result := 'autofire'; exit;
  end;

  result := '';
  if (shifts and emuSSCtrl) <> 0 then result := result+'ctrl+';
  if (shifts and emuSSAlt) <> 0 then result := result+'alt+';
  if (shifts and emuSSShift) <> 0 then result := result+'shift+';

  for f := 0 to High(pcVKNames) do
  begin
    if pcVKNames[f].code = vk then
    begin
      result := result+pcVKNames[f].name;
      exit;
    end;
  end;

  if (vk >= ord('0')) and (vk <= ord('9')) then
  begin
    result := result+chr(vk);
    exit;
  end;

  if (vk >= ord('A')) and (vk <= ord('Z')) then
  begin
    result := result+chr(vk+32);
    exit;
  end;

  result := result+'#'+IntToHex(vk, 2);
end;

function PCKeyNameToCode (name: string; out vk: Word; out shifts: Byte): Boolean;
var
  f: Integer;
  s: string;
begin
  if name = 'autofire' then
  begin
    vk := 0; shifts := 0; result := true; exit;
  end;

  shifts := 0; result := false;
  repeat
    name := Trim(name);
    f := Pos('+', name);
    if f = 0 then break;
    s := TrimRight(Copy(name, 1, f-1)); Delete(name, 1, f);
    if s = 'ctrl' then shifts := shifts or emuSSCtrl
    else if s = 'alt' then shifts := shifts or emuSSAlt
    else if s = 'shift' then shifts := shifts or emuSSShift
    else exit;
  until false;

  for f := 0 to High(pcVKNames) do
  begin
    if pcVKNames[f].name = name then
    begin
      vk := pcVKNames[f].code;
      result := true;
      exit;
    end;
  end;

  if (name >= '0') and (name <= '9') then
  begin
    vk := ord(name[1]);
    result := true;
    exit;
  end;

  if (name >= 'a') and (name <= 'z') then
  begin
    vk := ord(name[1])-32;
    result := true;
    exit;
  end;
end;

function BindPCKeyToZX (vk: Word; shifts: Byte; zxname: string): Boolean;
var
  f, c: Integer;
  n: string;
begin
  result := false; if (shifts and (not emuSSShift)) <> 0 then exit;
  //f := pcVKNames[f].code;

  //if vk = ord('J') then WriteLn('***J*** ', vk, '; ', shifts, '; ', zxname);

  zxname := Trim(zxname);
  if zxname = '' then exit;
  // find zx key(s)
  c := 0;
  while (zxname <> '') and (c < 2) do
  begin
    // get word
    f := Pos('+', zxname); if f < 1 then f := Length(zxname)+1;
    n := TrimRight(Copy(zxname, 1, f-1));
    zxname := Trim(Copy(zxname, f+1, Length(zxname)));

    // find zx key
    f := High(zxKNames); while (f >= 0) and (zxKNames[f].name <> n) do Dec(f);
    if f < 0 then break;

    // set masks
    pcKeyMap[shifts <> 0, vk].ports[c] := zxKNames[f].pvptr;
    pcKeyMap[shifts <> 0, vk].masks[c] := zxKNames[f].mask;
    Inc(c);
  end;

  if zxname <> '' then exit;
  result := true;
end;

function PCJoyNameValid (name: string): Boolean;
begin
  name := LowerCase(Trim(name));
  result := false;
  if (Length(name) < 4) or (name[1] <> 'j') or not (name[2] in ['1','2']) then exit;
  Delete(name, 1, 2);
  result := (name = 'left') or (name = 'right') or (name = 'up') or (name = 'down') or
    (name = 'b1') or (name = 'b2') or (name = 'b3') or (name = 'b4');
end;

function BindPCJoyToZX (name, zxname: string): Boolean;
var
  f, c, jn, ja: Integer;
  n: string;
begin
  result := false;

  name := LowerCase(Trim(name));
  if (Length(name) < 4) or (name[1] <> 'j') or not (name[2] in ['1','2']) then exit;
  jn := ord(name[2])-ord('1');
  Delete(name, 1, 2);
  ja := -1;
  if name = 'left' then ja := 0
  else if name = 'right' then ja := 1
  else if name = 'up' then ja := 2
  else if name = 'down' then ja := 3
  else if name = 'b1' then ja := 4
  else if name = 'b2' then ja := 5
  else if name = 'b3' then ja := 6
  else if name = 'b4' then ja := 7;
  if ja < 0 then exit;

  zxname := Trim(zxname);
  if zxname = '' then exit;

  // find zx key(s)
  c := 0;
  while (zxname <> '') and (c < 2) do
  begin
    // get word
    f := Pos('+', zxname); if f < 1 then f := Length(zxname)+1;
    n := TrimRight(Copy(zxname, 1, f-1));
    zxname := Trim(Copy(zxname, f+1, Length(zxname)));

    // find zx key
    f := High(zxKNames); while (f >= 0) and (zxKNames[f].name <> n) do Dec(f);
    if f < 0 then break;

    // set masks
    pcJoyMap[jn, ja].ports[c] := zxKNames[f].pvptr;
    pcJoyMap[jn, ja].masks[c] := zxKNames[f].mask;
    Inc(c);
  end;

  if zxname <> '' then exit;
  result := true;
end;


end.
