;(setq lasm_main *t*)


(setq asm-outfile-h -1)
(setq asm-fixfile-h *f*)
(setq asm-infile-h -1)
(setq asm-pc 0)    ; current PC
(setq asm-pc-st 0) ; PC for $



(defun lasm-emit-byte (mcode)
  (if lasm-real-labels (fwrite-byte asm-outfile-h (lasm-norm-byte mcode)))
  (setq asm-pc (incr asm-pc))
)

(defun lasm-emit-word (mcode)
  (if lasm-real-labels (fwrite-word asm-outfile-h (lasm-norm-word mcode)))
  (setq asm-pc (+ asm-pc 2))
)

(defun lasm-new-org (addr)
  (if lasm-real-labels (begin
      (fwrite-byte asm-fixfile-h 0)
      (fwrite-word asm-fixfile-h addr)
    )
  )
)

(defun lasm-emit-addr-byte (mcode)
  (if lasm-real-labels
    (if lasm-ext-label (begin
      (fwrite-byte asm-fixfile-h 3)
      (fwrite-word asm-fixfile-h (fcurpos asm-outfile-h))
      (fwrite-byte asm-fixfile-h (length lasm-ext-label))
      (fwrite-str asm-fixfile-h lasm-ext-label)
      )
     (begin
       (fwrite-byte asm-fixfile-h 1)
       (fwrite-word asm-fixfile-h (fcurpos asm-outfile-h))
     )
    )
  )
  (lasm-emit-byte mcode)
)

(defun lasm-emit-addr (mcode)
  (if lasm-real-labels
    (if lasm-ext-label (begin
      (fwrite-byte asm-fixfile-h 4)
      (fwrite-word asm-fixfile-h (fcurpos asm-outfile-h))
      (fwrite-byte asm-fixfile-h (length lasm-ext-label))
      (fwrite-str asm-fixfile-h lasm-ext-label)
      )
     (begin
       (fwrite-byte asm-fixfile-h 2)
       (fwrite-word asm-fixfile-h (fcurpos asm-outfile-h))
     )
    )
  )
  (lasm-emit-word mcode)
)

(defun lasm-emit-list (code-list)
  (for () code-list (setq code-list (cdr code-list))
    (lasm-emit-byte (car code-list))
  )
)

(defun assemble-str (str)
  (if (empty-str? str) (return))
  (asm-tokenize str)
  (setq asm-pc-st asm-pc)
  (if (eq lasm-cmd "equ") (lasm-do-equ)
   (begin (lasm-do-label) (asm-do-cmd))
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun lasm-next-line ()
  (setq lasm-line-n (incr lasm-line-n))
  (if (zero? (& lasm-line-n 31)) (print "\r" lasm-line-n))
  ;(print "\r" lasm-line-n)
  ;(trim-right (freadln asm-infile-h))
  (freadln asm-infile-h)
)

(defun lasm-do-macro (cmd codez | df s xx)
  ;(println "MACRO")
  (if (<> lasm-argc 1) (bad-syntax "invalid macro definition"))
  (setq df
    (|| (chr 40) "defun xx (cmd) " (chr 40) "begin "
        "(locals marglist margc)" (chr 10)
        "(setq marglist lasm-argv  margc lasm-argc)"
    )
  )
  (for (setq s *nil*) (<> (lower-case (trim-left s)) "endm") (setq s (lasm-next-line))
    (begin
      ;(println "s=!"" s "!"")
      (setq df (|| df (chr 10) s))
    )
  )
  (setq df (|| df (chr 10) (chr 41) (chr 41)))
  ;(println "df=!"" df "!"")
  (if (not lasm-real-labels) (begin
    (evalstr df)
    (setq df (evalstr "xx"))
    ;(lprintln "!"" df "!"")
    (lasm-add-operator (lower-case (car lasm-argv)) df)
  ))
)


(defun lasm-do-include (cmd codez)
  (switch lasm-argc
    1 (begin (locals fh oldfh oldlc)
        (setq fh (fopen (car lasm-argv) "r-"))
        (if (not fh) (bad-syntax (|| "invalid file in INCLUDE: " (car lasm-argv))))
        (fclose fh)
        (setq oldfh asm-infile-h  oldlc lasm-line-n)
        (println "\r" (car lasm-argv) "   ")
        (asm-compile-file (car lasm-argv))
        (setq asm-infile-h oldfh  lasm-line-n oldlc)
      )
    (bad-syntax "invalid INCLUDE syntax")
  )
)

(defun asm-compile-file (fname | s)
  (setq asm-infile-h (fopen fname "r-")  lasm-line-n 0)
  (if (not asm-infile-h) (throw (0 "no file" fname)))
  (setq err (catch
    (while (< (fcurpos asm-infile-h) (fsize asm-infile-h)) (begin
      (setq s (lasm-next-line))
      (assemble-str s)
    ))
    (println "\r" lasm-line-n " lines.")
  ))
  (fclose asm-infile-h)
  (if err (println "***ERROR at line " (car err) ": " (cadr err)))
)


(defun assemble-file (fname outname | oh lcnt n s fbuf)
  (if (not (extract-ext fname)) (setq fname (|| fname ".saz")))
  (if (not outname) (setq outname (change-ext fname ".bin")))
  (if (not (extract-ext outname)) (setq outname (|| outname ".bin")))

  (println "assembling " fname)
  (println "pass 1...")
  (setq lasm-real-labels *f*  asm-pc 0  asm-pc-st 0) (lasm-new-org 0)
  (asm-compile-file fname)

  (setq asm-outfile-h (fopen outname "crw"))
  (setq asm-fixfile-h (fopen (|| outname ".fix") "crw"))
  (println "pass 2...")
  (setq lasm-real-labels *t*  asm-pc 0  asm-pc-st 0) (lasm-new-org 0)
  (asm-compile-file fname)

  ;    
  ;  ,    ORG   
  (println "building object file...")

  ;  
  (for (setq lcnt 0  n lasm-labels) n (setq n (cdr n))
    (setq lcnt (+ lcnt (if (cadddr (car n)) 0 1) ))
  )

  (setq oh (fopen (change-ext outname ".zobj") "c"))
  (fwrite-str oh "ZOBJ 0.1")  ; signature
  (fwrite-int oh (fsize asm-outfile-h))
  (fwrite-int oh (fsize asm-fixfile-h))
  (fwrite-int oh lcnt)  ; # of labels
  ; code
  (fseek asm-outfile-h 0)
  (setq fbuf (fread-buf asm-outfile-h (fsize asm-outfile-h)))
  (fwrite-buf oh -1 fbuf)
  ; fixups
  (fseek asm-fixfile-h 0)
  (fread-buf asm-fixfile-h (fsize asm-fixfile-h) fbuf)
  (fwrite-buf oh -1 fbuf)
  ; labels
  (for (setq n lasm-labels) n (setq n (cdr n))
    (if (not (cadddr (car n)))
      (begin
        (setq s (caar n))
        (if (not (within (length s) 1 255)) (bottom "shit!!"))
        (fwrite-byte oh (if (caddar n) 1 0))
        (fwrite-word oh (cadar n))
        (fwrite-byte oh (length s)) (fwrite-str oh s)
      )
    )
  )
  (fclose oh)

  (fclose asm-outfile-h) (setq asm-outfile-h -1)
  (fclose asm-fixfile-h) (setq asm-fixfile-h -1)

  (println "deleting temporary files...")
  ;(delete-file outname)
  (delete-file (|| outname ".fix"))
  (println "done.")
)
