// taken from Project JEDI Code Library (JCL), so I thing that this is
// under JEDI's MPL? i'm not sure yet...
// I don't know who exactly was the author of the code
//
unit cpuSpeedU;

interface

uses
  Windows;


type
  TFreqInfo = record
    rawFreq: Cardinal;
    normFreq: Cardinal;
    inCycles: Cardinal;
    exTicks: Cardinal;
  end;


function GetCPUSpeed (var cpuSpeed: TFreqInfo): Boolean;


var
  processorCount: Cardinal = 0;


implementation


function ReadTimeStampCounter (): Int64; assembler;
asm
  dw  $310F
end;

function RoundFrequency (const frequency: Integer): Integer;
const
  nf: array [0..8] of Integer = (0, 20, 33, 50, 60, 66, 80, 90, 100);
var
  freq, rf, f: Integer;
  hi, lo: Byte;
begin
  rf := 0; freq := frequency mod 100;
  for f := 0 to 8 do
  begin
    if freq < nf[f] then
    begin
      hi := f; lo := f-1;
      if nf[hi]-freq > freq-nf[lo] then rf := nf[lo]-freq else rf := nf[hi]-freq;
      break;
    end;
  end;

  result := frequency+rf;
end;

function GetCPUSpeed (var cpuSpeed: TFreqInfo): Boolean;
var
  t0, t1: Int64;
  countFreq: Int64;
  freq, freq2, freq3, total: Int64;
  totalCycles, cycles: Int64;
  stamp0, stamp1: Int64;
  totalTicks, ticks: Double;
  tries, priority: Integer;
  thread: THandle;
begin
  stamp0 := 0; stamp1 := 0; freq  := 0; freq2 := 0; freq3 := 0;
  tries := 0; totalCycles := 0; totalTicks := 0; total := 0;

  thread := GetCurrentThread();
  result := QueryPerformanceFrequency(countFreq);
  if result then
  begin
    while ((tries < 3) or ((tries < 20) and ((abs(3*freq-total) > 3) or
          (abs(3*freq2-total) > 3) or (abs(3*freq3-total) > 3)))) do
    begin
      Inc(tries);
      freq3 := freq2; freq2 := freq;
      QueryPerformanceCounter(t0); t1 := t0;

      priority := GetThreadPriority(thread);
      if priority <> THREAD_PRIORITY_ERROR_RETURN then
        SetThreadPriority(thread, THREAD_PRIORITY_TIME_CRITICAL);
      try
        while t1 - t0 < 50 do
        begin
          QueryPerformanceCounter(t1);
          stamp0 := ReadTimeStampCounter;
        end;
        t0 := t1;

        while t1 - t0 < 1000 do
        begin
          QueryPerformanceCounter(t1);
          stamp1 := ReadTimeStampCounter;
        end;
      finally
        if priority <> THREAD_PRIORITY_ERROR_RETURN then
          SetThreadPriority(thread, priority);
      end;

      cycles := stamp1-stamp0;
      ticks := t1-t0;
      ticks := ticks*100000;

      // avoid division by zero
      if countFreq = 0 then ticks := High(Int64) else ticks := ticks/(countFreq/10);

      totalTicks := totalTicks+ticks; totalCycles := totalCycles+cycles;

      // avoid division by zero
      if ticks = 0 then freq := High(freq) else freq := round(cycles/ticks);

      total := freq+freq2+freq3;
    end;

    // avoid division by zero
    if totalTicks = 0 then
    begin
      freq3 := High(freq3);
      freq2 := High(freq2);
      cpuSpeed.rawFreq := High(cpuSpeed.rawFreq);
    end
    else
    begin
      freq3 := round((totalCycles*10)/totalTicks); // freq. in multiples of 10^5 Hz
      freq2 := round((totalCycles*100)/totalTicks); // freq. in multiples of 10^4 Hz
      cpuSpeed.rawFreq := round(totalCycles/totalTicks);
    end;

    cpuSpeed.normFreq := cpuSpeed.rawFreq;

    if freq2-(freq3*10) >= 6 then Inc(freq3);


    freq := cpuSpeed.rawFreq*10;
    if (freq3-freq) >= 6 then Inc(cpuSpeed.normFreq);

    cpuSpeed.exTicks := round(totalTicks);
    cpuSpeed.inCycles := totalCycles;

    cpuSpeed.normFreq := RoundFrequency(cpuSpeed.normFreq);

    result := true;
  end;
end;


procedure InitSysInfo ();
var
  systemInfo: TSystemInfo;
begin
  { processor information related initialization }
  FillChar(systemInfo, SizeOf(systemInfo), 0);
  GetSystemInfo(systemInfo);
  processorCount := systemInfo.dwNumberOfProcessors;
end;


begin
  InitSysInfo();
end.
