// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// Write/WriteLn driver for command console, v0.0.1
//
// control codes:
//   CR, LF, BS
//   TAB: tab space = 4
//   #0#0#n: set X (0-wconWidth-1)
//   #0#1#n: set/reset word wrap mode (until next CR/LF)
//
//   userData[1]: current x
//   userData[2]: current command (#0, #FF, #1)
//   userData[3]: do #10 before print any char
//   userData[4]: wordwrap mode
//
{$I-}
unit dspConDriverU;

interface

uses
  Windows;


const
  wconWidth = 52;
  wconMinHeight = 2;
    // plus one editor line
  wconBufSize = 512;


type
  TWConString = ShortString;


var
  wconLogFile: string = '';
  wconLogging: Boolean = false;
  wconLogDateTime: Boolean = false;
  wconBuf: packed array [0..wconBufSize] of TWConString;
  wconDumpToWinConsole: Boolean = false;

  wconOutputCount: Integer;
  wconScrollCount: Integer;


implementation


type
  TDevFunc = function (var f: TTextRec): Integer;


const
  udX = 1;
  udCmd = 2;
  udDoNL = 3;
  udWrap = 4;


function IntToStr2 (n: Integer): string;
begin
  Str(n:2, result);
  if n < 10 then result[1] := '0';
end;

function Int2Str (n: Integer): string;
begin
  Str(n, result);
end;

function LogTime (): string;
var
  st: TSystemTime;
begin
  GetLocalTime(st);
  result := IntToStr2(st.wHour)+':'+IntToStr2(st.wMinute)+':'+IntToStr2(st.wSecond)+', '+
    IntToStr2(st.wDay)+'/'+IntToStr2(st.wMonth)+'/'+Int2Str(st.wYear) + ': ';
end;

function FileExists (const fileName: string): Boolean;
var
  handle: THandle;
  findData: TWin32FindData;
begin
  handle := FindFirstFile(PChar(fileName), findData);
  if handle <> INVALID_HANDLE_VALUE then
  begin
    Windows.FindClose(handle);
    if (findData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then
    begin
      result := true;
      exit;
    end;
  end;
  result := false;
end;

function LogFileOpen (const fName: string): Integer;
var
  hd: Integer;
begin
  result := -1;
  if fName = '' then exit;
  // no such file, try to create one
  hd := Windows.CreateFile(PChar(fName), 0, FILE_SHARE_READ or FILE_SHARE_WRITE, nil,
    OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
  if hd = -1 then exit; // not a device, not a file
  CloseHandle(hd);

  if not FileExists(fName) then exit; // a device

  // normal file
  result := Windows.CreateFile(PChar(fName), GENERIC_WRITE, FILE_SHARE_READ, nil,
    OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
end;

procedure LogStr (s: string);
var
  fh: Integer;
  sp, ep: Integer;
  nl: Char;
  wrt: LongWord;

  procedure WriteTime ();
  var
    ts: string;
  begin
    ts := LogTime(); if ts = '' then exit;
    WriteFile(fh, ts[1], Length(ts), wrt, nil);
    //WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), ts[1], Length(ts), wrt, nil);
  end;

begin
  if not wconLogging then exit;
  if s = '' then exit;
  nl := #10;

  fh := LogFileOpen(wconLogFile);
  if fh = -1 then exit;
  if SetFilePointer(fh, 0, nil, FILE_END) <> $FFFFFFFF then
  begin
    //AllocConsole();
    //WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), s[1], Length(s), wrt, nil);

    sp := 1;
    while sp <= Length(s) do
    begin
      ep := sp;
      while (ep <= Length(s)) and not (s[ep] in [#10,#13]) do Inc(ep);
      if ep-sp > 0 then
      begin
        WriteFile(fh, s[sp], ep-sp, wrt, nil);
        //WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), s[sp], ep-sp, wrt, nil);
      end;
      sp := ep;
      if sp <= Length(s) then
      begin
        WriteFile(fh, nl, 1, wrt, nil);
        //WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), nl, 1, wrt, nil);
        if wconLogDateTime then WriteTime();
        Inc(sp);
        if (sp <= Length(s)) and (s[sp-1] = #13) and (s[sp] = #10) then Inc(sp);
      end;
    end;
  end;
  CloseHandle(fh);
end;


procedure ProcessOutput (var tf: TTextRec; buf: PChar; count: Integer);
var
  x: Integer;
  ch: Char;

  procedure Scroll ();
  begin
    Move(wconBuf[1], wconBuf[0], SizeOf(TWConString)*wconBufSize);
    FillChar(wconBuf[wconBufSize], SizeOf(TWConString), ' ');
    wconBuf[wconBufSize][0] := #0;
    Inc(wconScrollCount);
  end;

  procedure WWrap ();
  var
    n: Integer;
    s: ShortString;
  begin
    n := Byte(wconBuf[wconBufSize][0]); if n > wconWidth then n := wconWidth;
    if x > n then n := x;
    if n < wconWidth then FillChar(wconBuf[wconBufSize][n+1], wconWidth-n, ' ');
    wconBuf[wconBufSize][0] := Char(n);

    if tf.userData[udWrap] <> 0 then
    begin
      n := x+1; if n > wconWidth then n := wconWidth;
      while (n > 0) and (wconBuf[wconBufSize][n] <> ' ') do Dec(n);
      if n > 0 then
      begin
        s := Copy(wconBuf[wconBufSize], n+1, wconWidth);
        FillChar(wconBuf[wconBufSize][n], wconWidth-n, ' ');
        while (n > 0) and (wconBuf[wconBufSize][n] = ' ') do Dec(n);
        wconBuf[wconBufSize][0] := chr(n);
        n := 1; while (n < Length(s)) and (s[n] <= ' ') do Inc(n);
        Delete(s, 1, n-1);
      end;
      Scroll();
      wconBuf[wconBufSize] := s;
      x := Length(s);
    end
    else
    begin
      Scroll();
      x := 0;
    end;
  end;

  procedure NewLine ();
  begin
    WWrap(); tf.userData[udWrap] := 0; tf.userData[udDoNL] := 0;
  end;

  procedure EmitChar ();
  begin
    if tf.userData[udDoNL] <> 0 then NewLine();

    Inc(wconOutputCount);
    if x > wconWidth-1 then
    begin
      wconBuf[wconBufSize][0] := Char(wconWidth);
      WWrap();
    end;

    if (tf.userData[udWrap] <> 0) and (ch = ' ') and (x = 0) then exit;
    wconBuf[wconBufSize][x+1] := ch;
    Inc(x);
  end;

begin
  while count > 0 do
  begin
    x := tf.userData[udX]; if x > wconWidth then x := wconWidth;

    if tf.userData[udCmd] <> 0 then
    begin
      // prev was control code
      if tf.userData[udCmd] = 255 then
      begin
        if buf^ in [#0..#1] then
        begin
          tf.userData[udCmd] := Byte(buf^);
          Inc(buf); Dec(count);
        end
        else tf.userData[udCmd] := 0;
      end
      else
      begin
        if tf.userData[udDoNL] <> 0 then NewLine();
        // prev was command
        case tf.userData[udCmd] of
          0: // set x
            begin
              tf.userData[udX] := Byte(buf^);
              if tf.userData[udX] > wconWidth then tf.userData[udX] := wconWidth;
            end;
          1: // set wordwrap
            begin
              if buf^ <> #0 then tf.userData[udWrap] := 1 else tf.userData[udWrap] := 0;
            end;
        end;
        tf.userData[udCmd] := 0;
        Inc(buf); Dec(count);
      end;
    end
    else
    begin
      while count > 0 do
      begin
        ch := buf^; Inc(buf); Dec(count);
        case ch of
          #0: begin tf.userData[udCmd] := 255; break; end;
          #8:
            begin
              if x > Byte(wconBuf[wconBufSize][0]) then wconBuf[wconBufSize][0] := Char(x);
              if x > 0 then Dec(x);
            end;
          #9: // tab
            begin
              ch := ' ';
              repeat
                EmitChar();
              until (x mod 4) = 0;
            end;
          #10:
            begin
              if tf.userData[udDoNL] <> 0 then NewLine();
              tf.userData[udDoNL] := 1;
            end;
          #13:
            begin
              if x > Byte(wconBuf[wconBufSize][0]) then wconBuf[wconBufSize][0] := Char(x);
              x := 0;
              wconBuf[wconBufSize] := 'de'; x := 2;
            end
          else EmitChar();
        end;
      end;
    end;
    if x > Byte(wconBuf[wconBufSize][0]) then wconBuf[wconBufSize][0] := Char(x);
    tf.userData[udX] := x;
  end;
end;

function DevOpen (var f: TTextRec): Integer;
var
  c: Integer;
begin
  wconOutputCount := 0; wconScrollCount := 0;
  f.userData[udX] := 0;
  f.userData[udCmd] := 0;
  f.userData[udDoNL] := 0;
  f.userData[udWrap] := 0;
  FillChar(wconBuf, SizeOf(wconBuf), ' ');
  for c := 0 to wconBufSize do wconBuf[c][0] := #0;
  f.bufPos := 0;
  f.bufEnd := 0;
  result := 0;
end;

function DevInOut (var f: TTextRec): Integer;
var
  buf: PChar;
  sz: Integer;
  s: string;
  dm: LongWord;
begin
  result := 0;
  buf := f.BufPtr; sz := f.BufPos;
  if sz > 0 then
  begin
    if wconDumpToWinConsole then WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), buf^, sz, dm, nil);
    ProcessOutput(f, buf, sz);
    if wconLogging then
    begin
      SetString(s, buf, sz);
      LogStr(s);
    end;
  end;
  f.bufPos := 0;
  f.bufEnd := 0;
end;

function DevFlush (var f: TTextRec): Integer;
begin
  result := DevInOut(f);
end;

function DevClose (var f: TTextRec): Integer;
begin
  result := 0;
end;


procedure InitDriver ();
begin
  with TTextRec(output) do
  begin
    Mode := fmClosed;
    if BufPtr = nil then
    begin
      BufSize := SizeOf(Buffer);
      BufPtr := @Buffer;
    end;
    OpenFunc := @DevOpen;
    InOutFunc := @DevInOut;
    FlushFunc := @DevFlush;
    CloseFunc := @DevClose;
    Name[0] := #0;
  end;
  Rewrite(output);
end;


function ExtractFilePath (const s: string): string;
var
  pos: Integer;
begin
  result := s;
  pos := Length(result);
  while (pos > 0) and not (result[pos] in ['/', '\']) do Dec(pos);
  Delete(result, pos+1, Length(result));
end;


initialization
  wconLogFile := ExtractFilePath(ParamStr(0)) + 'dsp.log';
  InitDriver();
end.
