// high-performance timer
// based on ZenTimer by Vadim Bodrov, which is based on
// ZenTimer by KendallB
//
// hacking, choping, etc...
// Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
unit hpTimerU;

interface


const
  CPU_UNKNOWN     = 0;            // Unknown proccessor
  CPU_I386        = 1;            // Intel 80386 processor
  CPU_I486        = 2;            // Intel 80486 processor
  CPU_PENTIUM     = 3;            // Intel Pentium(R) processor
  CPU_PENTIUMPRO  = 4;            // Intel PentiumPro(R) processor
  CPU_PENTIUMII   = 5;            // Intel PentiumII(R) processor
  CPU_PENTIUMIII  = 6;            // Intel PentiumIII(R) processor
  CPU_UNKPENTIUM  = 7;            // Unknown Intel Pentium family processor

  cpuTypeNames: array [0..7] of string = (
    'Unknown',
    'Intel 80386',
    'Intel 80486',
    'Intel Pentium(R)',
    'Intel PentiumPro(R)',
    'Intel PentiumII(R)',
    'Intel PentiumIII(R)',
    'Unknown Pentium');


type
  PX86LargeInteger = ^TX86LargeInteger;
  TX86LargeInteger = packed record
    lowPart: LongWord;
    highPart: LongWord;
  end;


(* Routines to obtain CPU information *)

{ DESCRIPTION:
  Returns the type of processor in the system.

  RETURNS:
  Numerical identifier for the installed processor

  REMARKS:
  Returns the type of processor in the system. Note that if the CPU is an
  unknown Pentium family processor that we don't have an enumeration for,
  the return value will be greater than or equal to the value of CPU86UNKPENTIUM
  (depending on the value returned by the CPUID instruction).}
function CPU86GetProcessorType (): LongWord;

function CPU86GetProcessorId (): string;

{ DESCRIPTION:
  Returns TRUE if the processor is an Intel clone.}
function CPU86IsIntelClone (): Boolean;

{  DESCRIPTION:
  Returns true if the processor supports Intel MMX extensions.

  RETURNS:
  True if MMX is available, false if not.

  REMARKS:
  This function determines if the processor supports the Intel MMX extended
  instruction set. If the processor is not an Intel or Intel clone CPU, this
  function will always return false.}
function CPU86HaveMMX (): Boolean;

{ DESCRIPTION:
  Returns true if the processor supports the RDTSC instruction

  RETURNS:
  True if the RTSC instruction is available, false if not.

  REMARKS:
  This function determines if the processor supports the Intel RDTSC
  instruction, for high precision timing. If the processor is not an Intel or
  Intel clone CPU, this function will always return false.}
function CPU86HaveRDTSC (): Boolean;

{ REMARKS:
  On processors supporting the Read Time Stamp opcode, compare elapsed
  time on the High-Resolution Counter with elapsed cycles on the Time
  Stamp Register.

  The inner loop runs up to 20 times or until the average of the previous
  three calculated frequencies is within 1 MHz of each of the individual
  calculated frequencies. This resampling increases the accuracy of the
  results since outside factors could affect this calculation.}
//function CPU86GetRDTSCSpeed (): LongInt;

{ DESCRIPTION:
  Returns the speed of the processor in Mhz.

  RETURNS:
  processor speed in Mhz.

  REMARKS:
  This function returns the speed of the CPU in Mhz. Note that if the speed
  cannot be determined, this function will return 0.}
function CPU86GetProcessorSpeed (): LongWord; // better use CPU86GetDetectedSpeed

function CPU86GetDetectedSpeed (): LongWord;

{ determines the CPU type using the CPUID instruction.}
function CPU86GetCPUIDFeatures (): LongWord;


(* Long period timing routines (times up to 1 hour) *)
function HPTimerResolution (): Double;

procedure HPTimerOn (out tmStart: TX86LargeInteger);
function HPTimerLap (const tmStart: TX86LargeInteger): LongWord;
procedure HPTimerOff (out tmEnd: TX86LargeInteger);
function HPTimerCount (const tmStart, tmEnd: TX86LargeInteger): LongWord;


procedure HPTimerMainInit ();
  // call this before any other timer proc!


implementation


function GetCurrentThread (): LongWord; stdcall;
  external 'kernel32.dll' name 'GetCurrentThread';
function GetThreadPriority (hThread: LongWord): LongWord; stdcall;
  external 'kernel32.dll' name 'GetThreadPriority';
function SetThreadPriority(hThread: LongWord; nPriority: LongWord): LongBool; stdcall;
  external 'kernel32.dll' name 'SetThreadPriority';

function QueryPerformanceCounter (out lpPerformanceCount: TX86LargeInteger): LongBool; stdcall;
  external 'kernel32.dll' name 'QueryPerformanceCounter';
function QueryPerformanceFrequency (out lpFrequency: TX86LargeInteger): LongBool; stdcall;
  external 'kernel32.dll' name 'QueryPerformanceFrequency';

function timeGetTime (): LongWord; stdcall; external 'winmm.dll' name 'timeGetTime'


const
  GEN_INTEL_ID: packed array [0..11] of Char = 'GenuineIntel';

  TOLERANCE       = 1;
  MAXCLOCKS       = 150;
  ROUND_THRESHOLD = 6;
  ITERATIONS      = 16000;
  MAX_TRIES       = 20;
  SAMPLINGS       = 10;

  CPU86mask       = $7FFF;
  CPU86IntelClone = $8000;
  CPU86HaveMMX_   = $00800000;
  CPU86HaveRDTSC_ = $00000010;


var
  cpuSpeed: LongInt = -1;
  haveRDTSC: Boolean = FALSE;
  start: LongWord = 0;
  finish: LongWord = 0;
  zTimerBIOS: LongWord = 0;
  countFreq: TX86LargeInteger;
  havePerformanceCounter: Boolean;
  x86ID: string[12] = '';


function CPU86GetProcessorId (): string;
begin
  result := x86ID;
end;

function CPU86GetDetectedSpeed (): LongWord;
begin
  result := LongWord(cpuSpeed div 1000000);
end;


function CPU86Check80486 (): Boolean; assembler;
// determines if we have an i486 processor.
asm
  pushfd               // get original EFLAGS
  pop     eax
  mov     ecx,eax
  xor     eax,$200000  // flip ID bit in EFLAGS
  push    eax          // save new EFLAGS value on stack
  popfd                // replace current EFLAGS value
  pushfd               // get new EFLAGS
  pop     eax          // store new EFLAGS in EAX
  xor     eax,ecx      // can not toggle ID bit, processor=80486
  jnz     @@1
  mov     eax,1        // we dont have a pentium
  jmp     @@done
@@1:
  xor     eax,eax      // we have pentium or later
@@done:
end;

function CPU86HaveCPUID (): Boolean; assembler;
// determines if we have support for the CPUID instruction.
asm
  pushfd               // get original EFLAGS
  pop     eax
  mov     ecx,eax
  xor     eax,$200000  // flip ID bit in EFLAGS
  push    eax          // save new EFLAGS value on stack
  popfd                // replace current EFLAGS value
  pushfd               // get new EFLAGS
  pop     eax          // store new EFLAGS in EAX
  xor     eax,ecx      // can not toggle ID bit, processor=80486
  jnz     @@1
  xor     eax,eax      // we dont have CPUID support
  jmp     @@done
@@1:
  mov     eax,1        // we have CPUID support
@@done:
end;

function CPU86CheckCPUID (): LongWord; assembler;
// determines the CPU type using the CPUID instruction.
asm
  push    ebx  // make Delphi happy
  xor     eax,eax      // set up for CPUID instruction
  cpuid                // get and save vendor ID
  cmp     eax,1        // make sure 1 is valid input for CPUID
  jl      @@fail       // we dont have the CPUID instruction
  // store cpu id
  mov     byte ptr [x86ID],12
  mov     dword ptr [x86ID+1],ebx
  mov     dword ptr [x86ID+5],edx
  mov     dword ptr [x86ID+9],ecx
  //
  xor     eax,eax      // assume Genuine Intel
  cmp     dword ptr [GEN_INTEL_ID+0],ebx
  jne     @@notGenuineIntel
  cmp     dword ptr [GEN_INTEL_ID+4],edx
  jne     @@notGenuineIntel
  cmp     dword ptr [GEN_INTEL_ID+8],ecx
  je      @@haveGenuineIntel
@@notGenuineIntel:
  mov     eax,CPU86IntelClone   // set the clone flag
@@haveGenuineIntel:
  push    eax
  xor     eax,eax
  inc     eax
  cpuid               // get family/model/stepping/features
  and     eax,0F00h
  shr     eax,8       // isolate family
  and     eax,0Fh
  pop     ecx
  or      eax,ecx     // combine in the clone flag
  jmp     @@done
@@fail:
  xor     eax,eax
@@done:
  pop     ebx
end;

function CPU86getCPUIDModel (): LongWord; assembler;
// determines the CPU type using the CPUID instruction.
asm
  push    ebx  // make Delphi happy
  xor     eax,eax     // set up for CPUID instruction
  cpuid               // get and save vendor ID
  cmp     eax,1       // make sure 1 is valid input for CPUID
  jl      @@fail      // we dont have the CPUID instruction
  xor     eax,eax
  inc     eax
  cpuid               // get family/model/stepping/features
  and     eax,0F0h
  shr     eax,4       // isolate model
  jmp     @@done
@@fail:
  xor     eax,eax
@@done:
  pop     ebx
end;

function CPU86GetCPUIDFeatures (): LongWord; assembler;
// determines the CPU type using the CPUID instruction.
asm
  push    ebx  // make Delphi happy
  xor     eax,eax  // set up for CPUID instruction
  cpuid            // get and save vendor ID
  cmp     eax,1    // make sure 1 is valid input for CPUID
  jl      @@fail   // we dont have the CPUID instruction
  xor     eax,eax
  inc     eax
  cpuid            // get family/model/stepping/features
  mov     eax,edx
  jmp     @@done
@@fail:
  xor     eax,eax
@@done:
  pop     ebx
end;

function CPU86CheckClone (): LongWord; assembler;
// checks if the i386 or i486 processor is a clone or genuine Intel.
asm
  mov     ax,5555h    // check to make sure this is a 32-bit processor
  xor     dx,dx
  mov     cx,0002h
  div     cx          // perform division
  clc
  jnz     @@noClone
  jmp     @@clone
@@noClone:
  stc
@@clone:
  pushfd
  pop     eax         // get the flags
  and     eax,1
  xor     eax,1       // EAX=0 is probably Intel, EAX=1 is a clone
end;


procedure CPU86ReadTimeStamp (var time: TX86LargeInteger); assembler; register;
// reads the time stamp counter and returns the 64-bit result.
asm
  push    eax
  //db      00Fh,031h       // rdtsc. ???
  rdtsc
  pop     ecx
  //mov     ecx,[time]      // access directly without stack frame
  mov     [ecx],eax
  mov     [ecx+4],edx
end;

function CPU86DiffTime64 (const t1, t2: TX86LargeInteger; out t: TX86LargeInteger): LongWord; assembler; stdcall;
// computes the difference between two 64-bit numbers.
asm
  push    ebx
  mov     ecx,[t2]
  mov     eax,[ecx]       // EAX := t2.low
  mov     ecx,[t1]
  sub     eax,[ecx]
  mov     edx,eax         // EDX := low difference
  mov     ecx,[t2]
  mov     eax,[ecx+4]     // ECX := t2.high
  mov     ecx,[t1]
  sbb     eax,[ecx+4]     // EAX := high difference
  mov     ebx,[t]         // store the result
  mov     [ebx],edx       // store low part
  mov     [ebx+4],eax     // store high part
  mov     eax,edx         // return low part
  pop     ebx
end;

function CPU86CalcMicroSec (var count: TX86LargeInteger; freq: LongWord): LongWord; assembler; stdcall;
// computes the value in microseconds for the elapsed time with maximum
// precision. The formula we use is:
//
//      us = (((diff * 0x100000) / freq) * 1000000) / 0x100000)
//
// the power of two multiple before the first divide allows us to scale the
// 64-bit difference using simple shifts, and then the divide brings the
// final result into the range to fit into a 32-bit integer.
asm
  mov     ecx,[count]
  mov     eax,[ecx]        // EAX := low part
  mov     edx,[ecx+4]      // EDX := high part
  shld    edx,eax,20
  shl     eax,20           // diff*$100000
  div     [freq] // (diff*$100000)/freq
  mov     ecx,1000000
  xor     edx,edx
  mul     ecx              // ((diff*$100000)/freq)*1000000)
  shrd    eax,edx,20       // ((diff*$100000)/freq)*1000000)/$100000
end;

function GetCPUType (): LongWord;
var
  cpu, model, clone: LongWord;
begin
  if CPU86HaveCPUID() then
  begin
    cpu := CPU86CheckCPUID();
    clone := (cpu and CPU86IntelClone);
    case (cpu and CPU86mask) of
     4: cpu := CPU_I486;
     5: cpu := CPU_PENTIUM;
     6: begin
          model := CPU86getCPUIDModel();
          if model = 1 then cpu := CPU_PENTIUMPRO else cpu := CPU_PENTIUMII;
        end;
     7: cpu := CPU_PENTIUMIII;
     else cpu := CPU_UNKPENTIUM;
    end;
  end
  else
  begin
    clone := CPU86CheckClone()*CPU86IntelClone;
    if CPU86check80486 then cpu := CPU_I486
    else cpu := CPU_PENTIUM;
  end;
  result := cpu or clone;
end;

function CPU86GetProcessorType (): LongWord;
begin
  result := GetCPUType and $7FFF;
end;

function CPU86IsIntelClone: Boolean;
begin
  result := (GetCPUType and CPU86IntelClone) <> 0;
end;

function CPU86HaveMMX: Boolean;
begin
  if CPU86HaveCPUID then
   result:= (CPU86GetCPUIDFeatures and CPU86HaveMMX_) <> 0
  else
   result := FALSE;
end;

function CPU86HaveRDTSC (): Boolean;
begin
  if CPU86HaveCPUID then
   result := (CPU86GetCPUIDFeatures and CPU86HaveRDTSC_) <> 0
  else
   result := FALSE;
end;

procedure GetCounterFrequency (var freq: TX86LargeInteger);
begin
  if not QueryPerformanceFrequency(freq) then
  begin
    havePerformanceCounter := FALSE;
    freq.lowPart  := 100000;
    freq.highPart := 0;
  end else
    havePerformanceCounter := TRUE;
end;

procedure GetCounter (var t: TX86LargeInteger);
begin
  if havePerformanceCounter then QueryPerformanceCounter(t)
  else
  begin
    t.lowPart := timeGetTime()*100;
    t.highPart := 0;
  end;
end;

function CPU86GetRDTSCSpeed (): LongInt;
var
  t0, t1, count_freq: TX86LargeInteger;
  freq, freq2, freq3: LongWord;
  total, tries, total_cycles, cycles,
  stamp0, stamp1, total_ticks, ticks: LongWord;
var
  iPriority: LongInt;
  hThread: LongWord;
begin
  freq := 0; freq2 := 0;
  tries := 0; total_cycles := 0;
  stamp0 := 0; stamp1 := 0;
  total_ticks := 0;

  hThread := GetCurrentThread();
  iPriority := GetThreadPriority(hThread);
  if iPriority <> $7FFFFFFF then SetThreadPriority(hThread, 15); // time critical
  GetCounterFrequency(count_freq);
  repeat
    Inc(tries);      // increment number of times sampled
    freq3 := freq2;  // shift frequencies back
    freq2 := freq;

    // loop until 50 ticks have  passed since last read of hi-res counter.
    // this accounts for overhead later.
    GetCounter(t0);
    t1.lowPart := t0.lowPart; t1.highPart := t0.highPart;
    while (t1.lowPart-t0.lowPart) < 50 do
    begin
      GetCounter(t1);
      asm
        push  edx
        rdtsc
        mov   stamp0,eax
        pop   edx
      end;
    end;

    // loop until 1000 ticks have passed since last read of hi-res counter.
    // this allows for elapsed time for sampling.
    t0.lowPart := t1.lowPart; t0.highPart := t1.highPart;
    while (t1.lowPart-t0.lowPart) < 1000 do
    begin
      GetCounter(t1);
      asm
        push  edx
        rdtsc
        mov   stamp1,eax
        pop   edx
      end;
    end;

    // find the difference during the timing loop
    cycles := stamp1-stamp0;
    ticks := t1.lowPart-t0.lowPart;

    // note that some seemingly arbitrary mulitplies and divides are done below.
    // this is to maintain a high level of precision without truncating
    // the most significant data. according to what value ITERATIIONS is set
    // to, these multiplies and divides might need to be shifted for optimal
    // precision.

    ticks := ticks*100000;
    ticks := ticks div (count_freq.lowPart div 10);
    Inc(total_ticks, ticks);
    Inc(total_cycles, cycles);
    // round up if necessary
    if (ticks mod count_freq.lowPart) > (count_freq.lowPart div 2) then Inc(ticks);
    freq := cycles div ticks;          // Cycles / us  = MHz
    // round up if necessary
    if (cycles mod ticks) > (ticks div 2) then Inc(freq);
    total := (freq+freq2+freq3);   // total last three frequency calculations
  until ((tries < 3 ) or (tries < 20) and
        ((abs(3*freq -total) > (3*TOLERANCE)) or
        ( abs(3*freq2-total) > (3*TOLERANCE)) or
        ( abs(3*freq3-total) > (3*TOLERANCE))));
  if iPriority <> $7FFFFFFF then SetThreadPriority(hThread, iPriority);
  result := (total_cycles div total_ticks);
end;


function CPU86GetProcessorSpeed (): LongWord;
var
  cpuSpeed, i, tries: LongWord;
const
  known_speeds: array [0..23] of LongWord = (
    600, 550, 500, 450, 400, 375, 333, 300,
    266, 233, 200, 166, 150, 133, 120, 100,
    90, 75, 66, 60, 50, 33, 20, 0);
begin
  if not CPU86HaveRDTSC() then begin result := 0; exit; end;
  // Try 3 times for insurance...
  for tries := 1 to 3 do
   begin
    cpuSpeed := CPU86GetRDTSCSpeed();
    i := 0;
    repeat
     if (cpuSpeed >= (known_speeds[i]-2)) and (cpuSpeed <= (known_speeds[i]+2)) then
     begin
       result := known_speeds[i];
       exit;
     end;
     Inc(i);
    until known_speeds[i] = 0;
  end;
  result := cpuSpeed;
end;


procedure HPTimerOn (out tmStart: TX86LargeInteger);
begin
  if haveRDTSC then
  begin
    CPU86ReadTimeStamp(tmStart);
  end
  else
  begin
    if havePerformanceCounter then QueryPerformanceCounter(tmStart)
    else begin tmStart.lowPart := timeGetTime(); tmStart.highPart := 0; end;
  end;
end;

function HPTimerLap (const tmStart: TX86LargeInteger): LongWord;
var
  tmLap, tmCount: TX86LargeInteger;
begin
  if haveRDTSC then
  begin
    CPU86ReadTimeStamp(tmLap);
    CPU86DiffTime64(tmStart, tmLap, tmCount);
    result := CPU86CalcMicroSec(tmCount, cpuSpeed);
  end
  else
  begin
    if havePerformanceCounter then
    begin
      QueryPerformanceCounter(tmLap);
      CPU86DiffTime64(tmStart, tmLap, tmCount);
      result := CPU86CalcMicroSec(tmCount, countFreq.lowPart);
    end
    else
    begin
      tmLap.lowPart := timeGetTime();
      result := (tmLap.lowPart-tmStart.lowPart)*1000;
    end;
  end;
end;

procedure HPTimerOff (out tmEnd: TX86LargeInteger);
begin
  if haveRDTSC then begin CPU86ReadTimeStamp(tmEnd); end
  else
  begin
    if havePerformanceCounter then QueryPerformanceCounter(tmEnd)
    else begin tmEnd.lowPart := timeGetTime(); tmEnd.highPart := 0; end;
  end;
end;

function HPTimerCount (const tmStart, tmEnd: TX86LargeInteger): LongWord;
var
  tmCount: TX86LargeInteger;
begin
  if haveRDTSC then
  begin
    CPU86DiffTime64(tmStart, tmEnd, tmCount);
    result:=CPU86CalcMicroSec(tmCount, cpuSpeed);
  end
  else
  begin
    if havePerformanceCounter then
    begin
      CPU86DiffTime64(tmStart, tmEnd, tmCount);
      result := CPU86CalcMicroSec(tmCount, countFreq.lowPart);
    end
    else result := (tmEnd.lowPart-tmStart.lowPart)*1000;
  end;
end;

function HPTimerResolution (): Double;
begin
  result := 1E-6;
end;


procedure HPTimerMainInit ();
begin
  if cpuSpeed = -1 then
  begin
    cpuSpeed := Integer(CPU86GetProcessorSpeed()*1000000);
    haveRDTSC := (CPU86HaveRDTSC()) and (cpuSpeed > 0);
  end;
  havePerformanceCounter := QueryPerformanceFrequency(countFreq);
end;


begin
  cpuSpeed := -1;
  //MainInit();
end.
