// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// Speccy window interface
//
unit dspWinU;

interface

uses
  Windows, Messages, SysUtils, Classes, dspVScrU, dspEmuVarsU, dspEmuMouseU,
  alisp3U;


const
  winCursorImg: packed array [0..1, 0..11, 0..7] of Byte = (
   ((2,2,0,0,0,0,0,0),
    (2,1,2,0,0,0,0,0),
    (2,1,1,2,0,0,0,0),
    (2,1,1,1,2,0,0,0),
    (2,1,1,1,1,2,0,0),
    (2,1,1,1,1,1,2,0),
    (2,1,1,1,1,1,1,2),
    (2,1,1,1,1,1,1,2),
    (2,1,1,1,2,2,2,2),
    (2,1,1,2,1,1,2,0),
    (2,2,2,2,1,1,2,0),
    (0,0,0,2,2,2,2,0)),

   ((2,2,0,0,0,0,0,0),
    (2,1,2,2,0,0,0,0),
    (2,1,1,1,2,2,0,0),
    (2,1,1,1,1,1,2,2),
    (2,1,1,1,1,1,1,2),
    (2,1,1,1,2,2,1,2),
    (2,1,1,2,1,1,2,0),
    (2,1,1,2,1,1,2,0),
    (2,2,2,2,2,2,2,0),
    (0,0,0,0,0,0,0,0),
    (0,0,0,0,0,0,0,0),
    (0,0,0,0,0,0,0,0)));


// ctrls: prefix of the control variables
//
// paint event:
//  events.paint(ctrls)
//  events.prepaint(ctrls)
//  events.postpaint(ctrls)
//
// messages:
// events.<messagename>(ctrls msgname msx msy msb ib ii ic it)

const
  zxaLeft = $01;
  zxaRight = $02;
  zxaHCenter = $03;
  zxaTop = $10;
  zxaBottom = $20;
  zxaVCenter = $30;


type
  TZXControl = class;

  PZXEvent = ^TZXEvent;
  TZXEvent = record
    msg: string;
    dest: TZXControl;   // nil -- broadcast

    msX, msY: Integer;  // global
    msButtons: Byte;

    iBool: Boolean;
    iInt: Integer;
    iText: string;
  end;

  TZXControlList = class
  private
    fList: TList;

    function GetCount (): Integer;
    function GetItems (n: Integer): TZXControl;
    procedure SetItems (n: Integer; i: TZXControl);

  public
    constructor Create ();
    destructor Destroy (); override;

    procedure Clear ();
    procedure Add (i: TZXControl);
    procedure Delete (n: Integer); // Free()s <i>
      // accepts -1
    procedure Extract (n: Integer); // Delete, but w/o Free()
      // accepts -1
    procedure Insert (before: Integer; i: TZXControl);
    procedure Replace (n: Integer; i: TZXControl);
      // Free() previous item

    function Find (i: TZXControl): Integer;
    function FindByName (const name: string; stIndex, dir: Integer): Integer;
    function FindByCoords (x, y, stIndex, dir: Integer): Integer;

    property Count: Integer read GetCount;
    property Items [n: Integer]: TZXControl read GetItems write SetItems; default;
  end;


  TZXControl = class
  protected
    fOwner: TZXControl;
    fChildren: TZXControlList;

    procedure AddControl (ctrl: TZXControl);
    procedure DeleteControl (ctrl: TZXControl);

    function DoEvent (const name: string; var event: TZXEvent): Integer;
    procedure DoPaintEvent (const name: string);

  public
    constructor Create (aOwner: TZXControl);
    destructor Destroy (); override;

    procedure CoordsLocalToScr (var x, y: Integer); virtual;
    procedure CoordsScrToLocal (var x, y: Integer); virtual;

    function IsAccelerator (key: Integer): Boolean;

    function ControlAtXY (x, y: Integer): TZXControl;

    function GetPropertyAtom (const name: string): Integer;
    function GetIntProperty (const name: string): Integer;
    function GetStrProperty (const name: string): string;
    function GetBoolProperty (const name: string): Boolean;

    procedure SetPropertyAtom (const name: string; v: Integer);
    procedure SetIntProperty (const name: string; v: Integer);
    procedure SetStrProperty (const name: string; const v: string);
    procedure SetBoolProperty (const name: string; v: Boolean);

    procedure Paint (); virtual;
    procedure EventHandler (var event: TZXEvent); virtual;

    procedure Show ();
    procedure Hide ();
    function Visible (): Boolean;
    function Enabled (): Boolean;
    procedure Suicide ();
    function Active (): Boolean;

    function Left (): Integer;
    function Top (): Integer;
    function Width (): Integer;
    function Height (): Integer;
    function Name (): string;
  end;

  TZXTextInput = class (TZXControl)
  public
    constructor Create (aOwner: TZXControl);

    procedure Paint (); override;
    procedure EventHandler (var event: TZXEvent); override;
  end;

  TZXWindow = class (TZXControl)
  public
    constructor Create ();
    destructor Destroy (); override;

    procedure CoordsLocalToScr (var x, y: Integer); override;
    procedure CoordsScrToLocal (var x, y: Integer); override;

    procedure Activate ();
    procedure Close ();

    procedure Paint (); override;
    procedure EventHandler (var event: TZXEvent); override;
  end;


procedure DrawFrame (x, y, wdt, hgt: Integer; frC, fillC: Byte);
procedure DrawAlignedText (x, y, wdt, hgt: Integer; s: string; accelPos: Integer;
  align: Integer; fc: Byte; bc: Byte = 255);

function StrExtractAccel (const s: string; out accelPos: Integer): string;

procedure ZXWindowsDraw ();
procedure ZXWindowsMsg (var msg: TMsg);
procedure ZXWindowsPostMsg (name: string; dest: TZXControl;
  pBool: Boolean; pInt: Integer; pText: string;
  msX: Integer = -1; msY: Integer = -1; msB: Integer = -1);
procedure ZXWindowsProcessMessages ();


var
  zxWinList: TZXControlList = nil;
  zxWinActive: Boolean = true;


implementation


var
  zxwInternalActive: Boolean = false;
  zxwMsX, zxwMsY, zxwMsB: Integer;
  zxwMsgQueue: TList = nil;
  wMoving: TZXControl = nil;
  mvX, mvY: Integer;
  wasHidden: Boolean = false;


function StrExtractAccel (const s: string; out accelPos: Integer): string;
var
  f: Integer;
begin
  result := s;
  f := 1; accelPos := -1;
  while f <= Length(result) do
  begin
    if result[f] = '\' then Delete(result, f, 1)
    else if (accelPos = -1) and (result[f] = '&') then begin Delete(result, f, 1); accelPos := f; end;
    Inc(f);
  end;
end;

function PointInControl (x, y: Integer; v: TZXControl): Boolean;
var
  l, t: Integer;
begin
  l := v.Left; t := v.Top;
  result := (x >= l) and (y >= t) and
    (x < l+v.Width) and (y < t+v.Height);
end;

procedure DrawFrame (x, y, wdt, hgt: Integer; frC, fillC: Byte);
begin
  VFillRect(x, y, wdt, hgt, fillC, 255);
  if frC = 255 then exit;
  VVLine(x, y, hgt, frC);
  VVLine(x+wdt-1, y, hgt, frC);
  VHLine(x, y, wdt, frC);
  VHLine(x, y+hgt-1, wdt, frC);
end;

procedure DrawAlignedText (x, y, wdt, hgt: Integer; s: string; accelPos: Integer;
  align: Integer; fc: Byte; bc: Byte = 255);
begin
  if (align and $0F) = zxaHCenter then Inc(x, (wdt-Length(s)*6) div 2)
  else if (align and $0F) = zxaRight then Inc(x, wdt-Length(s)*6);

  if (align and $F0) = zxaVCenter then y := y+(hgt-8) div 2
  else if (align and $F0) = zxaBottom then y := y+hgt-8;

  VDrawString(x, y, s, fc, bc);
  if accelPos > 0 then VHLine(x+accelPos*6-6, y+7, 6, fc);
end;


{ TZXControlList }
constructor TZXControlList.Create ();
begin
  fList := TList.Create();
end;

destructor TZXControlList.Destroy ();
begin
  Clear();
  fList.Free();
end;

procedure TZXControlList.Clear ();
var
  f: Integer;
begin
  f := fList.Count-1;
  while f >= 0 do begin TZXControl(fList[f]).Free(); Dec(f); end;
  fList.Clear();
end;

function TZXControlList.GetCount (): Integer;
begin
  result := fList.Count;
end;

function TZXControlList.GetItems (n: Integer): TZXControl;
begin
  if (n < 0) or (n >= fList.Count) then result := nil
  else result := TZXControl(fList[n]);
end;

procedure TZXControlList.SetItems (n: Integer; i: TZXControl);
begin
  fList[n] := i;
end;

procedure TZXControlList.Replace (n: Integer; i: TZXControl);
begin
  TZXControl(fList[n]).Free();
  fList[n] := i;
end;

procedure TZXControlList.Add (i: TZXControl);
begin
  fList.Add(i);
end;

procedure TZXControlList.Delete (n: Integer); // Free()s <i>
// accepts -1
begin
  if n <> -1 then begin TZXControl(fList[n]).Free(); fList.Delete(n); end;
end;

procedure TZXControlList.Extract (n: Integer); // Delete, but w/o Free()
// accepts -1
begin
  if n <> -1 then fList.Delete(n);
end;

procedure TZXControlList.Insert (before: Integer; i: TZXControl);
begin
  fList.Insert(before, i);
end;

function TZXControlList.Find (i: TZXControl): Integer;
begin
  result := fList.Count-1;
  while (result >= 0) and (fList[result] <> i) do Dec(result);
end;

function TZXControlList.FindByName (const name: string; stIndex, dir: Integer): Integer;
begin
  if dir < 0 then dir := -1 else dir := 1;
  result := stIndex;
  if result < 0 then result := 0;
  if result >= fList.Count then result := fList.Count-1;
  while (result >= 0) and (result < fList.Count) do
  begin
    if TZXControl(fList[result]).Name = name then exit;
    Inc(result, dir);
  end;
  result := -1;
end;

function TZXControlList.FindByCoords (x, y, stIndex, dir: Integer): Integer;
begin
  if dir < 0 then dir := -1 else dir := 1;
  result := stIndex;
  if result < 0 then result := 0;
  if result >= fList.Count then result := fList.Count-1;
  while (result >= 0) and (result < fList.Count) do
  begin
    if PointInControl(x, y, TZXControl(fList[result])) then
    begin
      exit;
    end;
    Inc(result, dir);
  end;
  result := -1;
end;


{ TZXControl }
constructor TZXControl.Create (aOwner: TZXControl);
begin
  fChildren := TZXControlList.Create();
  try
    if aOwner <> nil then aOwner.AddControl(self);
  except
    fChildren.Free();
    raise;
  end;
  //fOwner := aOwner;
  if fOwner <> nil then
    SetStrProperty('owner', IntToHex(Integer(fOwner), 8)+'.');
  //else SetStrProperty('owner', '');
  SetIntProperty('left', 4); SetIntProperty('top', 4);
  SetIntProperty('width', 52); SetIntProperty('height', 12);
  SetBoolProperty('enabled', true); SetBoolProperty('visible', true);
  SetBoolProperty('clickable', true);
  SetIntProperty('align.caption', zxaHCenter or zxaVCenter);
  SetIntProperty('colors.frame', 15);
  SetIntProperty('colors.back', 0);
  SetIntProperty('colors.caption', 15);
  SetIntProperty('colors.disabled.frame', 6);
  SetIntProperty('colors.disabled.back', 0);
  SetIntProperty('colors.disabled.caption', 6);
  SetIntProperty('colors.hot.frame', 15);
  SetIntProperty('colors.hot.back', 15);
  SetIntProperty('colors.hot.caption', 0);
end;


type
  PEnumInfo = ^TEnumInfo;
  TEnumInfo = record
    prefix: string;
    sl: TStringList;
  end;

function SymbolEnumerator (const name: string; value: Integer; user: Pointer): Boolean;
var
  f, l: Integer;
begin
  l := Length(PEnumInfo(user).prefix);
  if Length(name) >= l then
  begin
    f := 1;
    while (f <= l) and (PEnumInfo(user).prefix[f] = name[f]) do Inc(f);
    if f > l then PEnumInfo(user).sl.AddObject(name, Pointer(value));
  end;
  result := true;
end;

destructor TZXControl.Destroy ();
var
  f: Integer;
  e: TZXEvent;
  sl: TStringList;
  inf: TEnumInfo;
begin
  for f := 0 to fChildren.Count-1 do
  begin
    e.msg := 'suicide'; e.dest := fChildren[f];
    e.msX := zxwMsX; e.msY := zxwMsY; e.msButtons := zxwMsB;
    e.iBool := false; e.iInt := 0; e.iText := '';
    fChildren[f].EventHandler(e);
  end;

  fChildren.Free();

  // delete all control-related symbols
  //ALDeleteGlobalSymbols(IntToHex(Integer(self), 8)+'.');
  sl := TStringList.Create();
  try
    inf.prefix := IntToHex(Integer(self), 8)+'.';
    inf.sl := sl;
    ALEnumSymbolsInCurrentContext(SymbolEnumerator, @inf);
    for f := sl.Count-1 downto 0 do ALDeleteSymbolFromCurrentContext(sl[f]);
  finally
    sl.Free();
  end;

  if fOwner <> nil then fOwner.DeleteControl(self);
end;

function TZXControl.IsAccelerator (key: Integer): Boolean;
var
  s: string;
  f: Integer;
  ch: Char;
begin
  result := false;
  if (key < 1) or (key > 255) then exit;
  if GetBoolProperty('noaccel') then exit;
  s := StrExtractAccel(GetStrProperty('caption'), f);
  if f < 1 then exit;
  ch := UpCase(s[f]);
  if not (ch in ['0'..'9', 'A'..'Z']) then exit;
  result := (key = ord(ch));
end;

function TZXControl.GetPropertyAtom (const name: string): Integer;
begin
  result := ALCdr(ALFindGlobalSymbol(IntToHex(Integer(self), 8)+'.'+name));
end;

function TZXControl.GetIntProperty (const name: string): Integer;
var
  n: Integer;
begin
  n := GetPropertyAtom(name);
  if ALGetAtomType(n) = DT_INTEGER then result := ALGetAtomInt(n)
  else result := 0;
end;

function TZXControl.GetStrProperty (const name: string): string;
var
  n: Integer;
begin
  n := GetPropertyAtom(name);
  if ALGetAtomType(n) = DT_STRING then result := ALGetAtomStr(n)
  else result := '';
end;

function TZXControl.GetBoolProperty (const name: string): Boolean;
var
  n: Integer;
begin
  n := GetPropertyAtom(name);
  result := not (ALIsNil(n) or ALIsFalse(n));
end;

procedure TZXControl.SetPropertyAtom (const name: string; v: Integer);
begin
  //ALAddGlobalSymbol(IntToHex(Integer(self), 8)+'.'+name, v);
  ALSetGlobalSymbol(IntToHex(Integer(self), 8)+'.'+name, v, true);
end;

procedure TZXControl.SetIntProperty (const name: string; v: Integer);
begin
  SetPropertyAtom(name, ALNewIntAtom(v));
end;

procedure TZXControl.SetStrProperty (const name: string; const v: string);
begin
  if v = '' then SetPropertyAtom(name, ALNil())
  else SetPropertyAtom(name, ALNewStrAtom(v));
end;

procedure TZXControl.SetBoolProperty (const name: string; v: Boolean);
begin
  if v then SetPropertyAtom(name, ALTrue())
  else SetPropertyAtom(name, ALFalse());
end;

procedure TZXControl.AddControl (ctrl: TZXControl);
begin
  ASSERT(ctrl.fOwner = nil);
  fChildren.Add(ctrl);
  ctrl.fOwner := self;
end;

procedure TZXControl.DeleteControl (ctrl: TZXControl);
begin
  ASSERT(ctrl.fOwner = self);
  fChildren.Extract(fChildren.Find(ctrl));
  ctrl.fOwner := nil;
end;

function TZXControl.Left (): Integer;
begin
  result := GetIntProperty('left');
end;

function TZXControl.Top (): Integer;
begin
  result := GetIntProperty('top');
end;

function TZXControl.Width (): Integer;
begin
  result := GetIntProperty('width');
end;

function TZXControl.Height (): Integer;
begin
  result := GetIntProperty('height');
end;

function TZXControl.Name (): string;
begin
  result := GetStrProperty('name');
end;

procedure TZXControl.Show ();
begin
  SetBoolProperty('visible', true);
end;

procedure TZXControl.Hide ();
begin
  SetBoolProperty('visible', false);
end;

function TZXControl.Visible (): Boolean;
begin
  result := GetBoolProperty('visible');
end;

function TZXControl.Enabled (): Boolean;
begin
  result := GetBoolProperty('enabled');
end;

procedure TZXControl.Suicide ();
begin
  ZXWindowsPostMsg('suicide', self, false, 0, '');
end;

function TZXControl.Active (): Boolean;
var
  w: TZXControl;
  f: Integer;
begin
  if fOwner <> nil then
  begin
    result := fOwner.Active();
    exit;
  end;
  result := not (self is TZXWindow);
  if result then exit;
  w := nil;
  f := zxWinList.Count-1;
  while (f >= 0) and not zxWinList[f].Visible do Dec(f);
  if f >= 0 then w := zxWinList[f];
  result := (w = self);
end;

procedure TZXControl.CoordsLocalToScr (var x, y: Integer);
begin
  Inc(x, Left); Inc(y, Top);
  if fOwner <> nil then fOwner.CoordsLocalToScr(x, y);
end;

procedure TZXControl.CoordsScrToLocal (var x, y: Integer);
begin
  if fOwner <> nil then fOwner.CoordsScrToLocal(x, y);
  Dec(x, Left); Dec(y, Top);
end;

function TZXControl.ControlAtXY (x, y: Integer): TZXControl;
var
  lx, ly: Integer;
  f: Integer;
begin
  CoordsLocalToScr(x, y);
  f := fChildren.Count-1;
  while f >= 0 do
  begin
    if fChildren[f].Visible then
    begin
      lx := x; ly := y; fChildren[f].CoordsScrToLocal(lx, ly);
      if (lx >= 0) and (ly >= 0) and
         (lx < fChildren[f].Width) and (ly < fChildren[f].Height) then break;
    end;
    Dec(f);
  end;
  if f >= 0 then result := fChildren[f] else result := nil;
end;

procedure TZXControl.Paint ();
var
  gsb: packed array [0..vClipOriginSaveBufSize-1] of Byte;
  f, ox, oy, w, h: Integer;
  frc, backc, textc: Integer;
  ds, s: string;
  clickable: Boolean;
begin
  if not GetBoolProperty('visible') then exit;
  ox := 0; oy := 0; w := Width; h := Height;
  CoordsLocalToScr(ox, oy);
  clickable := GetBoolProperty('clickable');
  ds := '';
  if Enabled then
  begin
    if clickable and zxwInternalActive and Active() then
    begin
      if (emuMouseX >= ox) and (emuMouseY >= oy) and
         (emuMouseX < ox+w) and (emuMouseY < oy+h) then ds := 'hot.';
    end;
  end
  else ds := 'disabled.';
  VSaveClipOrigin(gsb);
  //VResetOrigin(); VResetClip();
  VSetOrigin(ox, oy);
  VSetClip(ox, oy, ox+w-1, oy+h-1);
  try
    frc := GetIntProperty('colors.'+ds+'frame');
    backc := GetIntProperty('colors.'+ds+'back');
    DrawFrame(0, 0, w, h, frc, backc);
    if frc <> 255 then
    begin
      Dec(w, 2); Dec(h, 2);
      VSetOrigin(ox+1, oy+1);
      VSetClip(ox+1, oy+1, ox+w, oy+h);
    end;

    f := GetPropertyAtom('events.paint');
    if ALIsNil(f) then
    begin
      DoPaintEvent('events.prepaint');
      textc := GetIntProperty('colors.'+ds+'caption');
      s := GetStrProperty('caption');
      if not GetBoolProperty('noaccel') then s := StrExtractAccel(s, f)
      else f := -1;
      DrawAlignedText(0, 0, w, h, s, f, GetIntProperty('align.caption'), textc);
      //DoPaintEvent('events.postpaint');
    end
    else DoPaintEvent('events.paint');

    for f := 0 to fChildren.Count-1 do fChildren[f].Paint();
  finally
    VRestoreClipOrigin(gsb);
  end;
end;

procedure TZXControl.DoPaintEvent (const name: string);
var
  s: string;
  f: Integer;
begin
  f := GetPropertyAtom('events.'+name);
  if not ALIsNil(f) then
  begin
    s := '(eval (list (get-sym "'+IntToHex(Integer(self), 8)+'.events.'+name+'") ''"'+
      IntToHex(Integer(self), 8)+'."))';
    //WriteLn('<', s, '>');
    ALEvalStr(s);
    // events.<messagename>(ctrls)
  end;
end;

function TZXControl.DoEvent (const name: string; var event: TZXEvent): Integer;
const
  bstr: array [Boolean] of string = ('*f*', '*t*');
var
  s: string;
  f: Integer;
begin
  f := GetPropertyAtom('events.'+name);
  if not ALIsNil(f) then
  begin
    (*s := '(eval (list (get-sym "'+IntToHex(Integer(self), 8)+'.events.'+name+'") ''"'+
      IntToHex(Integer(self), 8)+'." ''"'+
      event.msg+'" '''+IntToStr(event.msX)+' '+IntToStr(event.msY)+' '''+
      IntToStr(event.msButtons)+' '''+bstr[event.iBool]+' '''+
      IntToStr(event.iInt)+' '''+s+'))';*)
    s := '('+ALEscapeStr(IntToHex(Integer(self), 8)+'.events.'+name)+
      ' "'+IntToHex(Integer(self), 8)+'."'+
      ' "'+event.msg+'" '+IntToStr(event.msX)+' '+IntToStr(event.msY)+' '+
      IntToStr(event.msButtons)+' '+bstr[event.iBool]+' '+
      IntToStr(event.iInt)+' '+ALEscapeStrEx(event.iText, true, '"')+')';
    //WriteLn('<', s, '>');
    result := ALEvalStr(s);
    // events.<messagename>(ctrls msgname msx msy msb ib ii ic it)
  end
  else result := ALNil();
end;

procedure TZXControl.EventHandler (var event: TZXEvent);
var
  f: Integer;
begin
  if event.dest = nil then DoEvent(event.msg, event);

  if event.dest = self then
  begin
    //WriteLn('<', IntToHex(Integer(self), 8), '>: ', event.msg);
    //if event.msg = 'suicide' then exit;
    f := DoEvent(event.msg, event);
    if (event.msg = 'die') and (ALIsNil(f) or ALIsFalse(f)) then
      ZXWindowsPostMsg('suicide', self, false, 0, '');
  end
  else
  begin
    if event.msg = 'suicide' then
    begin
      f := 0;
      while f < fChildren.Count do
      begin
        if fChildren[f] = event.dest then
        begin
          fChildren[f].EventHandler(event);
          fChildren[f].Free(); f := 0;
        end
        else Inc(f);
      end;
    end
    else for f := 0 to fChildren.Count-1 do fChildren[f].EventHandler(event);
  end;
end;


{ TZXTextInput }
constructor TZXTextInput.Create (aOwner: TZXControl);
begin
  inherited Create(aOwner);
  SetBoolProperty('wantkeys', true);
  SetBoolProperty('noaccel', true);
  SetBoolProperty('clickable', false);
  SetIntProperty('colors.caption', 14);
  SetIntProperty('align.caption', zxaVCenter or zxaLeft);
end;

procedure TZXTextInput.Paint ();
var
  s, ns: string;
  w, cw: Integer;
begin
  s := GetStrProperty('caption'); ns := s+#249;
  cw := GetIntProperty('width')-2; w := Length(ns)*6;
  if w <= cw then
  begin
    SetIntProperty('align.caption', zxaLeft or zxaVCenter);
  end
  else
  begin
    SetIntProperty('align.caption', zxaRight or zxaVCenter);
    w := cw div 6; if cw mod 6 = 0 then Inc(w);
    ns := Copy(ns, Length(ns)-w+1, w);
  end;
  SetStrProperty('caption', ns);
  inherited;
  SetStrProperty('caption', s);
end;

procedure TZXTextInput.EventHandler (var event: TZXEvent);
var
  s: string;
begin
  if event.dest <> self then exit;
  if event.msg = 'char' then
  begin
    //if event.iText = '&' then event.iText := '\&';
    s := GetStrProperty('caption');
    if event.iText = #8 then Delete(s, Length(s), 1)
    else if event.iText = ^Y then s := ''
    else if event.iText >= ' ' then s := s+event.iText;
    SetStrProperty('caption', s);
  end;
end;

{ TZXWindow }
constructor TZXWindow.Create ();
begin
  inherited Create(nil);
  SetIntProperty('colors.caption', 0);
  SetIntProperty('colors.frame', 15);
  SetIntProperty('colors.back', 1);
  SetIntProperty('colors.caption', 0);
  SetIntProperty('colors.inactive.frame', 6);
  SetIntProperty('colors.inactive.back', 1);
  SetIntProperty('colors.inactive.caption', 0);
  SetIntProperty('align.caption', zxaHCenter or zxaVCenter);
  zxWinList.Add(self);

  if zxWinList.Count = 1 then
  begin
    wasHidden := not emuMouseHidden;
    if wasHidden then ShowCursor(false);
  end;
end;

destructor TZXWindow.Destroy ();
begin
  inherited Destroy();
  zxWinList.Extract(zxWinList.Find(self));
  if zxWinList.Count = 0 then
  begin
    if wasHidden then ShowCursor(true);
  end;
end;

procedure TZXWindow.CoordsLocalToScr (var x, y: Integer);
begin
  //Inc(y, 8);
  inherited CoordsLocalToScr(x, y);
end;

procedure TZXWindow.CoordsScrToLocal (var x, y: Integer);
begin
  inherited CoordsScrToLocal(x, y);
  //Dec(y, 8);
end;

procedure TZXWindow.Activate ();
var
  f: Integer;
begin
  Show();
  f := zxWinList.Find(self);
  if (f = -1) or (f = zxWinList.Count-1) then exit;
  zxWinList.fList.Move(f, zxWinList.Count-1);
end;

procedure TZXWindow.Close ();
begin
  Suicide();
end;

procedure TZXWindow.Paint ();
var
  gsb: packed array [0..vClipOriginSaveBufSize-1] of Byte;
  f, ox, oy, w, h: Integer;
  frc, backc, textc: Integer;
  ds: string;
begin
  if not GetBoolProperty('visible') then exit;
  ox := 0; oy := 0; w := Width; h := Height;
  CoordsLocalToScr(ox, oy);
  ds := '';
  if not Active() then ds := 'inactive.';
  VSaveClipOrigin(gsb);

  VSetOrigin(0, 0);
  VSetClip(ox-1, oy-1, ox+w, oy+h);
  VVLine(ox-1, oy-1, h+2, 0);
  VVLine(ox+w, oy-1, h+2, 0);
  VHLine(ox-1, oy-1, w+2, 0);
  VHLine(ox-1, oy+h, w+2, 0);

  VSetOrigin(ox, oy);
  VSetClip(ox, oy, ox+w-1, oy+h-1);
  try
    frc := GetIntProperty('colors.'+ds+'frame');
    backc := GetIntProperty('colors.'+ds+'back');
    textc := GetIntProperty('colors.'+ds+'caption');
    //VFillRect(0, 0, w, h, backc, frc);
    DrawFrame(0, 0, w, h, frc, backc);
    if frc <> 255 then
    begin
      Dec(w, 2); Dec(h, 2); Inc(ox); Inc(oy);
      VSetOrigin(ox, oy);
      VSetClip(ox, oy, ox-1+w, oy+7);
      VFillRect(0, 0, w, 8, frc, 255);
      DrawAlignedText(0, 0, w, 8, GetStrProperty('caption'), -1,
        GetIntProperty('align.caption'), textc);
      VSetOrigin(ox, oy+8);
      VSetClip(ox, oy+8, ox-1+w, oy-1+h);
    end;

    DoPaintEvent('events.prepaint');
    for f := 0 to fChildren.Count-1 do fChildren[f].Paint();
    //DoPaintEvent('events.postpaint');
  finally
    VRestoreClipOrigin(gsb);
  end;
end;

procedure TZXWindow.EventHandler (var event: TZXEvent);
var
  f: Integer;
  cc: TZXControl;
begin
  if event.dest = self then
  begin
    if event.msg = 'click' then
    begin
      cc := ControlAtXY(event.msX-Left, event.msY-Top);
      //WriteLn(Integer(cc));
      if (cc <> nil) and cc.Enabled then
        ZXWindowsPostMsg('click', cc, false, 0, '', event.msX, event.msY, event.msButtons);
      exit;
    end
    else if event.msg = 'keydown' then
    begin
      if (event.iInt = VK_RETURN) or (event.iInt = VK_ESCAPE) then
      begin
        f := fChildren.Count-1;
        while f >= 0 do
        begin
          if {fChildren[f].Visible and }fChildren[f].Active then
          begin
            if ((event.iInt = VK_RETURN) and fChildren[f].GetBoolProperty('default')) or
               ((event.iInt = VK_ESCAPE) and fChildren[f].GetBoolProperty('cancel')) then
            begin
              event.dest := fChildren[f];
              event.msg := 'click';
              fChildren[f].EventHandler(event);
            end;
          end;
          Dec(f);
        end;
        exit;
      end
      else if ((GetKeyState(VK_SHIFT) and $8000) = 0) and
              ((GetKeyState(VK_MENU) and $8000) <> 0) and
              ((GetKeyState(VK_CONTROL) and $8000) = 0) then
      begin
        f := fChildren.Count-1;
        while f >= 0 do
        begin
          if fChildren[f].Visible and fChildren[f].Active and
             fChildren[f].IsAccelerator(event.iInt) then
          begin
            event.dest := fChildren[f];
            event.msg := 'click';
            fChildren[f].EventHandler(event);
            break;
          end
          else Dec(f);
        end;
        exit;
      end;
    end;
    if ((event.msg = 'keydown') or (event.msg = 'keyup') or (event.msg = 'char')) and
       ((GetKeyState(VK_MENU) and $8000) = 0) then
    begin
      f := fChildren.Count-1;
      while f >= 0 do
      begin
        if fChildren[f].Visible and fChildren[f].Active and
           fChildren[f].GetBoolProperty('wantkeys') then
        begin
          event.dest := fChildren[f];
          fChildren[f].EventHandler(event);
          break;
        end;
        Dec(f);
      end;
      exit;
    end;
  end;
  inherited EventHandler(event);
end;


///////////////////////////////////////////////////////////
procedure DrawMsCursor ();
var
  curN: Integer;
begin
  if (emuMouseButtons <> 0) then curN := 1 else curN := 0;
  VDrawCursor(emuMouseX, emuMouseY, 8, 12, winCursorImg[curN]);
end;

procedure ZXWindowsDraw ();
var
  f: Integer;
begin
  zxwInternalActive := (GetForegroundWindow() = emuHWnd);
  for f := 0 to zxWinList.Count-1 do
  begin
    zxWinList[f].Paint();
  end;
  if zxwInternalActive and zxWinActive and (zxWinList.Count > 0) then DrawMsCursor();
end;

procedure ZXWindowsPostMsg (name: string; dest: TZXControl;
  pBool: Boolean; pInt: Integer; pText: string;
  msX: Integer = -1; msY: Integer = -1; msB: Integer = -1);
var
  msg: PZXEvent;
begin
  if msX = -1 then msX := zxwMsX;
  if msY = -1 then msY := zxwMsY;
  if msB = -1 then msB := zxwMsB;
  New(msg);
  try
    msg.msg := name; msg.dest := dest;
    msg.msX := msX; msg.msY := msY; msg.msButtons := msB;
    msg.iBool := pBool; msg.iInt := pInt; msg.iText := pText;
    zxwMsgQueue.Add(msg);
  except
    Dispose(msg);
    raise;
  end;
end;

procedure ZXWindowsMsg (var msg: TMsg);
var
  w, cc: TZXControl;
  f, ob: Integer;
  x, y, wx{, wy}: Integer;
  mps: TPoint;
  msg1: TMsg;
begin
  //Write(msg.message, ' ');
  zxwInternalActive := (GetForegroundWindow() = emuHWnd);
  if not zxwInternalActive then
  begin
    zxwMsX := 0; zxwMsY := 0; zxwMsB := 0; wMoving := nil;
    exit;
  end;
  w := nil;
  f := zxWinList.Count-1;
  while (f >= 0) and not zxWinList[f].Visible do Dec(f);
  if f >= 0 then w := zxWinList[f];
  case msg.message of
    WM_LBUTTONDOWN:
      begin
        ob := zxwMsB;
        zxwMsB := zxwMsB or $01;
        if w <> nil then ZXWindowsPostMsg('msdown', w, false, 0, '');
        if (w <> nil) and (ob = 0) and PointInControl(zxwMsX, zxwMsY, w) then
        begin
          //cc := w.ControlAtXY(zxwMsX-w.Left, zxwMsY-w.Top);
          //if cc = nil then cc := w;
          //if cc.Enabled then ZXWindowsPostMsg('click', cc, false, 0, '');
          //WriteLn('!');
          x := w.Left; y := w.Top;
          wx := w.Width; //wy := w.Height;
          if (zxwMsX >= x) and (zxwMsY >= y) and
             (zxwMsX < x+wx) and (zxwMsY <= y+10) then
          begin
            //WriteLn('moving!');
            if wMoving <> w then
            begin
              wMoving := w;
              mvX := zxwMsX-x; mvY := zxwMsY-y;
            end;
          end
          else ZXWindowsPostMsg('click', w, false, 0, '');
        end
        else if ob = 0 then
        begin
          f := maxInt;
          repeat
            f := zxWinList.FindByCoords(zxwMsX, zxwMsY, f, -1);
            if (f <> -1) and zxWinList[f].Visible then
            begin
              if (zxWinList[f] is TZXWindow) or zxWinList[f].Enabled then break;
            end;
          until f = -1;
          if f <> -1 then
          begin
            cc := zxWinList[f];
            if (cc is TZXWindow) then
            begin
              TZXWindow(cc).Activate();
            end
            else ZXWindowsPostMsg('click', cc, false, 0, '');
          end;
        end;
      end;
    WM_LBUTTONUP:
      begin
        wMoving := nil;
        zxwMsB := zxwMsB and (not $01);
        if w <> nil then ZXWindowsPostMsg('msup', w, false, 0, '');
      end;
    WM_RBUTTONDOWN:
      begin
        wMoving := nil;
        if (w <> nil) and (zxwMsB = 0) then ZXWindowsPostMsg('die', w, false, 0, '');
        zxwMsB := zxwMsB or $02;
      end;
    WM_RBUTTONUP:
      begin
        wMoving := nil;
        zxwMsB := zxwMsB and (not $02);
      end;
    WM_MOUSEMOVE:
      begin
        zxwMsX := LoWord(msg.lParam) div emuScale;
        zxwMsY := HiWord(msg.lParam) div emuScale;
        if wMoving <> nil then
        begin
          if zxWinList.Find(wMoving) = -1 then wMoving := nil
          else
          begin
            wMoving.SetIntProperty('left', zxwMsX-mvX);
            wMoving.SetIntProperty('top', zxwMsY-mvY);
          end;
        end;
      end;
    WM_KEYDOWN:
      begin
        if w <> nil then
        begin
          x := 0; y := 0;
          case msg.wParam and $FFFF of
            VK_LEFT: x := -1;
            VK_RIGHT: x := 1;
            VK_UP: y := -1;
            VK_DOWN: y := 1;
            VK_HOME: begin x := -1; y := -1; end;
            VK_PRIOR: begin x := 1; y := -1; end;
            VK_END: begin x := -1; y := 1; end;
            VK_NEXT: begin x := 1; y := 1; end;
            VK_INSERT:
              begin
                msg1.message := WM_LBUTTONDOWN;
                ZXWindowsMsg(msg1);
                exit;
              end;
            VK_DELETE:
              begin
                msg1.message := WM_RBUTTONDOWN;
                ZXWindowsMsg(msg1);
                msg1.message := WM_RBUTTONUP;
                ZXWindowsMsg(msg1);
                exit;
              end;
          end;
          if (x <> 0) or (y <> 0) then
          begin
            GetCursorPos(mps);
            if (GetKeyState(VK_CONTROL) and $8000) <> 0 then f := 8 else f := 1;
            mps.x := mps.x+x*f*emuScale; mps.y := mps.y+y*f*emuScale;
            SetCursorPos(mps.x, mps.y);
          end
          else ZXWindowsPostMsg('keydown', w, false, msg.wParam and $FFFF, '');
        end;
     end;
    WM_KEYUP:
      begin
        if w <> nil then
        begin
          case msg.wParam and $FFFF of
            VK_INSERT:
              begin
                msg1.message := WM_LBUTTONUP;
                ZXWindowsMsg(msg1);
                exit;
              end;
          end;
          ZXWindowsPostMsg('keyup', w, false, msg.wParam and $FFFF, '');
        end;
      end;
    WM_CHAR:
      begin
        if w <> nil then ZXWindowsPostMsg('char', w, false, msg.wparam and $FF, chr(msg.wparam and $FF));
      end;

    //WM_KEYDOWN: WriteLn('key_down');
      //begin WriteLn('kf!'); zxwInternalActive := false; end;
    //WM_SETFOCUS: begin WriteLn('sf!'); zxwInternalActive := true; end;
    //WM_LBUTTONDOWN: begin emuMouseButtons := emuMouseButtons or 1; exit; end;
    //WM_LBUTTONUP: begin emuMouseButtons := emuMouseButtons and (not 1); exit; end;
    //WM_RBUTTONDOWN: begin emuMouseButtons := emuMouseButtons or 2; exit; end;
    //WM_RBUTTONUP: begin emuMouseButtons := emuMouseButtons and (not 2); exit; end;
    //WM_MBUTTONDOWN: begin emuMouseButtons := emuMouseButtons or 4; exit; end;
    //WM_MBUTTONUP: begin emuMouseButtons := emuMouseButtons and (not 4); exit; end;
    //WM_MOUSEMOVE:
    //WM_KEYDOWN:
    //WM_KEYUP:
    //WM_CHAR:
  end;
end;

procedure ZXWindowsProcessMessages ();
var
  f: Integer;
  pe: PZXEvent;
begin
  while zxwMsgQueue.Count > 0 do
  begin
    pe := PZXEvent(zxwMsgQueue[0]);
    //WriteLn('ZXM: ', pe.msg);
    f := 0; while f < zxWinList.Count do
    begin
      zxWinList[f].EventHandler(pe^);
      if (pe.msg = 'suicide') and (pe.dest = zxWinList[f]) then
      begin
        zxWinList[f].Free();
        //zxWinList.Delete(f);
        f := 0;
      end
      else Inc(f);
    end;
    Dispose(PZXEvent(zxwMsgQueue[0])); zxwMsgQueue.Delete(0);
  end;
end;


begin
  zxWinList := TZXControlList.Create();
  zxwMsX := 0; zxwMsY := 0; zxwMsB := 0;
  zxwMsgQueue := TList.Create();
end.
